/*
 *        spWave_Caf.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spKanji.h>

#include <sp/spChunkP.h>
#include <sp/spWave_Caf.h>

#ifdef SP_SUPPORT_64BIT_INT
spBool spEqCafChunkType(spCafChunk *chunk, const char *type)
{
    return (spBool)strneq(chunk->header.type, type, 4);
}

spChunkSize spGetCafChunkHeaderContentSize(spCafChunkHeader *chunk_header)
{
    return chunk_header->mChunkSize;
}

spBool spSetCafChunkHeaderContentSize(spCafChunkHeader *chunk_header, spChunkSize size)
{
    spDebug(100, "spSetCafChunkHeaderContentSize", "%c%c%c%c: set size = %ld\n",
            chunk_header->type[0], chunk_header->type[1], chunk_header->type[2], chunk_header->type[3], (long)size);
    
    chunk_header->mChunkSize = size;
    
    return SP_TRUE;
}

spBool spSetCafChunkContentSize(spCafChunk *chunk, spChunkSize size)
{
    return spSetCafChunkHeaderContentSize(&chunk->header, size);
}

spChunkSize spGetCafChunkContentSize(spCafChunk *chunk, char *string)
{
    if (chunk == NULL) return 0;
    
    if (string != NULL) {
        memcpy(string, chunk->header.type, 4);
        string[4] = NUL;
    }
    
    return spGetCafChunkHeaderContentSize(&chunk->header);
}

spChunkSize spGetCafChunkContentSizeForChunkFileSpec(spCafChunk *chunk)
{
    return spGetCafChunkHeaderContentSize(&chunk->header);
}

spChunkSize spGetCafChunkSize(spCafChunk *chunk, char *string)
{
    if (chunk == NULL) return 0;
    
    if (string != NULL) {
        memcpy(string, chunk->header.type, 4);
        string[4] = NUL;
    }
    
    if (strneq(chunk->header.type, "caff", 4)) {
        return chunk->header.mChunkSize + 8;
    } else {
        return chunk->header.mChunkSize + 12;
    }
}

spChunkSize spGetCafChunkHeaderHeaderSize(spCafChunkHeader *chunk_header)
{
    if (strneq(chunk_header->type, "caff", 4)) {
        return 8;
    } else {
        return 12;
    }
}

spChunkSize spGetCafChunkHeaderSize(spCafChunk *chunk)
{
    return spGetCafChunkHeaderHeaderSize(&chunk->header);
}

spChunkSize spGetCafChunkContentMarginSize(spCafChunk *chunk)
{
    return chunk->header.mMarginSize;
}

spBool spSetCafChunkContentMarginSize(spCafChunk *chunk, spChunkSize marginSize)
{
    spDebug(100, "spSetCafChunkContentMarginSize", "marginSize = %ld\n", (long)marginSize);
    chunk->header.mMarginSize = marginSize;
    return SP_TRUE;
}

spChunkSize spReadCafChunkHeader(spChunk *parent, spCafChunkHeader *chunk_header, int swap, FILE *fp)
{
    spChunkSize nread;

    memset(chunk_header, 0, sizeof(spCafChunkHeader));
    
    if (fread(chunk_header->type, 1, 4, fp) != 4) {
        spDebug(10, "spReadCafChunkHeader", "Can't read chunk type.\n");
        return 0;
    }
    nread = 4;
    spDebug(10, "spReadCafChunkHeader", "type = %c%c%c%c\n",
            chunk_header->type[0], chunk_header->type[1], chunk_header->type[2], chunk_header->type[3]);
    
    if (freadlong64(&chunk_header->mChunkSize, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafChunkHeader", "Can't read chunk size.\n");
        return 0;
    }
    nread += 8;
    spDebug(10, "spReadCafChunkHeader", "mChunkSize = %ld\n", (long)chunk_header->mChunkSize);
    
    return nread;
}

spChunkSize spWriteCafChunkHeader(spChunk *chunk, spCafChunkHeader *header, int swap, FILE *fp)
{
    long nwrite;
    spChunkSize total_nwrite;

    if (fwrite(header->type, 1, 4, fp) != 4) {
        spDebug(10, "spWriteCafChunkHeader", "Can't write chunk type.\n");
        return 0;
    }
    total_nwrite = 4;

    if (strneq(header->type, "caff", 4)) {
        spCafHeader *cafheader;

        spDebug(10, "spWriteCafChunkHeader", "top ('caff') chunk\n");
        
        cafheader = (spCafHeader *)chunk;
        
        if ((nwrite = fwriteshort((short *)&cafheader->mFileVersion, 1, SP_AIFF_NEED_SWAP, fp)) != 1) {
            spDebug(10, "spWriteCafChunkHeader", "Can't write mFileVersion: %ld\n", nwrite);
            return nwrite;
        }
        total_nwrite += 2;
    
        if ((nwrite = fwriteshort((short *)&cafheader->mFileFlags, 1, SP_AIFF_NEED_SWAP, fp)) != 1) {
            spDebug(10, "spWriteCafChunkHeader", "Can't write mFileFlags: %ld\n", nwrite);
            return nwrite;
        }
        total_nwrite += 2;
    } else {
        spLong64 chunkSize;
        
        spDebug(10, "spWriteCafChunkHeader", "mChunkSize = %ld, '%c%c%c%c' chunk\n",
                (long)header->mChunkSize, header->type[0], header->type[1], header->type[2], header->type[3]);

        chunkSize = header->mChunkSize;
        if (header->mMarginSize > 0) {
            chunkSize -= header->mMarginSize;
        }
        spDebug(10, "spWriteCafChunkHeader", "chunkSize = %ld, mMarginSize = %ld, '%c%c%c%c' chunk\n",
                (long)chunkSize, (long)header->mMarginSize, header->type[0], header->type[1], header->type[2], header->type[3]);
        
        if ((nwrite = fwritelong64(&chunkSize, 1, swap, fp)) != 1) {
            spDebug(10, "spWriteCafChunkHeader", "Can't write chunk size: %ld\n", nwrite);
            return nwrite;
        }
        total_nwrite += 8;
    }
    
    spDebug(10, "spWriteCafChunkHeader", "nwrite = %ld\n", (long)total_nwrite);
    
    return total_nwrite;
}

spChunkSize spSkipCafChunk(spChunk *parent, spCafChunkHeader *chunk_header, spChunkSize remain_size, int swap, FILE *fp)
{
    spChunkSize len;
    
    spDebug(10, "spSkipCafChunk", "chunk_header->mChunkSize = %ld, chunk_header->type = %c%c%c%c\n",
            (long)chunk_header->mChunkSize, chunk_header->type[0], chunk_header->type[1], chunk_header->type[2], chunk_header->type[3]);
    
    if (chunk_header->mChunkSize == 0) {
        if (remain_size < 0) {
            if (spSeekFile(fp, 0, SEEK_END) != 0) {
                return 0;
            }
        } else {
            if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
                return 0;
            }
        }
        len = 0;
    } else {
        len = chunk_header->mChunkSize;
        if (spSeekFile(fp, len, SEEK_CUR) != 0) {
            return 0;
        }
    }

    return len;
}

spChunkSize spWriteCafContentRest(spCafChunk *chunk, long depth, spChunkSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long nwrite;
    spLong64 nremain;
    spLong64 new_content_size;
    spLong64 total_nwrite;
    static char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};

    spDebug(100, "spWriteCafContentRest", "%c%c%c%c: remain_size = %ld\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
            (long)remain_size);
            
    if (strneq(chunk->header.type, "caff", 4) || strneq(chunk->header.type, "data", 4)) {
        return remain_size;
    }

    total_nwrite = 0;
    nremain = remain_size;
    new_content_size = chunk->header.mChunkSize - remain_size;
    
    if (nremain >= 12) {
        spDebug(10, "spWriteCafContentRest", "chunk->offset = %ld\n", (long)chunk->offset);
        if (chunk->offset > 0) {
            spFileOffset current_offset;

            current_offset = spTellFile(fp);
            spDebug(10, "spWriteCafContentRest", "current_offset = %ld, new_content_size = %ld\n",
                    (long)current_offset, (long)new_content_size);

            if (spSeekFile(fp, chunk->offset + 4, SEEK_SET) == 0) {
                fwritelong64(&new_content_size, 1, swap, fp);
                spSeekFile(fp, current_offset, SEEK_SET);
            }
        }
        
        if (fwrite("free", 1, 4, fp) != 4) {
            spDebug(10, "spWriteCafContentRest", "Can't write 'free' chunk type.\n");
            return 0;
        }
        total_nwrite += 4;
            
        nremain -= 12;
        if (fwritelong64(&nremain, 1, swap, fp) != 1) {
            spDebug(10, "spWriteCafContentRest", "Can't write chunk size.\n");
            return 0;
        }
        total_nwrite += 8;
        spDebug(100, "spWriteCafContentRest", "write 'free' header done: total_nwrite = %ld\n", (long)total_nwrite);
    }
        
    spDebug(100, "spWriteCafContentRest", "nremain = %ld\n", (long)nremain);
    
    while (nremain > 0) {
        len = MIN(8, (long)nremain);
        if ((nwrite = (long)fwrite(buf, 1, len, fp)) != len) {
            break;
        }

        nremain -= nwrite;
        total_nwrite += nwrite;
    }

    spDebug(100, "spWriteCafContentRest", "done: total_nwrite = %ld\n", (long)total_nwrite);
    
    return (spChunkSize)total_nwrite;
}

spChunkSize spReadCafAudioDescChunk(spChunk *parent, long depth, void *chunk, int swap, FILE *fp)
{
    spChunkSize nread;
    spCafAudioDescChunk *desc = (spCafAudioDescChunk *)chunk;

    spDebug(80, "spReadCafAudioDescChunk", "in\n");

    if (freaddouble((double *)&desc->mSampleRate, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mSampleRate in 'desc' chunk.\n");
        return 0;
    }
    nread = 8;
    spDebug(80, "spReadCafAudioDescChunk", "mSampleRate = %f\n", desc->mSampleRate);
    
    if (fread(desc->mFormatID, 1, 4, fp) != 4) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mFormatID in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mFormatID = %c%c%c%c\n",
            desc->mFormatID[0], desc->mFormatID[1], desc->mFormatID[2], desc->mFormatID[3]);
    
    if (freadlong32((long *)&desc->mFormatFlags, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mFormatFlags in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mFormatFlags = %lx\n", desc->mFormatFlags);
    
    if (freadlong32((long *)&desc->mBytesPerPacket, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mBytesPerPacket in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mBytesPerPacket = %ld\n", desc->mBytesPerPacket);
    
    if (freadlong32((long *)&desc->mFramesPerPacket, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mFramesPerPacket in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mFramesPerPacket = %ld\n", desc->mFramesPerPacket);
    
    if (freadlong32((long *)&desc->mChannelsPerFrame, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mChannelsPerFrame in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mChannelsPerFrame = %ld\n", desc->mChannelsPerFrame);
    
    if (freadlong32((long *)&desc->mBitsPerChannel, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mBitsPerChannel in 'desc' chunk.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafAudioDescChunk", "mBitsPerChannel = %ld\n", desc->mBitsPerChannel);
    
    spDebug(80, "spReadCafAudioDescChunk", "done: nread = %ld\n", (long)nread);
    
    return nread;
}

spChunkSize spWriteCafAudioDescChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spChunkSize nwrite;
    long value;
    spBool compressed = SP_FALSE;
    spCafAudioDescChunk *desc = (spCafAudioDescChunk *)chunk;

    spDebug(80, "spWriteCafAudioDescChunk", "in\n");

    spDebug(80, "spWriteCafAudioDescChunk", "mSampleRate = %f\n", desc->mSampleRate);
    if (fwritedouble((double *)&desc->mSampleRate, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mSampleRate in 'desc' chunk.\n");
        return 0;
    }
    nwrite = 8;

    if (!strneq(desc->mFormatID, "lpcm", 4)) {
        compressed = SP_TRUE;
    }
    
    spDebug(80, "spWriteCafAudioDescChunk", "mFormatID = %c%c%c%c\n",
            desc->mFormatID[0], desc->mFormatID[1], desc->mFormatID[2], desc->mFormatID[3]);
    if (fwrite(desc->mFormatID, 1, 4, fp) != 4) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mFormatID in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;
    
    spDebug(80, "spWriteCafAudioDescChunk", "mFormatFlags = %lx\n", desc->mFormatFlags);
    if (fwritelong32((long *)&desc->mFormatFlags, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mFormatFlags in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;
    
    spDebug(80, "spWriteCafAudioDescChunk", "mBytesPerPacket = %ld\n", desc->mBytesPerPacket);
    if (fwritelong32((long *)&desc->mBytesPerPacket, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mBytesPerPacket in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;

    if (compressed == SP_FALSE) {
        desc->mFramesPerPacket = 1;
    }
    spDebug(80, "spWriteCafAudioDescChunk", "mFramesPerPacket = %ld\n", desc->mFramesPerPacket);
    if (fwritelong32((long *)&desc->mFramesPerPacket, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mFramesPerPacket in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;
    
    spDebug(80, "spWriteCafAudioDescChunk", "mChannelsPerFrame = %ld\n", desc->mChannelsPerFrame);
    if (fwritelong32((long *)&desc->mChannelsPerFrame, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mChannelsPerFrame in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;
    
    if (compressed == SP_TRUE) {
        value = 0;
    } else {
        value = desc->mBitsPerChannel;
    }
    spDebug(80, "spWriteCafAudioDescChunk", "mBitsPerChannel = %ld\n", value);
    if (fwritelong32(&value, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mBitsPerChannel in 'desc' chunk.\n");
        return 0;
    }
    nwrite += 4;
    
    spDebug(80, "spWriteCafAudioDescChunk", "done: nwrite = %ld\n", (long)nwrite);
    
    return nwrite;
}

spChunkSize spReadCafGeneralDataChunk(spChunk *parent, long depth, void *chunk, int swap, FILE *fp)
{
    spChunkSize nread;
    spCafGeneralDataChunk *gdata = (spCafGeneralDataChunk *)chunk;

    spDebug(80, "spReadCafGeneralDataChunk", "in\n");

    gdata->allocFlag = SP_TRUE;
    gdata->mData = xalloc(gdata->header.mChunkSize, char);

    if ((nread = fread(gdata->mData, 1, (long)gdata->header.mChunkSize, fp)) <= 0) {
        spDebug(10, "spReadCafGeneralDataChunk", "Can't read data in general data chunk.\n");
        return 0;
    }
    
    spDebug(80, "spReadCafGeneralDataChunk", "done: nread = %ld\n", (long)nread);
    
    return nread;
}

spBool spFreeCafGeneralDataChunk(void *chunk)
{
    spCafGeneralDataChunk *gdata = (spCafGeneralDataChunk *)chunk;

    if (gdata->allocFlag && gdata->mData != NULL) {
        xfree(gdata->mData);
        gdata->mData = NULL;
    }
    
    return SP_TRUE;
}

spChunkSize spWriteCafGeneralDataChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spChunkSize nwrite;
    spCafGeneralDataChunk *gdata = (spCafGeneralDataChunk *)chunk;

    spDebug(80, "spWriteCafGeneralDataChunk", "in\n");

    if ((nwrite = fwrite(gdata->mData, 1, (long)gdata->header.mChunkSize, fp)) <= 0) {
        spDebug(10, "spWriteCafGeneralDataChunk", "Can't write data in general data chunk.\n");
        return 0;
    }
    
    spDebug(80, "spWriteCafGeneralDataChunk", "done: nwrite = %ld\n", (long)nwrite);
    
    return nwrite;
}

spChunkSize spCopyCafGeneralDataChunk(spChunk *dest_chunk, spChunk *src_chunk, long depth)
{
    spChunkSize ncopy = 0;
    spCafGeneralDataChunk *dest_gdata = (spCafGeneralDataChunk *)dest_chunk;
    spCafGeneralDataChunk *src_gdata = (spCafGeneralDataChunk *)src_chunk;

    spDebug(80, "spCopyCafGeneralDataChunk", "in\n");

    if (dest_gdata->header.mChunkSize > 0) {
        dest_gdata->allocFlag = SP_TRUE;
        ncopy = dest_gdata->header.mChunkSize;
        dest_gdata->mData = xalloc((long)ncopy, char);
        memcpy(dest_gdata->mData, src_gdata->mData, (long)ncopy);
    } else {
        dest_gdata->allocFlag = SP_FALSE;
        dest_gdata->mData = NULL;
    }
    ncopy += sizeof(dest_gdata->allocFlag) + sizeof(dest_gdata->mData);

    spDebug(80, "spCopyCafGeneralDataChunk", "done: ncopy = %ld\n", (long)ncopy);
    
    return ncopy;
}

void spUpdateCafPacketTableChunkBuffer(spCafAudioDescChunk *desc, spCafPacketTableChunk *pakt, spBool free_buffer)
{
    spLong64 newBufferSize;
    
    spDebug(80, "spUpdateCafPacketTableChunkBuffer", "mNumberPackets = %ld\n", (long)pakt->mNumberPackets);
    
    if (pakt->mNumberPackets > 0) {
        pakt->mTableSize = pakt->mNumberPackets;
        if (desc->mBytesPerPacket == 0 && desc->mFramesPerPacket == 0) {
            pakt->mTableSize *= 2;
        }
        spDebug(80, "spUpdateCafPacketTableChunkBuffer", "mNumberPackets = %ld, mTableSize = %ld\n",
                (long)pakt->mNumberPackets, (long)pakt->mTableSize);

        newBufferSize = pakt->mTableSize * sizeof(spLong64);

        if (newBufferSize > pakt->mBufferSize) {
            pakt->mBufferSize = newBufferSize;
            
            pakt->mBuffer = xrealloc(pakt->mBuffer, (long)pakt->mBufferSize, char);
            memset(pakt->mBuffer, 0, (long)pakt->mBufferSize);
            pakt->mTable = (spLong64 *)pakt->mBuffer;
        }
    } else {
        pakt->mTableSize = 0;
        pakt->mTable = NULL;

        if (free_buffer == SP_TRUE && pakt->mBuffer != NULL) {
            xfree(pakt->mBuffer);
            pakt->mBuffer = NULL;
            pakt->mBufferSize = 0;
        }
    }

    return;
}

spChunkSize spReadCafPacketTableChunk(spChunk *parent, long depth, void *chunk, int swap, FILE *fp)
{
    spChunkSize nread;
    spCafHeader *header = (spCafHeader *)parent;
    spCafPacketTableChunk *pakt = (spCafPacketTableChunk *)chunk;

    if (pakt == NULL || header == NULL || header->desc == NULL) return 0;

    spDebug(80, "spReadCafPacketTableChunk", "in\n");

    pakt->mBufferSize = 0;
    pakt->mBuffer = NULL;
    pakt->mTableSize = 0;
    pakt->mTable = NULL;

    if (freadlong64(&pakt->mNumberPackets, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafPacketTableChunk", "Can't read mNumberPackets.\n");
        return 0;
    }
    nread = 8;
    spDebug(80, "spReadCafPacketTableChunk", "mNumberPackets = %ld\n", (long)pakt->mNumberPackets);
    
    if (freadlong64(&pakt->mNumberValidFrames, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafPacketTableChunk", "Can't read mNumberValidFrames.\n");
        return 0;
    }
    nread += 8;
    spDebug(80, "spReadCafPacketTableChunk", "mNumberValidFrames = %ld\n", (long)pakt->mNumberValidFrames);
    
    if (freadlong32((long *)&pakt->mPrimingFrames, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafPacketTableChunk", "Can't read mPrimingFrames.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafPacketTableChunk", "mPrimingFrames = %ld\n", pakt->mPrimingFrames);

    if (freadlong32((long *)&pakt->mRemainderFrames, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafPacketTableChunk", "Can't read mRemainderFrames.\n");
        return 0;
    }
    nread += 4;
    spDebug(80, "spReadCafPacketTableChunk", "mRemainderFrames = %ld\n", pakt->mRemainderFrames);

    if (pakt->mNumberPackets > 0) {
        unsigned char c;
        spLong64 count;
        spLong64 value;

        spUpdateCafPacketTableChunkBuffer(header->desc, pakt, SP_FALSE);
        
        count = 0;
        value = 0;
        
        for (; nread < pakt->header.mChunkSize;) {
            if (fread(&c, 1, 1, fp) != 1) {
                spDebug(10, "spReadCafPacketTableChunk", "Can't read packet table in 'pakt' chunk.\n");
                return 0;
            }
            nread += 1;

            value = (value << 7) | (spLong64)(c & 0x7f);

            if (!(c & 0x80)) {
                pakt->mTable[count] = value;
                spDebug(100, "spReadCafPacketTableChunk", "nread = %ld, mTable[%ld] = %ld\n",
                        (long)nread, (long)count, (long)value);
                value = 0;
                ++count;
                if (count >= pakt->mTableSize) {
                    break;
                }
            }
        }
    }
    
    spDebug(80, "spReadCafPacketTableChunk", "done: nread = %ld / %ld\n",
            (long)nread, (long)pakt->header.mChunkSize);
    
    return nread;
}

spBool spFreeCafPacketTableChunk(void *chunk)
{
    spCafPacketTableChunk *pakt = (spCafPacketTableChunk *)chunk;

    if (pakt->mBuffer != NULL) {
        xfree(pakt->mBuffer);
        pakt->mBuffer = NULL;
    }
    pakt->mTable = NULL;
    
    return SP_TRUE;
}

static spChunkSize spWriteCafPacketTableContent(spCafPacketTableChunk *pakt, int swap, FILE *fp)
{
    unsigned char c;
    spChunkSize nwrite = 0;
    
    spDebug(80, "spWriteCafPacketTableContent", "mTableSize = %ld\n", (long)pakt->mTableSize);
    
    if (pakt->mTableSize > 0) {
        int shift;
        spLong64 k;
        spLong64 mask;
        unsigned char contflag;
        spLong64 value;

        for (k = 0; k < pakt->mTableSize; k++) {
            spDebug(100, "spWriteCafPacketTableContent", "mTable[%ld] = %ld\n", (long)k, (long)pakt->mTable[k]);
            
            /*mask = 0x7f80000000000000;*/
            mask = 0x7f80L;
            mask <<= 48;
            contflag = 0;
            for (shift = 56; shift >= 0; shift -= 7) {
                value = (mask & pakt->mTable[k]) >> shift;
                if (value) {
                    contflag = 0x80;
                }
                spDebug(100, "spWriteCafPacketTableContent", "shift = %d, mask = %lx, value = %lx, contflag = %x\n",
                        shift, (long)mask, (long)value, contflag);
                
                if (contflag || shift == 0) {
                    c = (unsigned char)(0x7f & value);
                    if (shift > 0) {
                        c |= contflag;
                    }
                    spDebug(100, "spWriteCafPacketTableContent", "write %x\n", c);
                    
                    if (fp != NULL) {
                        if (fwrite(&c, 1, 1, fp) != 1) {
                            spDebug(10, "spWriteCafPacketTableContent", "Can't write packet table in 'pakt' chunk.\n");
                            return 0;
                        }
                    }
                    nwrite += 1;
                }
                
                mask >>= 7;
            }
        }
    }

    spDebug(80, "spWriteCafPacketTableContent", "done: nwrite = %ld / %ld\n",
            (long)nwrite, (long)pakt->header.mChunkSize);
    
    return nwrite;
}

spChunkSize spWriteCafPacketTableChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spChunkSize nwrite;
    spCafPacketTableChunk *pakt = (spCafPacketTableChunk *)chunk;

    if (pakt == NULL) return 0;

    spDebug(80, "spWriteCafPacketTableChunk", "in: offset = %ld\n", (long)pakt->offset);

    spDebug(80, "spWriteCafPacketTableChunk", "mNumberPackets = %ld\n", (long)pakt->mNumberPackets);
    if (fwritelong64(&pakt->mNumberPackets, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafPacketTableChunk", "Can't write mNumberPackets.\n");
        return 0;
    }
    nwrite = 8;
    
    spDebug(80, "spWriteCafPacketTableChunk", "mNumberValidFrames = %ld\n", (long)pakt->mNumberValidFrames);
    if (fwritelong64(&pakt->mNumberValidFrames, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafPacketTableChunk", "Can't write mNumberValidFrames.\n");
        return 0;
    }
    nwrite += 8;
    
    spDebug(80, "spWriteCafPacketTableChunk", "mPrimingFrames = %ld\n", pakt->mPrimingFrames);
    if (fwritelong32((long *)&pakt->mPrimingFrames, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafPacketTableChunk", "Can't write mPrimingFrames.\n");
        return 0;
    }
    nwrite += 4;

    spDebug(80, "spWriteCafPacketTableChunk", "mRemainderFrames = %ld\n", pakt->mRemainderFrames);
    if (fwritelong32((long *)&pakt->mRemainderFrames, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafPacketTableChunk", "Can't write mRemainderFrames.\n");
        return 0;
    }
    nwrite += 4;

    nwrite += spWriteCafPacketTableContent(pakt, swap, fp);

    spDebug(80, "spWriteCafPacketTableChunk", "done: nwrite = %ld / %ld\n",
            (long)nwrite, (long)pakt->header.mChunkSize);
    
    return nwrite;
}

spChunkSize spCopyCafPacketTableChunk(spChunk *dest_chunk, spChunk *src_chunk, long depth)
{
    spChunkSize ncopy;
    spCafPacketTableChunk *dest_pakt = (spCafPacketTableChunk *)dest_chunk;
    spCafPacketTableChunk *src_pakt = (spCafPacketTableChunk *)src_chunk;

    if (dest_pakt == NULL || src_pakt == NULL) return 0;

    spDebug(80, "spCopyCafPacketTableChunk", "in\n");

    dest_pakt->mNumberPackets = src_pakt->mNumberPackets;
    ncopy = sizeof(dest_pakt->mNumberPackets);

    dest_pakt->mNumberValidFrames = src_pakt->mNumberValidFrames;
    ncopy += sizeof(dest_pakt->mNumberValidFrames);

    dest_pakt->mPrimingFrames = src_pakt->mPrimingFrames;
    ncopy += sizeof(dest_pakt->mPrimingFrames);

    dest_pakt->mRemainderFrames = src_pakt->mRemainderFrames;
    ncopy += sizeof(dest_pakt->mRemainderFrames);

    ncopy += sizeof(dest_pakt->mBufferSize) + sizeof(dest_pakt->mBuffer) + sizeof(dest_pakt->mTableSize) + sizeof(dest_pakt->mTable);
    
    dest_pakt->mBufferSize = 0;
    dest_pakt->mBuffer = NULL;
    dest_pakt->mTableSize = 0;
    dest_pakt->mTable = NULL;

    if (dest_pakt->mNumberPackets > 0) {
        dest_pakt->mTableSize = src_pakt->mTableSize;
        dest_pakt->mBufferSize = src_pakt->mBufferSize;
        
        dest_pakt->mBuffer = xalloc((long)dest_pakt->mBufferSize, char);
        memcpy(dest_pakt->mBuffer, src_pakt->mBuffer, (long)dest_pakt->mBufferSize);
        dest_pakt->mTable = (spLong64 *)dest_pakt->mBuffer;
        
        ncopy += dest_pakt->mBufferSize;
    }
    
    return ncopy;
}

spChunkSize spUpdateCafPacketTableChunkSize(void *chunk)
{
    spCafPacketTableChunk *pakt = (spCafPacketTableChunk *)chunk;

    return MAX(24 + spWriteCafPacketTableContent(pakt, SP_AIFF_NEED_SWAP, NULL), pakt->header.mChunkSize);
}

spChunkSize spReadCafAudioDataChunk(spChunk *parent, long depth, void *chunk, int swap, FILE *fp)
{
    spLong64 size;
    spCafAudioDataChunk *data = (spCafAudioDataChunk *)chunk;

    spDebug(80, "spReadCafAudioDataChunk", "in: mChunkSize = %ld\n", (long)data->header.mChunkSize);

    if (freadlong32((long *)&data->mEditCount, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafAudioDescChunk", "Can't read mEditCount in 'data' chunk.\n");
        return 0;
    }
    spDebug(80, "spReadCafAudioDataChunk", "mEditCount = %ld\n", data->mEditCount);
    
    data->mOffset = spTellFile(fp);

    if (data->header.mChunkSize <= 0) {
        if (spSeekFile(fp, 0, SEEK_END) != 0) {
            return 0;
        }
        data->header.mChunkSize = spTellFile(fp) - data->mOffset + 4;
    } else {
        size = data->header.mChunkSize - 4;

        if (size > 0) {
            if (spSeekFile(fp, size, SEEK_CUR) != 0) {
                return 0;
            }
        }
    }
    
    return data->header.mChunkSize;
}

spChunkSize spWriteCafAudioDataChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spCafAudioDataChunk *data = (spCafAudioDataChunk *)chunk;

    spDebug(80, "spWriteCafAudioDataChunk", "in: mChunkSize = %ld\n", (long)data->header.mChunkSize);

    spDebug(80, "spWriteCafAudioDataChunk", "mEditCount = %ld\n", data->mEditCount);
    if (fwritelong32((long *)&data->mEditCount, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafAudioDescChunk", "Can't write mEditCount in 'data' chunk.\n");
        return 0;
    }
    
    data->mOffset = spTellFile(fp);
    spDebug(80, "spWriteCafAudioDataChunk", "done: mOffset = %ld\n", (long)data->mOffset);
    
    return 4;
}

spChunkSize spReadCafInformationChunk(spChunk *parent, long depth, void *chunk, int swap, FILE *fp)
{
    long k;
    long pos;
    long nread;
    spChunkSize size;
    char *data;
    spCafInformationChunk *info = (spCafInformationChunk *)chunk;

    spDebug(80, "spReadCafInformationChunk", "in\n");

    info->mInformations = NULL;

    if (freadlong32((long *)&info->mNumEntries, 1, swap, fp) != 1) {
        spDebug(10, "spReadCafInformationChunk", "Can't read mNumEntries in 'info' chunk.\n");
        return 0;
    }
    
    size = info->header.mChunkSize - 4;
    data = xalloc(size, char);
    spDebug(80, "spReadCafInformationChunk", "mNumEntries = %ld, size = %ld\n", info->mNumEntries, (long)size);

    if ((nread = (long)fread(data, 1, (long)size, fp)) <= 0) {
        spDebug(10, "spReadCafInformationChunk", "Can't read data in 'info' chunk.\n");
    } else {
        info->mInformations = xalloc(info->mNumEntries, spCafInformation);
    
        pos = 0;
        for (k = 0; k < info->mNumEntries; k++) {
            info->mInformations[k].mKey = strclone(&data[pos]);
            pos += (long)strlen(info->mInformations[k].mKey) + 1;
            info->mInformations[k].mValue = strclone(&data[pos]);
            pos += (long)strlen(info->mInformations[k].mValue) + 1;
        }
        spDebug(80, "spReadCafInformationChunk", "pos = %ld, nread = %ld\n", pos, nread);
    }

    xfree(data);
    
    spDebug(80, "spReadCafInformationChunk", "done: nread = %ld\n", nread);
    
    return info->header.mChunkSize;
}

spBool spFreeCafInformationChunk(void *chunk)
{
    spCafInformationChunk *info = (spCafInformationChunk *)chunk;

    if (info->mInformations != NULL) {
        long k;

        for (k = 0; k < info->mNumEntries; k++) {
            if (info->mInformations[k].mKey != NULL) xfree(info->mInformations[k].mKey);
            if (info->mInformations[k].mValue != NULL) xfree(info->mInformations[k].mValue);
        }
        
        xfree(info->mInformations);
        info->mInformations = NULL;
    }
    
    return SP_TRUE;
}
    
spChunkSize spWriteCafInformationChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long k;
    long len;
    spChunkSize nwrite;
    spCafInformationChunk *info = (spCafInformationChunk *)chunk;

    spDebug(80, "spWriteCafInformationChunk", "in\n");

    spDebug(80, "spWriteCafInformationChunk", "mNumEntries = %ld\n", info->mNumEntries);
    if (fwritelong32((long *)&info->mNumEntries, 1, swap, fp) != 1) {
        spDebug(10, "spWriteCafInformationChunk", "Can't write mNumEntries in 'info' chunk.\n");
        return 0;
    }
    nwrite = 4;
    
    for (k = 0; k < info->mNumEntries; k++) {
        len = (long)strlen(info->mInformations[k].mKey) + 1;
        if ((long)fwrite(info->mInformations[k].mKey, 1, len, fp) != len) {
            return 0;
        }
        nwrite += len;
        
        len = (long)strlen(info->mInformations[k].mValue) + 1;
        if ((long)fwrite(info->mInformations[k].mValue, 1, len, fp) != len) {
            return 0;
        }
        nwrite += len;
    }
    
    spDebug(80, "spWriteCafInformationChunk", "done: nwrite = %ld / %ld\n",
            (long)nwrite, (long)info->header.mChunkSize);
    
    return nwrite;
}

spChunkSize spCopyCafInformationChunk(spChunk *dest_chunk, spChunk *src_chunk, long depth)
{
    long k;
    long ncopy;
    spCafInformationChunk *dest_info = (spCafInformationChunk *)dest_chunk;
    spCafInformationChunk *src_info = (spCafInformationChunk *)src_chunk;

    spDebug(80, "spCopyCafInformationChunk", "in\n");

    dest_info->mNumEntries = src_info->mNumEntries;
    ncopy = sizeof(dest_info->mNumEntries);
    
    ncopy += sizeof(dest_info->mInformations);
    
    if (src_info->mInformations == NULL) {
        dest_info->mInformations = NULL;
    } else {
        dest_info->mInformations = xalloc(dest_info->mNumEntries, spCafInformation);
    
        for (k = 0; k < dest_info->mNumEntries; k++) {
            dest_info->mInformations[k].mKey = strclone(src_info->mInformations[k].mKey);
            ncopy += (long)strlen(src_info->mInformations[k].mKey) + 1;
            dest_info->mInformations[k].mValue = strclone(src_info->mInformations[k].mValue);
            ncopy += (long)strlen(src_info->mInformations[k].mValue) + 1;
        }
    }

    spDebug(80, "spCopyCafInformationChunk", "done: ncopy = %ld\n", ncopy);
    
    return ncopy;
}

const char *spGetCafInformationValue(spCafInformationChunk *info, const char *key, long *index)
{
    long k;
    
    if (info == NULL || strnone(key) || info->mInformations == NULL) return NULL;
    
    for (k = 0; k < info->mNumEntries; k++) {
        if (streq(info->mInformations[k].mKey, key)) {
            if (index != NULL) *index = k;
            return info->mInformations[k].mValue;
        }
    }

    return NULL;
}

long spGetCafInformationValueInLocaleCode(spCafInformationChunk *info, const char *key, unsigned char *str_in_locale, long size)
{
    long index;
    const unsigned char *utf8str;
    
    if ((utf8str = (const unsigned char *)spGetCafInformationValue(info, key, &index)) != NULL) {
        spStrCopy((char *)str_in_locale, size, (const char *)utf8str);
        spConvertKanjiToLocaleCode(str_in_locale, size, SP_KANJI_CODE_UTF8);
        return index;
    }

    return -1;
}

spChunkSize spWriteCafChildChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp);
spChunkSize spCopyCafChildChunk(spChunk *dest_parent, spChunk *src_parent, long depth);

static spChunkInfoTable sp_caf_chunk_info_list[] = {
    {"caff", "", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK, 0, sizeof(spCafHeader), 0, NULL, spWriteCafChildChunk, NULL, NULL, SP_FALSE, NULL, spCopyCafChildChunk},
    {"desc", "caff", SP_TRUE, 0, spOffsetOf(spCafHeader, desc), sizeof(spCafAudioDescChunk), 32, spReadCafAudioDescChunk, spWriteCafAudioDescChunk, NULL, NULL, SP_TRUE, NULL, NULL},
    {"pakt", "caff", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK, spOffsetOf(spCafHeader, pakt), sizeof(spCafPacketTableChunk), 0, spReadCafPacketTableChunk, spWriteCafPacketTableChunk, NULL, spFreeCafPacketTableChunk, SP_TRUE, spUpdateCafPacketTableChunkSize, spCopyCafPacketTableChunk},
    {"kuki", "caff", SP_FALSE, 0, spOffsetOf(spCafHeader, kuki), sizeof(spCafGeneralDataChunk), 0, spReadCafGeneralDataChunk, spWriteCafGeneralDataChunk, NULL, spFreeCafGeneralDataChunk, SP_TRUE, NULL, spCopyCafGeneralDataChunk},
    {"info", "caff", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK, 0, sizeof(spCafInformationChunk), 0, spReadCafInformationChunk, spWriteCafInformationChunk, NULL, spFreeCafInformationChunk, SP_TRUE, NULL, spCopyCafInformationChunk},
    {"data", "caff", SP_TRUE, SP_CHUNK_INFO_OPTION_PAUSE_AFTER_WRITE_MASK | SP_CHUNK_INFO_OPTION_REWRITE_MASK, spOffsetOf(spCafHeader, data), sizeof(spCafAudioDataChunk), 0, spReadCafAudioDataChunk, spWriteCafAudioDataChunk, NULL, NULL, SP_TRUE, NULL, NULL},
    {"free", "", SP_FALSE, SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
};
static int sp_num_caf_chunk_info_list = 0;

spChunkInfoTable *spFindCafChunkInfoTable(spCafChunkHeader *parent_header, spCafChunkHeader *header)
{
    if (sp_num_caf_chunk_info_list == 0) {
        sp_num_caf_chunk_info_list = spArraySize(sp_caf_chunk_info_list);
    }

    return spFindChunkInfoTable(sp_caf_chunk_info_list, sp_num_caf_chunk_info_list, (spChunkHeader *)parent_header, (spChunkHeader *)header);
}

spChunkFileSpec sp_caf_file_spec = {
    sizeof(struct _spCafChunkHeader),
    sizeof(struct _spCafChunk),
    sp_caf_chunk_info_list,
    0,

    NULL,
    (spChunkHeaderGetSizeFunc)spGetCafChunkHeaderHeaderSize,
    (spChunkHeaderGetSizeFunc)spGetCafChunkHeaderContentSize,
    (spChunkGetSizeFunc)spGetCafChunkHeaderSize,
    (spChunkGetSizeFunc)spGetCafChunkContentSizeForChunkFileSpec,
    (spChunkSetSizeFunc)spSetCafChunkContentSize,
    (spChunkGetSizeFunc)spGetCafChunkContentMarginSize,
    (spChunkSetSizeFunc)spSetCafChunkContentMarginSize,
    (spChunkReadHeaderFunc)spReadCafChunkHeader,
    NULL,
    (spChunkSkipFunc)spSkipCafChunk,

    (spChunkWriteHeaderFunc)spWriteCafChunkHeader,
    NULL,
    (spChunkWriteContentRestFunc)spWriteCafContentRest,
        
    NULL,
};

static spCafChunk *spFindOrCreateCafChunk(spCafHeader *header, const char *type)
{
    spCafChunk *chunk;
    
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }
    
    if ((chunk = spFindCafChunkFromHeader(header, type)) == NULL) {
        if ((chunk = (spCafChunk *)spCreateChunk(&sp_caf_file_spec, (spChunk *)header, NULL, type,
                                                 SP_TRUE, SP_FALSE, NULL, NULL)) == NULL) {
            return NULL;
        }
    }

    return chunk;
}

spChunkSize spReadCafChildChunk(spChunk *parent, long max_count, spChunkSize remain_size, long depth, int swap, FILE *fp)
{
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }
    return spReadChildChunk(&sp_caf_file_spec, parent, max_count, remain_size, depth, swap, fp);
}

spChunkSize spReadCafHeader(spCafHeader *header, FILE *fp)
{
    spChunkSize nread;
    spChunkSize total_nread;

    memset(header, 0, sizeof(spCafHeader));

    if ((nread = fread(header->header.type, 1, 4, fp)) != 4
        || !strneq(header->header.type, "caff", 4)) {
        spDebug(10, "spReadCafHeader", "Can't read first 'caff': %ld\n", (long)nread);
        if (nread > 0) {
            spSeekFile(fp, -nread, SEEK_CUR);
        }
        return 0;
    }
    total_nread = nread;

    if ((nread = freadshort((short *)&header->mFileVersion, 1, SP_AIFF_NEED_SWAP, fp)) != 1) {
        spDebug(10, "spReadCafHeader", "Can't read mFileVersion: %ld\n", (long)nread);
        return nread;
    }
    total_nread += 2;
    
    if ((nread = freadshort((short *)&header->mFileFlags, 1, SP_AIFF_NEED_SWAP, fp)) != 1) {
        spDebug(10, "spReadCafHeader", "Can't read mFileFlags: %ld\n", (long)nread);
        return nread;
    }
    total_nread += 2;
    
    if ((nread = spReadCafChildChunk((spChunk *)header, 0, 0, 0, SP_AIFF_NEED_SWAP, fp)) <= 0) {
        spDebug(10, "spReadCafHeader", "spReadCafChildChunk failed: %ld\n", (long)nread);
        return nread;
    }
    total_nread += nread;
    
    header->header.mChunkSize = total_nread - 4;

    spDebug(80, "spReadCafHeader", "done: total_nread = %ld\n", (long)total_nread);
    
    return total_nread;
}

spBool spFreeCafHeader(spCafHeader *header)
{
    spDebug(80, "spFreeCafHeader", "in\n");
    
    if (header != NULL) {
        return spFreeChunk(&sp_caf_file_spec, (spChunk *)header);
    } else {
        spDebug(80, "spFreeCafHeader", "header == NULL\n");
        return SP_FALSE;
    }
}

spChunkSize spWriteCafChildChunk(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }
    return spWriteChildChunk(&sp_caf_file_spec, chunk, depth, SP_TRUE, rewrite_flag, swap, NULL, fp);
}

spChunkSize spWriteCafHeader(spCafHeader *header, FILE *fp)
{
    if (header == NULL || fp == NULL) return 0;
    
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }

    return spWriteChunk(&sp_caf_file_spec, (spChunk *)header, 0, SP_TRUE, SP_FALSE, SP_AIFF_NEED_SWAP, NULL, fp);
}

spChunkSize spRewriteCafHeader(spCafHeader *header, long editCount, spLong64 dataSize, FILE *fp)
{
    if (header == NULL || fp == NULL) return 0;
    
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }

    spUpdateCafAudioDataChunk(header, editCount, dataSize);

    if (spSeekFile(fp, 0, SEEK_SET) != 0) {
        return 0;
    }
    
    return spWriteChunk(&sp_caf_file_spec, (spChunk *)header, 0, SP_TRUE, SP_TRUE, SP_AIFF_NEED_SWAP, NULL, fp);
}

spChunkSize spCopyCafChildChunk(spChunk *dest_parent, spChunk *src_parent, long depth)
{
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }
    return spCopyChildChunk(&sp_caf_file_spec, dest_parent, src_parent, depth);
}

spChunkSize spCopyCafHeader(spCafHeader *dest_header, spCafHeader *src_header)
{
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }
    return spCopyChunk(&sp_caf_file_spec, (spChunk *)dest_header, (spChunk *)src_header, 0);
}

spBool spInitCafHeaderWithCB(spCafHeader *header, unsigned short fileVersion, unsigned short fileFlags,
                             spCafChunkUpdateFunc update_func, void *data)
{
    if (header == NULL || fileVersion > 1) return SP_FALSE;
    
    spDebug(80, "spInitCafHeaderWithCB", "fileVersion = %d, fileFlags = %d\n", fileVersion, fileFlags);
    
    if (sp_caf_file_spec.num_list <= 0) {
        sp_caf_file_spec.num_list = spArraySize(sp_caf_chunk_info_list);
    }

    memset(header, 0, sizeof(spCafHeader));

    memcpy(header->header.type, "caff", 4);
    header->header.mChunkSize = 4;
    header->mFileVersion = fileVersion;
    header->mFileFlags = fileFlags;

#if 0
    spCreateChildChunk(&sp_caf_file_spec, (spChunk *)header, SP_TRUE,
                       (spChunkUpdateFunc)update_func, data);
#endif

    spDebug(80, "spInitCafHeaderWithCB", "done\n");
    
    return SP_TRUE;
}

spBool spInitCafHeader(spCafHeader *header, unsigned short fileVersion, unsigned short fileFlags)
{
    return spInitCafHeaderWithCB(header, fileVersion, fileFlags, NULL, NULL);
}

spBool spUpdateCafAudioDescChunk(spCafHeader *header, double sampleRate, const char *formatID,
                                 long formatFlags, long bytesPerPacket, long framesPerPacket,
                                 long channelsPerFrame, long bitsPerChannel)
{
    spCafAudioDescChunk *desc;

    if (header == NULL) return SP_FALSE;

    if ((desc = (spCafAudioDescChunk *)spFindOrCreateCafChunk(header, "desc")) == NULL) {
        return SP_FALSE;
    }

    desc->mSampleRate = sampleRate;
    memcpy(desc->mFormatID, formatID, 4);
    desc->mFormatFlags = formatFlags;
    desc->mBytesPerPacket = bytesPerPacket;
    desc->mFramesPerPacket = framesPerPacket;
    desc->mChannelsPerFrame = channelsPerFrame;
    desc->mBitsPerChannel = bitsPerChannel;

    return SP_TRUE;
}

spBool spUpdateCafGeneralDataChunk(spCafHeader *header, const char *type, char *data, spLong64 size, spBool copy_flag)
{
    spCafGeneralDataChunk *gdata;

    if (header == NULL || header->desc == NULL) return SP_FALSE;
    
    spDebug(80, "spUpdateCafGeneralDataChunk", "type = %c%c%c%c, size = %ld\n",
            type[0], type[1], type[2], type[3], (long)size);
    
    if ((gdata = (spCafGeneralDataChunk *)spFindOrCreateCafChunk(header, type)) == NULL) {
        return SP_FALSE;
    }

    if (copy_flag) {
        gdata->allocFlag = SP_TRUE;
        gdata->mData = xalloc((long)size, char);
        memcpy(gdata->mData, data, (long)size);
    } else {
        gdata->allocFlag = SP_FALSE;
        gdata->mData = data;
    }

    /*gdata->header.mChunkSize = size;*/
    spSetChunkContentSize(&sp_caf_file_spec, (spChunk *)gdata, size, SP_TRUE);

    spDebug(80, "spUpdateCafGeneralDataChunk", "done (type = %c%c%c%c)\n", type[0], type[1], type[2], type[3]);
    
    return SP_TRUE;
}

spBool spUpdateCafPacketTableChunkForFixedFrame(spCafHeader *header, long packetMarginSize, spLong64 totalFrames, long *maxPacketSize)
{
    spCafPacketTableChunk *pakt;
    long bytesPerFrame;
    long byteSizeTableEntry = 0;
    long maxPacketSizeLocal = 0;
    spChunkSize maxPacketTableSize = 0;

    if (header == NULL || header->desc == NULL) return SP_FALSE;

    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame", "in: packetMarginSize = %ld, totalFrames = %ld\n",
            packetMarginSize, (long)totalFrames);

    if (header->desc->mFramesPerPacket <= 0) {
        spDebug(10, "spUpdateCafPacketTableChunkForFixedFrame", "desc->mFramesPerPacket must be more than 0\n");
        return SP_FALSE;
    }
    
    if ((pakt = (spCafPacketTableChunk *)spFindOrCreateCafChunk(header, "pakt")) == NULL) {
        return SP_FALSE;
    }

    bytesPerFrame = (header->desc->mBitsPerChannel * header->desc->mChannelsPerFrame) / 8;
    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame",
            "mBitsPerChannel = %ld, mChannelsPerFrame = %ld, bytesPerFrame = %ld\n",
            header->desc->mBitsPerChannel, header->desc->mChannelsPerFrame, bytesPerFrame);

    pakt->mNumberValidFrames = totalFrames;
    pakt->mNumberPackets = pakt->mNumberValidFrames / header->desc->mFramesPerPacket;
    pakt->mPrimingFrames = 0;
    pakt->mRemainderFrames = (long)(pakt->mNumberValidFrames - pakt->mNumberPackets * header->desc->mFramesPerPacket);
    pakt->mRemainderFrames = header->desc->mFramesPerPacket - pakt->mRemainderFrames;
    if (pakt->mRemainderFrames) pakt->mNumberPackets += 1;
    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame",
            "mNumberValidFrames = %ld, mNumberPackets = %ld, mRemainderFrames = %ld\n",
            (long)pakt->mNumberValidFrames, (long)pakt->mNumberPackets, pakt->mRemainderFrames);

    maxPacketSizeLocal = bytesPerFrame * header->desc->mFramesPerPacket + packetMarginSize/*kALACMaxEscapeHeaderBytes*/;
    if (maxPacketSize != NULL) *maxPacketSize = maxPacketSizeLocal;
    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame", "maxPacketSizeLocal = %ld\n", maxPacketSizeLocal);
    
    spUpdateCafPacketTableChunkBuffer(header->desc, pakt, SP_FALSE);

    /*spUpdateChunkContentSize(&sp_caf_file_spec, (spChunk *)pakt, SP_TRUE);*/
    
    if (maxPacketSizeLocal < 16384) {
        byteSizeTableEntry = 2;
    } else {
        byteSizeTableEntry = 3;
    }
    maxPacketTableSize =  byteSizeTableEntry * (long)pakt->mNumberPackets;
    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame", "maxPacketTableSize = %ld\n", maxPacketTableSize);
    
    spSetChunkContentSize(&sp_caf_file_spec, (spChunk *)pakt, maxPacketTableSize + 24, SP_TRUE);

    spDebug(80, "spUpdateCafPacketTableChunkForFixedFrame", "done\n");
    
    return SP_TRUE;
}

spBool spUpdateCafPacketTableChunkNumberPackets(spCafHeader *header, spLong64 numberPackets, spLong64 totalFrames, spBool *insufficientBuffer)
{
    spCafPacketTableChunk *pakt;
    
    if (header == NULL || header->desc == NULL) return SP_FALSE;
    
    if ((pakt = (spCafPacketTableChunk *)spFindCafChunkFromHeader(header, "pakt")) == NULL) {
        return SP_FALSE;
    }

    if (totalFrames > 0) {
        pakt->mNumberValidFrames = totalFrames;
        
        if (header->desc->mFramesPerPacket > 0) {
            pakt->mRemainderFrames = (long)(numberPackets * header->desc->mFramesPerPacket - totalFrames);
        }
    }
    
    pakt->mNumberPackets = numberPackets;
    
    if (header->desc->mBytesPerPacket == 0 && header->desc->mFramesPerPacket == 0) {
        pakt->mTableSize = pakt->mNumberPackets * 2;
    } else {
        pakt->mTableSize = pakt->mNumberPackets;
    }

    if (insufficientBuffer != NULL) {
        if ((pakt->mBufferSize / (spLong64)sizeof(spLong64)) < pakt->mTableSize) {
            *insufficientBuffer = SP_TRUE;
        } else {
            *insufficientBuffer = SP_FALSE;
        }
    }

    spUpdateChunkContentSize(&sp_caf_file_spec, (spChunk *)pakt, SP_TRUE);
    
    return SP_TRUE;
}

spBool spUpdateCafAudioDataChunk(spCafHeader *header, long editCount, spLong64 dataSize)
{
    spCafAudioDataChunk *data;

    if (header == NULL || header->desc == NULL) return SP_FALSE;
    
    spDebug(80, "spUpdateCafAudioDataChunk", "in: dataSize = %ld\n", (long)dataSize);
    
    if ((data = (spCafAudioDataChunk *)spFindOrCreateCafChunk(header, "data")) == NULL) {
        return SP_FALSE;
    }

    data->mEditCount = editCount;
    spSetChunkContentSize(&sp_caf_file_spec, (spChunk *)data, dataSize + 4, SP_TRUE);

    spDebug(80, "spUpdateCafAudioDataChunk", "done\n");
    
    return SP_TRUE;
}

spCafInformationChunk *spCreateCafInformationChunk(spCafHeader *header, spLong64 sizeWithFree)
{
    spCafInformationChunk *info;

    if (header == NULL || header->desc == NULL) return NULL;
    
    if ((info = (spCafInformationChunk *)spFindOrCreateCafChunk(header, "info")) == NULL) {
        return NULL;
    }
    
    if (sizeWithFree > 0) {
        spLong64 marginSize;
        
        marginSize = sizeWithFree - spGetCafChunkHeaderContentSize(&info->header) + info->header.mMarginSize;
        spDebug(80, "spCreateCafInformationChunk", "sizeWithFree = %ld, marginSize = %ld\n", (long)sizeWithFree, (long)marginSize);
        spSetCafChunkContentSize((spCafChunk *)info, sizeWithFree);
        spSetCafChunkContentMarginSize((spCafChunk *)info, marginSize);
    }

    spDebug(80, "spCreateCafInformationChunk", "done\n");
    
    return info;
}

spBool spUpdateCafInformationChunk(spCafInformationChunk *info, const char *key, const char *value)
{
    long k;
    spChunkSize size;
    spBool found = SP_FALSE;

    if (info == NULL || strnone(key)) return SP_FALSE;
    
    size = 4;

    if (info->mInformations != NULL) {
        for (k = 0; k < info->mNumEntries; k++) {
            if (streq(info->mInformations[k].mKey, key)) {
                if (info->mInformations[k].mValue != NULL) xfree(info->mInformations[k].mValue);
                info->mInformations[k].mValue = strclone(value);
                found = SP_TRUE;
            }
            size += strlen(info->mInformations[k].mKey) + 1;
            size += strlen(info->mInformations[k].mValue) + 1;
        }
    } else {
        info->mNumEntries = 0;
    }

    if (found == SP_FALSE) {
        info->mNumEntries++;
        if (info->mNumEntries > info->mAllocCount) {
            info->mAllocCount = (1 + info->mNumEntries / SP_CAF_MIN_ENTRY_ALLOC_COUNT) * SP_CAF_MIN_ENTRY_ALLOC_COUNT;
            info->mInformations = xrealloc(info->mInformations, info->mAllocCount, spCafInformation);
        }

        info->mInformations[info->mNumEntries - 1].mKey = strclone(key);
        info->mInformations[info->mNumEntries - 1].mValue = strclone(value);
        
        size += strlen(key) + 1;
        size += strlen(value) + 1;
    }

    /*info->header.mChunkSize = size;*/
    spSetChunkContentSize(&sp_caf_file_spec, (spChunk *)info, size, SP_TRUE);

    return SP_TRUE;
}

spBool spUpdateCafInformationChunkFromLocaleCode(spCafInformationChunk *info, const char *key, const unsigned char *str_in_locale)
{
    int size;
    unsigned char *utf8str;
    spBool flag;
    
    if (info == NULL || strnone(key)) return SP_FALSE;
    
    spDebug(50, "spUpdateCafInformationChunkFromLocaleCode", "str_in_locale = %s\n", str_in_locale);

    size = (int)(strlen((char *)str_in_locale) + 1) * 2;
    utf8str = xalloc(size, unsigned char);
    spStrCopy((char *)utf8str, size, (const char *)str_in_locale);
    
    spConvertKanjiFromLocaleCode(utf8str, size, SP_KANJI_CODE_UTF8);
    spDebug(50, "spUpdateCafInformationChunkFromLocaleCode", "utf8str = %s\n", utf8str);

    flag = spUpdateCafInformationChunk(info, key, (const char *)utf8str);

    xfree(utf8str);

    return flag;
}

spBool spAppendCafInformationChunk(spCafHeader *header, const char *key, const char *value)
{
    spCafInformationChunk *info;

    if (header == NULL || header->desc == NULL) return SP_FALSE;
    
    if ((info = (spCafInformationChunk *)spFindOrCreateCafChunk(header, "info")) == NULL) {
        return SP_FALSE;
    }

    return spUpdateCafInformationChunk(info, key, value);
}

spBool spAppendCafInformationChunkFromLocaleCode(spCafHeader *header, const char *key, const unsigned char *str_in_locale)
{
    spCafInformationChunk *info;

    if (header == NULL || header->desc == NULL) return SP_FALSE;
    
    if ((info = (spCafInformationChunk *)spFindOrCreateCafChunk(header, "info")) == NULL) {
        return SP_FALSE;
    }

    return spUpdateCafInformationChunkFromLocaleCode(info, key, str_in_locale);
}

static int parseSongInfoNumber(const char *str, spLong32 *number, spLong32 *total)
{
    int i, j;
    int count;
    char buf[SP_MAX_LINE];
    
    *number = 0;
    *total = 0;

    if (str[0] == NUL) return 0;
    
    for (i = 0, j = 0, count = 0;; i++) {
        buf[j] = str[i];
        if (str[i] == NUL || !isdigit(str[i])) {
            buf[j] = NUL;

            if (j > 0 && count == 0) {
                *number = atol(buf);
                ++count;
            } else if (j > 0 && count == 1) {
                *total = atol(buf);
                ++count;
                break;
            }

            if (str[i] == NUL) {
                break;
            }
            
            j = 0;
        } else {
            ++j;
        }
    }

    return count;
}

spBool spParseCafSongInfo(spCafHeader *header, spSongInfoV2 *song_info)
{
    spCafInformationChunk *info;
    const char *str;
    int count;
    
    if (header == NULL || header->desc == NULL || song_info == NULL) return SP_FALSE;

    if ((info = (spCafInformationChunk *)spFindCafChunkFromHeader(header, "info")) == NULL) {
        return SP_FALSE;
    }

    spInitSongInfoV2(song_info);

    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_TITLE, (unsigned char *)song_info->title, sizeof(song_info->title)) >= 0) {
        song_info->info_mask |= SP_SONG_TITLE_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_ARTIST, (unsigned char *)song_info->artist, sizeof(song_info->artist)) >= 0) {
        song_info->info_mask |= SP_SONG_ARTIST_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_ALBUM, (unsigned char *)song_info->album, sizeof(song_info->album)) >= 0) {
        song_info->info_mask |= SP_SONG_ALBUM_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_ALBUM_ARTIST, (unsigned char *)song_info->album_artist, sizeof(song_info->album_artist)) >= 0) { /* unofficial */
        song_info->info_mask |= SP_SONG_ALBUM_ARTIST_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_RELEASE, (unsigned char *)song_info->release, sizeof(song_info->release)) >= 0) {
        song_info->info_mask |= SP_SONG_RELEASE_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_COMPOSER, (unsigned char *)song_info->composer, sizeof(song_info->composer)) >= 0) {
        song_info->info_mask |= SP_SONG_COMPOSER_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_GENRE, (unsigned char *)song_info->genre, sizeof(song_info->genre)) >= 0) {
        song_info->info_mask |= SP_SONG_GENRE_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_COMMENT, (unsigned char *)song_info->comment, sizeof(song_info->comment)) >= 0) {
        song_info->info_mask |= SP_SONG_COMMENT_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_COPYRIGHT, (unsigned char *)song_info->copyright, sizeof(song_info->copyright)) >= 0) {
        song_info->info_mask |= SP_SONG_COPYRIGHT_MASK;
    }
    if (spGetCafInformationValueInLocaleCode(info, SP_CAF_INFORMATION_KEY_SOFTWARE, (unsigned char *)song_info->software, sizeof(song_info->software)) >= 0) {
        song_info->info_mask |= SP_SONG_SOFTWARE_MASK;
    }
    
    if ((str = spGetCafInformationValue(info, SP_CAF_INFORMATION_KEY_TRACK, NULL)) != NULL) {
        count = parseSongInfoNumber(str, &song_info->track, &song_info->track_total);
        if (count >= 1) song_info->info_mask |= SP_SONG_TRACK_MASK;
        if (count >= 2) song_info->info_mask |= SP_SONG_TRACK_TOTAL_MASK;
    }
    if ((str = spGetCafInformationValue(info, SP_CAF_INFORMATION_KEY_DISC, NULL)) != NULL) { /* unofficial */
        count = parseSongInfoNumber(str, &song_info->disc, &song_info->disc_total);
        if (count >= 1) song_info->info_mask |= SP_SONG_DISC_MASK;
        if (count >= 2) song_info->info_mask |= SP_SONG_DISC_TOTAL_MASK;
    }
    if ((str = spGetCafInformationValue(info, SP_CAF_INFORMATION_KEY_TEMPO, NULL)) != NULL) {
        song_info->info_mask |= SP_SONG_TEMPO_MASK;
        song_info->tempo = atol(str);
    }

    if (song_info->info_mask) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spAppendCafSongInfo(spCafHeader *header, spSongInfoV2 *song_info)
{
    spCafInformationChunk *info;
    char buf[SP_MAX_LINE];

    if (header == NULL || header->desc == NULL || song_info == NULL || song_info->info_mask == 0) return SP_FALSE;
    
    if ((info = (spCafInformationChunk *)spFindOrCreateCafChunk(header, "info")) == NULL) {
        return SP_FALSE;
    }

    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_TITLE, (unsigned char *)song_info->title);
    }
    if (song_info->info_mask & SP_SONG_ARTIST_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_ARTIST, (unsigned char *)song_info->artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_ALBUM, (unsigned char *)song_info->album);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_ARTIST_MASK) { /* unofficial */
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_ALBUM_ARTIST, (unsigned char *)song_info->album_artist);
    }
    if (song_info->info_mask & SP_SONG_RELEASE_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_RELEASE, (unsigned char *)song_info->release);
    }
    if (song_info->info_mask & SP_SONG_COMPOSER_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_COMPOSER, (unsigned char *)song_info->composer);
    }
    if (song_info->info_mask & SP_SONG_GENRE_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_GENRE, (unsigned char *)song_info->genre);
    }
    if (song_info->info_mask & SP_SONG_COMMENT_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_COMMENT, (unsigned char *)song_info->comment);
    }
    if (song_info->info_mask & SP_SONG_COPYRIGHT_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_COPYRIGHT, (unsigned char *)song_info->copyright);
    }
    if (song_info->info_mask & SP_SONG_SOFTWARE_MASK) {
        spUpdateCafInformationChunkFromLocaleCode(info, SP_CAF_INFORMATION_KEY_SOFTWARE, (unsigned char *)song_info->software);
    }
    
    if (song_info->info_mask & SP_SONG_TRACK_MASK) {
        if ((song_info->info_mask & SP_SONG_TRACK_TOTAL_MASK) && song_info->track_total >= song_info->track) {
            sprintf(buf, "%ld/%ld", (long)song_info->track, (long)song_info->track_total); /* unofficial */
        } else {
            sprintf(buf, "%ld", (long)song_info->track);
        }
        spUpdateCafInformationChunk(info, SP_CAF_INFORMATION_KEY_TRACK, buf);
    }
    if (song_info->info_mask & SP_SONG_DISC_MASK) { /* unofficial */
        if ((song_info->info_mask & SP_SONG_DISC_TOTAL_MASK) && song_info->disc_total >= song_info->disc) {
            sprintf(buf, "%ld/%ld", (long)song_info->disc, (long)song_info->disc_total); /* unofficial */
        } else {
            sprintf(buf, "%ld", (long)song_info->disc);
        }
        spUpdateCafInformationChunk(info, SP_CAF_INFORMATION_KEY_DISC, buf);
    }
    if (song_info->info_mask & SP_SONG_TEMPO_MASK && song_info->tempo > 0) {
        sprintf(buf, "%ld", (long)song_info->tempo);
        spUpdateCafInformationChunk(info, SP_CAF_INFORMATION_KEY_TEMPO, buf);
    }

    return SP_TRUE;
}

spCafChunk *spFindCafChunkFromHeader(spCafHeader *header, const char *type)
{
    return (spCafChunk *)spFindChildChunk((spChunk *)header, type, NULL);
}

spFileOffset spConvertCafTimeToFileOffset(spCafHeader *header, spLong64 t,
                                          long *packet, long *remainder, long *packet_frames)
{
    long tpacket;
    spLong64 bytes;
    spLong64 frames;
    long k, m;
    spLong64 pos, next_pos;
    spFileOffset offset;
    
    if (header == NULL || header->desc == NULL) return 0;

    spDebug(80, "spConvertCafTimeToFileOffset", "in: t = %ld\n", (long)t);
    
    if (header->pakt != NULL && header->pakt->mPrimingFrames != 0) {
        spDebug(80, "spConvertCafTimeToFileOffset", "header->pakt->mPrimingFrames = %ld\n", (long)header->pakt->mPrimingFrames);
        t += header->pakt->mPrimingFrames;
    }

    spDebug(80, "spConvertCafTimeToFileOffset",
            "header->desc->mFramesPerPacket = %ld, header->desc->mBytesPerPacket = %ld\n",
            (long)header->desc->mFramesPerPacket, (long)header->desc->mBytesPerPacket);
    
    if (header->desc->mFramesPerPacket > 0 && header->desc->mBytesPerPacket > 0) {
        tpacket = (long)(t / header->desc->mFramesPerPacket);
        
        if (packet != NULL) *packet = tpacket;
        if (remainder != NULL) *remainder = (long)(t - tpacket * header->desc->mFramesPerPacket);
        if (packet_frames != NULL) *packet_frames = header->desc->mFramesPerPacket;

        return header->data->mOffset + tpacket * header->desc->mBytesPerPacket;
    }

    if (header->pakt == NULL || header->pakt->mTableSize <= 0) {
        spDebug(80, "spConvertCafTimeToFileOffset", "error: no 'pakt' chunk\n");
        return 0;
    }

    for (k = 0, pos = 0, offset = 0; k < header->pakt->mNumberPackets; k++) {
        if (header->pakt->mTableSize > header->pakt->mNumberPackets) {
            m = k * 2 + 1;
            bytes = header->pakt->mTable[k * 2];
        } else {
            m = k;
            if (header->desc->mBytesPerPacket <= 0) {
                bytes = header->pakt->mTable[m];
            } else {
                bytes = header->desc->mBytesPerPacket;
            }
        }
        spDebug(100, "spConvertCafTimeToFileOffset", "k = %ld, header->pakt->mTable[%ld] = %ld\n",
                k, m, (long)header->pakt->mTable[m]);

        if (header->desc->mFramesPerPacket <= 0) {
            frames = header->pakt->mTable[m];
        } else {
            frames = header->desc->mFramesPerPacket;
        }
        
        next_pos = pos + frames;
        spDebug(100, "spConvertCafTimeToFileOffset", "t = %ld, pos = %ld, next_pos = %ld, offset = %ld ( + %ld)\n",
                (long)t, (long)pos, (long)next_pos, (long)offset, (long)header->data->mOffset);

        if (next_pos > t) {
            if (packet != NULL) *packet = k;
            if (remainder != NULL) *remainder = (long)(t - pos);
            if (packet_frames != NULL) *packet_frames = (long)frames;

            return header->data->mOffset + offset;
        }

        offset += bytes;
        pos = next_pos;
    }

    spDebug(10, "spConvertCafTimeToFileOffset", "error: can't find packet\n");
    
    return 0;
}

spLong64 spGetCafTotalDuration(spCafHeader *header)
{
    if (header == NULL || header->desc == NULL || header->data == NULL) return 0;

    if (header->pakt != NULL) {
        return header->pakt->mNumberValidFrames;
    }

    if (header->data->header.mChunkSize <= 0) {
        return -1;
    }
    
    return (header->data->header.mChunkSize - 4)
        / ((header->desc->mChannelsPerFrame * header->desc->mBitsPerChannel) / 8);
}

spLong64 spGetCafPacketSize(spCafHeader *header, long packet)
{
    if (header == NULL || header->desc == NULL) return 0;

    if (header->desc->mBytesPerPacket > 0) {
        return header->desc->mBytesPerPacket;
    }
    
    if (header->pakt != NULL) {
        if (packet >= header->pakt->mNumberPackets) {
            return 0;
        }
        
        if (header->pakt->mTableSize > header->pakt->mNumberPackets) {
            return header->pakt->mTable[packet * 2];
        } else {
            return header->pakt->mTable[packet];
        }
    } else {
        return (header->desc->mChannelsPerFrame * header->desc->mBitsPerChannel) / 8;
    }
}

spBool spSetCafPacketSize(spCafHeader *header, long packet, spLong64 size)
{
    if (header == NULL || header->desc == NULL
        || header->desc->mBytesPerPacket > 0) return SP_FALSE;

    if (header->pakt != NULL) {
        spDebug(80, "spSetCafPacketSize", "packet = %ld / %ld, size = %ld\n",
                packet, (long)header->pakt->mNumberPackets, (long)size);
        
        if (packet < header->pakt->mNumberPackets) {
            if (header->pakt->mTableSize > header->pakt->mNumberPackets) {
                header->pakt->mTable[packet * 2] = size;
            } else {
                header->pakt->mTable[packet] = size;
            }
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

spBool spSetCafFrameLength(spCafHeader *header, long packet, spLong64 length)
{
    if (header == NULL || header->desc == NULL
        || header->desc->mFramesPerPacket > 0) return SP_FALSE;

    if (header->pakt != NULL) {
        if (packet < header->pakt->mNumberPackets) {
            if (header->pakt->mTableSize > header->pakt->mNumberPackets) {
                header->pakt->mTable[packet * 2 + 1] = length;
            } else {
                header->pakt->mTable[packet] = length;
            }
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}
#endif
