/*
 *        spMp4.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spKanji.h>

#include <sp/spChunkP.h>
#include <sp/spMp4.h>

static long sp_mp4_default_matrix[9] = { 0x00010000, 0, 0, 0, 0x00010000, 0, 0, 0, 0x40000000 };

extern spChunkInfoTable *spFindMp4BoxInfoTable(spMp4BoxHeader *parent_header, spMp4BoxHeader *header);
extern spMp4BoxSize spReadMp4ChildBox(spChunk *parent, long max_count, spMp4BoxSize remain_size, long depth, int swap, FILE *fp);
extern spMp4BoxSize spCopyMp4ChildBox(spChunk *dest_parent, spChunk *src_parent, long depth);

extern spMp4BoxSize spReadMp4ChildOnlyBox(spChunk *parent, long depth, void *box, int swap, FILE *fp);
extern spMp4BoxSize spWriteMp4ChildOnlyBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp);
extern spMp4BoxSize spCopyMp4ChildOnlyBox(spChunk *dest_chunk, spChunk *src_chunk, long depth);

unsigned long spGetMp4Time(time_t t)
{
    t += 2082844800;

    return (unsigned long)t;
}

void spGetMp4TimeC64(time_t t, char tc[8])
{
    t += 2082844800;
#ifdef SP_SUPPORT_64BIT_INT
    spLong64ToC64((spLong64)t, tc, SP_AIFF_NEED_SWAP);
#else
    spULongToC64((unsigned long)t, tc, SP_AIFF_NEED_SWAP);
#endif

    return;
}

time_t spConvMp4Time(unsigned long t)
{
    time_t ot;
    
    ot = (time_t)t;
    ot -= 2082844800;

    return ot;
}

void spConvMp4TimeC64(unsigned long t, char tc[8])
{
    time_t ot;
    
    ot = (time_t)t;
    ot -= 2082844800;
#ifdef SP_SUPPORT_64BIT_INT
    spLong64ToC64((spLong64)ot, tc, SP_AIFF_NEED_SWAP);
#else
    spULongToC64((unsigned long)ot, tc, SP_AIFF_NEED_SWAP);
#endif

    return;
}

unsigned short spGetMp4LangCode(const char *lang)
{
    unsigned short c0, c1, c2;
    unsigned short code;

    c0 = lang[0];
    c0 -= 0x60;
    c0 <<= 10;
    
    c1 = lang[1];
    c1 -= 0x60;
    c1 <<= 5;
    
    c2 = lang[2];
    c2 -= 0x60;

    code = c0 | c1 | c2;
    
    return code;
}

void spGetMp4LangStr(unsigned short code, char *lang, int lang_size)
{
    unsigned short c0, c1, c2;

    c0 = 0x7c00 & code;
    c0 >>= 10;
    c0 += 0x60;
    lang[0] = (char)c0;
    
    c1 = 0x3e0 & code;
    c1 >>= 5;
    c1 += 0x60;
    lang[1] = (char)c1;
    
    c2 = 0x1f & code;
    c2 += 0x60;
    lang[2] = (char)c2;

    if (lang_size >= 4) {
        lang[3] = NUL;
    }
    
    return;
}

static void incrementEntryCount(spMp4Box *box, long size_in_file, char **entries, long size_of_entry,
                                unsigned long *alloc_count, unsigned long *entry_count)
{
    spMp4BoxSize box_size;
    unsigned long orig_alloc_count;
    
    box_size = spGetMp4BoxContentSize(box, NULL);
    
    *entry_count = *entry_count + 1;
    if (*entry_count >= *alloc_count) {
        orig_alloc_count = *alloc_count;
        *alloc_count = (1 + *entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        *entries = xspRemalloc(*entries, size_of_entry * (*alloc_count));
        memset((char *)*entries + size_of_entry * orig_alloc_count, 0,
               size_of_entry * (*alloc_count - orig_alloc_count));
    }

    box_size += size_in_file;
    spSetMp4BoxContentSize(box, box_size, SP_TRUE);
    
    return;
}

static long readMp4String(char *buf, long bufsize, FILE *fp)
{
    int c;
    long k;

    k = 0;
    while ((c = getc(fp)) != EOF) {
        if (k == bufsize - 1) {
            buf[k] = NUL;
        } else if (k < bufsize) {
            buf[k] = c;
        }
        k++;
        if (c == NUL) {
            break;
        }
    }

    return k;
}

static long writeMp4String(char *buf, long bufsize, FILE *fp)
{
    int c;
    long k;

    k = 0;
    while (1) {
        if (k == bufsize - 1) {
            c = NUL;
        } else if (k < bufsize) {
            c = buf[k];
        } else {
            c = NUL;
        }
        if (fputc(c, fp) == EOF) {
            return -1;
        }
        
        k++;
        if (c == NUL) {
            break;
        }
    }

    return k;
}

static void showDataInAscii(char *data, long datasize)
{
    long k;

    for (k = 0; k < datasize; k++) {
        if (data[k] >= 32 && data[k] <= 126) {
            fputc(data[k], stdout);
        } else {
            fputc('.', stdout);
        }
    }
    fputc('\n', stdout);

    return;
}

static void showDataInHex(char *data, long datasize)
{
    long k;

    for (k = 0;;) {
        printf("%04x", data[k++]);
        if (k < datasize) {
            fputc(' ', stdout);
        } else {
            break;
        }
    }
    fputc('\n', stdout);

    return;
}

spBool spInitMp4Box(spChunk *parent, void *chunk, unsigned long options)
{
    spBool fullbox_flag;
    spMp4Box *box = (spMp4Box *)chunk;

    fullbox_flag = (spBool)(0x1 & options);
    if (fullbox_flag) {
        box->header.fullbox_flag = SP_TRUE;
    } else {
        box->header.fullbox_flag = SP_FALSE;
    }

    return SP_TRUE;
}

spBool spEqMp4BoxType(spMp4Box *box, const char *type)
{
    return (spBool)strneq(box->header.type, type, 4);
}

spBool spSetMp4BoxHeaderContentSize(spMp4BoxHeader *box_header, spMp4BoxSize size)
{
    if (box_header->fullbox_flag == SP_TRUE) {
        size += 4;
    }

    if (size >= 2147483647) {
        size += 16;
        box_header->size = 1;
        box_header->largesize = size;
    } else {
        size += 8;
        box_header->size = (unsigned long)size;
        box_header->largesize = 0;
    }

    return SP_TRUE;
}

spBool spSetMp4BoxContentSizeForChunkFileSpec(spMp4Box *box, spMp4BoxSize size)
{
    return spSetMp4BoxHeaderContentSize(&box->header, size);
}

spMp4BoxSize spGetMp4BoxContentMarginSize(spMp4Box *box)
{
    return box->header.marginsize;
}

spBool spSetMp4BoxContentMarginSize(spMp4Box *box, spMp4BoxSize marginsize)
{
    box->header.marginsize = marginsize;
    return SP_TRUE;
}

spMp4BoxSize spGetMp4BoxHeaderContentSize(spMp4BoxHeader *box_header)
{
    spMp4BoxSize size;
    
    if (box_header->size <= 0) {
        return 0;
    } else if (box_header->size == 1) {
        size = box_header->largesize - 16;
    } else {
        size = box_header->size - 8;
    }
    if (box_header->fullbox_flag == SP_TRUE) {
        size -= 4;
    }

    return size;
}

spMp4BoxSize spGetMp4BoxContentSizeForChunkFileSpec(spMp4Box *box)
{
    return spGetMp4BoxHeaderContentSize(&box->header);
}

spMp4BoxSize spGetMp4BoxHeaderBoxSize(spMp4BoxHeader *box_header)
{
    if (box_header->size <= 0) {
        return 0;
    } else if (box_header->size == 1) {
        return box_header->largesize;
    } else {
        return box_header->size;
    }
}

spMp4BoxSize spGetMp4BoxHeaderHeaderSize(spMp4BoxHeader *box_header)
{
    spMp4BoxSize size = 8;
    
    if (box_header->size == 1) {
        size += 8;
    }

    if (box_header->fullbox_flag == SP_TRUE) {
        size += 4;
    }

    return size;
}

spMp4BoxSize spGetMp4BoxHeaderSize(spMp4Box *box)
{
    spMp4BoxSize size;

    size = spGetMp4BoxHeaderHeaderSize(&box->header);
    
    spDebug(100, "spGetMp4BoxHeaderSize", "header.size = %lu, fullbox_flag = %d, size = %lu, type = %c%c%c%c\n",
            box->header.size, box->header.fullbox_flag, (unsigned long)size,
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    return size;
}

spMp4BoxSize spGetMp4BoxContentSize(spMp4Box *box, char *string)
{
    if (box == NULL) return 0;
    
    if (string != NULL) {
        memcpy(string, box->header.type, 4);
        string[4] = NUL;
    }
    
    return spGetMp4BoxHeaderContentSize(&box->header);
}

spMp4BoxSize spGetMp4BoxSize(spMp4Box *box, char *string)
{
    if (box == NULL) return 0;
    
    if (string != NULL) {
        memcpy(string, box->header.type, 4);
        string[4] = NUL;
    }

    return spGetMp4BoxHeaderBoxSize(&box->header);
}

spMp4BoxSize spCopyMp4BoxHeader(spMp4BoxHeader *header_dest, spMp4BoxHeader *header_src)
{
    if (header_dest != header_src && header_src != NULL) {
        spMp4BoxSize ncopy;

        ncopy = sizeof(spMp4BoxHeader);
        memcpy(header_dest, header_src, (long)ncopy);
        
        return ncopy;
    }
    
    return 0;
}

spMp4BoxSize spReadMp4BoxHeader(spMp4Box *parent, spMp4BoxHeader *box_header, int swap, FILE *fp)
{
    long nread;

    if (freadulong32(&box_header->size, 1, swap, fp) != 1) {
        spDebug(10, "spReadMp4BoxHeader", "Can't read box size.\n");
        return 0;
    }
    nread = 4;
    
    if (fread(box_header->type, 1, 4, fp) != 4) {
        spDebug(10, "spReadMp4BoxHeader", "Can't read box type.\n");
        return 0;
    }
    nread += 4;
    spDebug(10, "spReadMp4BoxHeader", "size = %lu, type = %c%c%c%c\n",
            box_header->size, box_header->type[0], box_header->type[1], box_header->type[2], box_header->type[3]);
    
    if (box_header->size == 1) {
        char largesize[8];
        
        /* large size */
        if (fread(largesize, 1, 8, fp) != 8) {
            spDebug(10, "spReadMp4BoxHeader", "Can't read largesize info.\n");
            return 0;
        }

#ifdef SP_SUPPORT_64BIT_INT
        box_header->largesize = spC64ToLong64(largesize, SP_AIFF_NEED_SWAP);
#else
        box_header->largesize = spC64ToULong(largesize, SP_AIFF_NEED_SWAP);
#endif
        nread += 8;
    }
    
    spDebug(80, "spReadMp4BoxHeader", "done: nread = %ld\n", nread);
    
    return nread;
}

spMp4BoxSize spReadMp4FullBoxHeaderRest(spMp4Box *parent, spMp4BoxHeader *fullbox_header, spMp4BoxHeader *box_header,
                                       unsigned long options, int swap, FILE *fp)
{
    spBool fullbox_flag;
    
    fullbox_flag = (spBool)(0x1 & options);

    spCopyMp4BoxHeader(fullbox_header, box_header);
    
    if (fullbox_flag == SP_FALSE) {
        return 0;
    }
    
    if (fread(&fullbox_header->version, 1, 1, fp) != 1) {
        spDebug(10, "spReadMp4FullBoxHeaderRest", "Can't read box version.\n");
        return 0;
    }
    if (fread(&fullbox_header->flags, 1, 3, fp) != 3) {
        spDebug(10, "spReadMp4FullBoxHeaderRest", "Can't read box flags.\n");
        return 0;
    }
    fullbox_header->fullbox_flag = SP_TRUE;

    return 4;
}

spMp4BoxSize spWriteMp4BoxHeader(spMp4Box *box, spMp4BoxHeader *box_header, int swap, FILE *fp)
{
    long nwrite;
    unsigned long size;

    size = box_header->size;
    if (size != 1 && box_header->marginsize > 0) {
        size -= (unsigned long)box_header->marginsize;
    }
    
    if (fwriteulong32(&size, 1, swap, fp) != 1) {
        spDebug(10, "spWriteMp4BoxHeader", "Can't write box size.\n");
        return 0;
    }
    nwrite = 4;
    
    spDebug(10, "spWriteMp4BoxHeader", "size = %ld, type = %c%c%c%c\n",
            box_header->size, box_header->type[0], box_header->type[1], box_header->type[2], box_header->type[3]);
    if (fwrite(box_header->type, 1, 4, fp) != 4) {
        spDebug(10, "spWriteMp4BoxHeader", "Can't write box type.\n");
        return 0;
    }
    nwrite += 4;
    
    if (box_header->size == 1) {
        char largesize[8];

#ifdef SP_SUPPORT_64BIT_INT
        spLong64ToC64(box_header->largesize, largesize, SP_AIFF_NEED_SWAP);
#else
        spULongToC64(box_header->largesize, largesize, SP_AIFF_NEED_SWAP);
#endif
        
        if (box_header->marginsize > 0) {
            spAddC64(largesize, (long)-box_header->marginsize, SP_AIFF_NEED_SWAP);
        }
        
        /* large size */
        if (fwrite(largesize, 1, 8, fp) != 8) {
            spDebug(10, "spWriteMp4BoxHeader", "Can't write largesize info.\n");
            return 0;
        }
        nwrite += 8;
    }
    
    spDebug(80, "spWriteMp4BoxHeader", "done: nwrite = %ld\n", nwrite);
    
    return nwrite;
}

spMp4BoxSize spWriteMp4FullBoxHeaderRest(spMp4Box *box, spMp4BoxHeader *box_header, unsigned long options, int swap, FILE *fp)
{
    spBool fullbox_flag;
    
    fullbox_flag = (spBool)(0x1 & options);
    spDebug(100, "spWriteMp4FullBoxHeaderRest", "options = %lx, fullbox_flag = %d\n", options, fullbox_flag);

    if (fullbox_flag == SP_FALSE) {
        spDebug(100, "spWriteMp4FullBoxHeaderRest", "not full box, return 0\n");
        return 0;
    }
    
    if (fwrite(&box_header->version, 1, 1, fp) != 1) {
        spDebug(10, "spWriteMp4FullBoxHeaderRest", "Can't write box version.\n");
        return 0;
    }
    if (fwrite(&box_header->flags, 1, 3, fp) != 3) {
        spDebug(10, "spWriteMp4FullBoxHeaderRest", "Can't write box flags.\n");
        return 0;
    }

    spDebug(100, "spWriteMp4FullBoxHeaderRest", "full box, return 4\n");
    
    return 4;
}

spMp4BoxSize spSkipMp4Box(spMp4Box *parent, spMp4BoxHeader *box_header, spMp4BoxSize remain_size, int swap, FILE *fp)
{
    spMp4BoxSize len;
    
    spDebug(10, "spSkipMp4Box", "box_header->size = %lu, box_header->type = %c%c%c%c\n",
            box_header->size, box_header->type[0], box_header->type[1], box_header->type[2], box_header->type[3]);
    
    if (strneq(box_header->type, "wide", 4)) {
        len = 0;                /* content size of 'wide' is always 0 bytes */
    } else if (box_header->size == 0) {
        if (remain_size < 0) {
            if (spSeekFile(fp, 0, SEEK_END) != 0) {
                return 0;
            }
        } else {
            if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
                return 0;
            }
        }
        len = 0;
    } else {
        len = spGetMp4BoxHeaderContentSize(box_header);

        if (len != 0) {
            if (spSeekFile(fp, len, SEEK_CUR) != 0) {
                return 0;
            }
        }
    }

    spDebug(10, "spSkipMp4Box", "done: len = %lu\n", (unsigned long)len);
    
    return len;
}

spMp4BoxSize spWriteMp4ContentRest(spMp4Box *box, long depth, spMp4BoxSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    spMp4BoxSize len;
    spMp4BoxSize nwrite;
    spMp4BoxSize nremain;
    spMp4BoxSize total_nwrite;
    static char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};

    spDebug(80, "spWriteMp4ContentRest", "box->header.size = %lu, box->header.type = %c%c%c%c\n",
            box->header.size, box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    spDebug(80, "spWriteMp4ContentRest", "depth = %ld, remain_size %lu\n", depth, (unsigned long)remain_size);

    total_nwrite = 0;
    nremain = remain_size;
    if (nremain >= 8) {
        spMp4BoxHeader box_header;
        spMp4BoxSize content_size;
        
        if (nremain - 16 >= 2147483647) {
            content_size = nremain - 16;
        } else {
            content_size = nremain - 8;
        }
        spDebug(80, "spWriteMp4ContentRest", "content size of 'free' = %lu\n", (unsigned long)content_size);
        memset(&box_header, 0, sizeof(spMp4BoxHeader));
        memcpy(box_header.type, "free", 4);
        spSetMp4BoxHeaderContentSize(&box_header, content_size);

        if ((nwrite = spWriteMp4BoxHeader(NULL, &box_header, swap, fp)) <= 0) {
            return 0;
        }
        total_nwrite = nwrite;
        nremain -= nwrite;
    }
    
    spDebug(80, "spWriteMp4ContentRest", "nremain = %lu\n", (unsigned long)nremain);
    
    while (nremain > 0) {
        len = MIN(8, nremain);
        spDebug(100, "spWriteMp4ContentRest", "len = %lu, nremain = %lu\n", (unsigned long)len, (unsigned long)nremain);
        if ((nwrite = (long)fwrite(buf, 1, (long)len, fp)) != len) {
            spDebug(10, "spWriteMp4ContentRest", "fwrite zeros failed: nwrite = %lu\n", (unsigned long)nwrite);
            break;
        }
        spDebug(100, "spWriteMp4ContentRest", "fwrite zeros done: nwrite = %lu\n", (unsigned long)nwrite);

        nremain -= nwrite;
        total_nwrite += nwrite;
    }

    spDebug(80, "spWriteMp4ContentRest", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spBool spGetMp4BoxHandleType(spMp4Box *box, char *type)
{
    spMp4Box *parent;
    
    parent = box;
    while (parent != NULL) {
        spDebug(10, "spGetMp4BoxHandleType", "parent->header.type = %c%c%c%c\n",
                parent->header.type[0], parent->header.type[1], parent->header.type[2], parent->header.type[3]);

        if (strneq("trak", parent->header.type, 4)) {
            spMp4TrackBox *trak = (spMp4TrackBox *)parent;
            parent = (spMp4Box *)trak->mdia;
            if (parent == NULL) {
                break;
            }
        }        
        if (strneq("mdia", parent->header.type, 4)) {
            spMp4MediaBox *mdia;
            mdia = (spMp4MediaBox *)parent;
            parent = (spMp4Box *)mdia->hdlr;
            if (parent == NULL) {
                break;
            }
        }
        
        if (strneq("hdlr", parent->header.type, 4)) {
            spMp4HandlerReferenceBox *hdlr = (spMp4HandlerReferenceBox *)parent;
            memcpy(type, hdlr->handler_type, 4);
            return SP_TRUE;
        }
        parent = parent->parent;
    }

    return SP_FALSE;
}

spMp4BoxSize spReadMp4ULongOnlyBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4ULongOnlyBox *ulobox = (spMp4ULongOnlyBox *)box;

    spDebug(50, "spReadMp4ULongOnlyBox", "parent = %c%c%c%c, type = %c%c%c%c\n",
            parent->header.type[0], parent->header.type[1], parent->header.type[2], parent->header.type[3],
            ulobox->header.type[0], ulobox->header.type[1], ulobox->header.type[2], ulobox->header.type[3]);
    
    if ((nread = freadulong32(&ulobox->value, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4ULongOnlyBox", "value = %ld\n", ulobox->value);
    
    spDebug(50, "spReadMp4ULongOnlyBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spMp4BoxSize spUpdateMp4ULongOnlyBox(spMp4ULongOnlyBox *ulobox, unsigned long value)
{
    ulobox->value = value;
    spSetMp4BoxHeaderContentSize(&ulobox->header, 4);

    return 4;
}

spMp4Box *spAppendMp4ULongOnlyBox(spMp4Box *parent, const char *type, unsigned long value)
{
    spMp4ULongOnlyBox *ulobox;

    if ((ulobox = (spMp4ULongOnlyBox *)spFindMp4Box(parent, type, parent->header.type)) == NULL) {
        if ((ulobox = (spMp4ULongOnlyBox *)spCreateMp4Box(parent, type)) == NULL) {
            return NULL;
        }
    }
    spUpdateMp4ULongOnlyBox(ulobox, value);

    return (spMp4Box *)ulobox;
}

spMp4BoxSize spWriteMp4ULongOnlyBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4ULongOnlyBox *ulobox = (spMp4ULongOnlyBox *)box;

    spDebug(50, "spWriteMp4ULongOnlyBox", "type = %c%c%c%c\n",
            ulobox->header.type[0], ulobox->header.type[1], ulobox->header.type[2], ulobox->header.type[3]);
    
    spDebug(50, "spWriteMp4ULongOnlyBox", "value = %ld\n", ulobox->value);
    if ((nwrite = fwriteulong32(&ulobox->value, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    spDebug(50, "spWriteMp4ULongOnlyBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spReadMp4DataOnlyBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4BoxSize content_size;
    spMp4DataOnlyBox *dobox = (spMp4DataOnlyBox *)box;

    total_nread = 0;
    
    spDebug(50, "spReadMp4DataOnlyBox", "parent = %c%c%c%c, type = %c%c%c%c\n",
            parent->header.type[0], parent->header.type[1], parent->header.type[2], parent->header.type[3],
            dobox->header.type[0], dobox->header.type[1], dobox->header.type[2], dobox->header.type[3]);
    
    if ((content_size = spGetMp4BoxHeaderContentSize(&dobox->header)) > 0) {
        dobox->data = xalloc((long)content_size + 1, char);
        if ((nread = (long)fread(dobox->data, 1, (long)content_size, fp)) != content_size) {
            xfree(dobox->data);
            dobox->data = NULL;
            return nread;
        }
        dobox->data[content_size] = NUL;
        spDebug(50, "spReadMp4DataOnlyBox", "data = %s\n", dobox->data);
        total_nread += nread;
    } else {
        dobox->data = NULL;
    }
    
    spDebug(50, "spReadMp4DataOnlyBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spFreeMp4DataOnlyBox(void *box)
{
    spMp4DataOnlyBox *dobox = (spMp4DataOnlyBox *)box;

    spDebug(100, "spFreeMp4DataOnlyBox", "in: type = %c%c%c%c\n",
            dobox->header.type[0], dobox->header.type[1], dobox->header.type[2], dobox->header.type[3]);
    
    if (dobox->data != NULL) {
        xfree(dobox->data);
        dobox->data = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4DataOnlyBox(spMp4DataOnlyBox *dobox, const char *data, spMp4BoxSize size, spBool propagate_size)
{
    if (dobox->data != NULL) xfree(dobox->data);

    dobox->data = xalloc((long)size + 1, char);
    memcpy(dobox->data, data, (long)size);
    dobox->data[size] = NUL;

    spSetMp4BoxContentSize((spMp4Box *)dobox, size, propagate_size);

    return size;
}

spMp4Box *spAppendMp4DataOnlyBox(spMp4Box *parent, const char *type, const char *data, spMp4BoxSize size)
{
    spMp4DataOnlyBox *dobox;

    spDebug(50, "spAppendMp4DataOnlyBox", "type = %c%c%c%c, size = %lu\n",
            type[0], type[1], type[2], type[3], (unsigned long)size);
    
    if ((dobox = (spMp4DataOnlyBox *)spFindMp4Box(parent, type, parent->header.type)) == NULL) {
        if ((dobox = (spMp4DataOnlyBox *)spCreateMp4Box(parent, type)) == NULL) {
            return NULL;
        }
    }
    spUpdateMp4DataOnlyBox(dobox, data, size, SP_TRUE);

    return (spMp4Box *)dobox;
}

spMp4BoxSize spWriteMp4DataOnlyBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4BoxSize content_size;
    spMp4DataOnlyBox *dobox = (spMp4DataOnlyBox *)box;

    total_nwrite = 0;
    
    spDebug(50, "spWriteMp4DataOnlyBox", "type = %c%c%c%c\n",
            dobox->header.type[0], dobox->header.type[1], dobox->header.type[2], dobox->header.type[3]);
    
    if ((content_size = spGetMp4BoxHeaderContentSize(&dobox->header)) > 0) {
        if ((nwrite = (long)fwrite(dobox->data, 1, (long)content_size, fp)) != content_size) {
            return nwrite;
        }
        total_nwrite += nwrite;
    }
    
    spDebug(50, "spWriteMp4DataOnlyBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4DataOnlyBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    spMp4BoxSize content_size;
    spMp4DataOnlyBox *dest_dobox = (spMp4DataOnlyBox *)dest_box;
    spMp4DataOnlyBox *src_dobox = (spMp4DataOnlyBox *)src_box;

    if ((content_size = spGetMp4BoxHeaderContentSize(&src_dobox->header)) > 0) {
        dest_dobox->data = xalloc((long)content_size + 1, char);
        memcpy(dest_dobox->data, src_dobox->data, (long)content_size);
        dest_dobox->data[content_size] = NUL;
    } else {
        dest_dobox->data = NULL;
    }

    return content_size;
}

spMp4MetaBox *spAppendMp4UserDataMetaBox(spMp4Box *parent, spMp4BoxSize meta_size, const char *handler_type, const char *handler_name)
{
    spMp4UserDataBox *udta;
    spMp4MetaBox *meta;
    spMp4Box *ilst;

    spDebug(50, "spAppendMp4UserDataMetaBox", "in: meta_size = %lu\n", (unsigned long)meta_size);
    
    if ((udta = (spMp4UserDataBox *)spFindMp4Box(parent, "udta", NULL)) == NULL) {
        if ((udta = (spMp4UserDataBox *)spCreateMp4Box(parent, "udta")) == NULL) {
            spDebug(10, "spAppendMp4UserDataMetaBox", "error: cannot create 'udta' box\n");
            return NULL;
        }
    }
    
    if ((meta = (spMp4MetaBox *)spFindMp4Box((spMp4Box *)udta, "meta", "udta")) == NULL) {
        if ((meta = (spMp4MetaBox *)spCreateMp4Box((spMp4Box *)udta, "meta")) == NULL) {
            spDebug(10, "spAppendMp4UserDataMetaBox", "error: cannot create 'meta' box\n");
            return NULL;
        }
    }
    
    spUpdateMp4HandlerReferenceBox(meta->hdlr, handler_type, handler_name, SP_TRUE);
    
    if ((ilst = spFindMp4Box((spMp4Box *)meta, "ilst", "meta")) == NULL) {
        if ((ilst = spCreateMp4Box((spMp4Box *)meta, "ilst")) == NULL) {
            spDebug(10, "spAppendMp4UserDataMetaBox", "error: cannot create 'ilst' box\n");
            return NULL;
        }
    }

    if (meta_size > 0) {
        spMp4BoxSize marginsize;
        
        marginsize = meta_size - spGetMp4BoxHeaderContentSize(&meta->header) + meta->header.marginsize;
        spSetMp4BoxContentSize((spMp4Box *)meta, meta_size, SP_TRUE);
        spSetMp4BoxContentMarginSize((spMp4Box *)meta, marginsize);
    }
    
    spDebug(80, "spAppendMp4UserDataMetaBox", "done\n");
    
    return meta;
}

spMp4BoxSize spReadMp4UserDataStringBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4BoxSize content_size;
    spMp4BoxSize remain_size;
    spMp4UserDataStringBox *udsbox = (spMp4UserDataStringBox *)box;

    if ((nread = freadshort((short *)&udsbox->size, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;
    spDebug(50, "spReadMp4UserDataStringBox", "size = %d\n", udsbox->size);

    if ((nread = freadshort((short *)&udsbox->language, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4UserDataStringBox", "language = %d\n", udsbox->language);

    udsbox->string = xalloc(udsbox->size + 1, char);
    if ((nread = (long)fread(udsbox->string, 1, udsbox->size, fp)) != udsbox->size) {
        return nread;
    }
    udsbox->string[udsbox->size] = NUL;
    spDebug(50, "spReadMp4UserDataStringBox", "string = %s\n", udsbox->string);
    total_nread += nread;

    if ((content_size = spGetMp4BoxHeaderContentSize(&udsbox->header)) > total_nread) {
        spDebug(50, "spReadMp4UserDataStringBox", "content_size = %lu, total_nread = %lu\n", (unsigned long)content_size, (unsigned long)total_nread);
        remain_size = content_size - total_nread;
        if (remain_size < 8) {
            spDebug(50, "spReadMp4UserDataStringBox", "skip: remain_size = %lu\n", (unsigned long)remain_size);
            if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
                return 0;
            }
            total_nread += remain_size;
        }
    }
    
    spDebug(50, "spReadMp4UserDataStringBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spFreeMp4UserDataStringBox(void *box)
{
    spMp4UserDataStringBox *udsbox = (spMp4UserDataStringBox *)box;

    if (udsbox->string != NULL) {
        xfree(udsbox->string);
        udsbox->string = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4UserDataStringBox(spMp4UserDataStringBox *udsbox, short language, const char *string, spBool propagate_size)
{
    if (udsbox->string != NULL) {
        xfree(udsbox->string);
    }
    
    udsbox->size = (unsigned short)(strlen(string) + 1);
    udsbox->language = language;
    udsbox->string = strclone(string);

    spSetMp4BoxContentSize((spMp4Box *)udsbox, udsbox->size + 4, propagate_size);
    
    return udsbox->size + 4;
}

spMp4Box *spAppendMp4UserDataStringBox(spMp4Box *parent, const char *type, short language, const char *string)
{
    spMp4UserDataStringBox *udsbox;

    if ((udsbox = (spMp4UserDataStringBox *)spFindMp4Box(parent, type, parent->header.type)) == NULL) {
        if ((udsbox = (spMp4UserDataStringBox *)spCreateMp4Box(parent, type)) == NULL) {
            return NULL;
        }
    }
    spUpdateMp4UserDataStringBox(udsbox, language, string, SP_TRUE);

    return (spMp4Box *)udsbox;
}

spMp4BoxSize spWriteMp4UserDataStringBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4BoxSize content_size;
    spMp4BoxSize remain_size;
    spMp4UserDataStringBox *udsbox = (spMp4UserDataStringBox *)box;

    spDebug(50, "spWriteMp4UserDataStringBox", "size = %d\n", udsbox->size);
    if ((nwrite = fwriteshort((short *)&udsbox->size, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;

    spDebug(50, "spWriteMp4UserDataStringBox", "language = %d\n", udsbox->language);
    if ((nwrite = fwriteshort((short *)&udsbox->language, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    spDebug(50, "spWriteMp4UserDataStringBox", "string = %s\n", udsbox->string);
    if ((nwrite = (long)fwrite(udsbox->string, 1, udsbox->size, fp)) != udsbox->size) {
        return nwrite;
    }
    total_nwrite += nwrite;

    if (rewrite_flag == SP_TRUE) {
        if ((content_size = spGetMp4BoxHeaderContentSize(&udsbox->header)) > total_nwrite) {
            spDebug(50, "spWriteMp4UserDataStringBox", "content_size = %lu, total_nwrite = %lu\n", (unsigned long)content_size, (unsigned long)total_nwrite);
            remain_size = content_size - total_nwrite;
            if (remain_size < 8) {
                spDebug(50, "spWriteMp4UserDataStringBox", "skip: remain_size = %lu\n", (unsigned long)remain_size);
                if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
                    return 0;
                }
                total_nwrite += remain_size;
            }
        }
    }
    
    spDebug(50, "spWriteMp4UserDataStringBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4UserDataStringBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    spMp4UserDataStringBox *dest_udsbox = (spMp4UserDataStringBox *)dest_box;
    spMp4UserDataStringBox *src_udsbox = (spMp4UserDataStringBox *)src_box;

    dest_udsbox->size = src_udsbox->size;
    dest_udsbox->language = src_udsbox->language;
    
    dest_udsbox->string = xalloc(dest_udsbox->size + 1, char);
    memcpy(dest_udsbox->string, src_udsbox->string, dest_udsbox->size);
    dest_udsbox->string[dest_udsbox->size] = NUL;
    
    return sizeof(dest_udsbox->size) + sizeof(dest_udsbox->language) + dest_udsbox->size;
}

/* 'data' */
spMp4BoxSize spReadMp4MetaDataValueBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4BoxSize content_size;
    spMp4BoxSize value_size;
    spMp4MetaDataValueBox *databox = (spMp4MetaDataValueBox *)box;

    if ((nread = freadulong32(&databox->type_indicator, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4MetaDataValueBox", "type_indicator = %ld\n", databox->type_indicator);
    
    if ((nread = freadshort((short *)&databox->country, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4MetaDataValueBox", "country = %d\n", databox->country);

    if ((nread = freadshort((short *)&databox->language, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4MetaDataValueBox", "language = %d\n", databox->language);

    if ((content_size = spGetMp4BoxHeaderContentSize(&databox->header)) > total_nread) {
        value_size = content_size - total_nread;
        spDebug(50, "spReadMp4MetaDataValueBox", "value_size = %lu\n", (unsigned long)value_size);
        databox->value = xalloc((long)value_size + 1, char);
        if ((nread = (long)fread(databox->value, 1, (long)value_size, fp)) != value_size) {
            return nread;
        }
        databox->value[value_size] = NUL;
        if (databox->type_indicator == 1) { /* UTF8 */
            spDebug(50, "spReadMp4MetaDataValueBox", "value = %s\n", databox->value);
        }
        
        total_nread += nread;
    } else {
        databox->value = NULL;
    }

    spDebug(50, "spReadMp4MetaDataValueBox", "done: total_nread = %lu / %lu\n",
            (unsigned long)total_nread, (unsigned long)content_size);
    
    return total_nread;
}

spBool spFreeMp4MetaDataValueBox(void *box)
{
    spMp4MetaDataValueBox *databox = (spMp4MetaDataValueBox *)box;

    if (databox->value != NULL) {
        xfree(databox->value);
        databox->value = NULL;
    }
    
    return SP_TRUE;
}

static spMp4BoxSize updateMp4MetaDataValueBox(spMp4MetaDataValueBox *databox,
                                              unsigned long type_indicator, unsigned short country, unsigned short language,
                                              const char *value, spMp4BoxSize value_size, spBool alloc_flag, spBool propagate_size)
{
    spDebug(50, "updateMp4MetaDataValueBox", "type_indicator = %lu, country = %d, language = %d\n",
            type_indicator, country, language);
    
    if (databox->value != NULL) {
        xfree(databox->value);
    }
    
    databox->type_indicator = type_indicator;
    databox->country = country;
    databox->language = language;
    if (alloc_flag == SP_TRUE) {
        databox->value = xalloc((long)value_size + 1, char);
        memcpy(databox->value, value, (long)value_size);
        databox->value[value_size] = NUL;
    } else {
        databox->value = (char *)value;
    }

    spSetMp4BoxContentSize((spMp4Box *)databox, value_size + 8, propagate_size);
    
    return value_size + 8;
}

spMp4BoxSize spUpdateMp4MetaDataValueBox(spMp4MetaDataValueBox *databox,
                                         unsigned long type_indicator, unsigned short country, unsigned short language,
                                         const char *value, spMp4BoxSize value_size, spBool propagate_size)
{
    return updateMp4MetaDataValueBox(databox, type_indicator, country, language,
                                     value, value_size, SP_TRUE, propagate_size);
}

static spMp4MetaDataValueBox *appendMp4MetaDataValueBox(spMp4Box *parent, const char *type, unsigned long type_indicator,
                                                        unsigned short country, unsigned short language,
                                                        const char *value, spMp4BoxSize value_size, spBool alloc_flag)
{
    spMp4Box *container;
    spMp4MetaDataValueBox *databox;

    spDebug(50, "spAppendMp4MetaDataValueBox", "type = %c%c%c%c\n",
            type[0], type[1], type[2], type[3]);
    
    if ((container = spFindMp4Box(parent, type, parent->header.type)) == NULL) {
        if ((container = spCreateMp4Box(parent, type)) == NULL) {
            return NULL;
        }
    }
    
    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box(container, "data", type)) == NULL) {
        if ((databox = (spMp4MetaDataValueBox *)spCreateMp4Box(container, "data")) == NULL) {
            return NULL;
        }
    }
    updateMp4MetaDataValueBox(databox, type_indicator, country, language, value, value_size, alloc_flag, SP_TRUE);

    return databox;
}

spMp4MetaDataValueBox *spAppendMp4MetaDataValueBox(spMp4Box *parent, const char *type, unsigned long type_indicator,
                                                   unsigned short country, unsigned short language,
                                                   const char *value, spMp4BoxSize value_size)
{
    return appendMp4MetaDataValueBox(parent, type, type_indicator, country, language,
                                     value, value_size, SP_TRUE);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataUTF8String(spMp4MetaBox *meta, const char *type,
                                                     unsigned short country, unsigned short language, const char *utf8str)
{
    if (meta == NULL) return NULL;

    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return NULL;
        }
    }

    return appendMp4MetaDataValueBox(meta->ilst, type, SP_MP4_META_DATA_TYPE_INDICATOR_UTF8, country, language,
                                     utf8str, strlen(utf8str), SP_TRUE);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataUTF8StringFromLocaleCode(spMp4MetaBox *meta, const char *type,
                                                                   unsigned short country, unsigned short language, const unsigned char *str_in_locale)
{
    long size;
    unsigned char *utf8str;
    
    if (meta == NULL) return NULL;
    
    spDebug(50, "spAppendMp4MetaDataUTF8StringFromLocaleCode", "str_in_locale = %s\n", str_in_locale);

    size = (long)(strlen((char *)str_in_locale) + 1) * 2;
    utf8str = xalloc(size, unsigned char);
    spStrCopy((char *)utf8str, size, (const char *)str_in_locale);
    
    spConvertKanjiFromLocaleCode(utf8str, size, SP_KANJI_CODE_UTF8);
    spDebug(50, "spAppendMp4MetaDataUTF8StringFromLocaleCode", "utf8str = %s\n", utf8str);
    
    return appendMp4MetaDataValueBox(meta->ilst, type, SP_MP4_META_DATA_TYPE_INDICATOR_UTF8, country, language,
                                     (const char *)utf8str, strlen((char *)utf8str), SP_FALSE);
}

spBool spParseMp4MetaDataUTF8StringToLocaleCode(spMp4MetaDataValueBox *databox, unsigned char *str_in_locale, long size)
{
    if (databox == NULL || databox->value == NULL
        || databox->type_indicator != SP_MP4_META_DATA_TYPE_INDICATOR_UTF8) return SP_FALSE;

    spStrCopy((char *)str_in_locale, size, databox->value);
    
    spConvertKanjiToLocaleCode((unsigned char *)str_in_locale, size, SP_KANJI_CODE_UTF8);
    
    return SP_TRUE;
}

char *xspParseMp4MetaDataUTF8StringToLocaleCode(spMp4MetaDataValueBox *databox)
{
    long size;
    unsigned char *str_in_locale;
    
    if (databox == NULL || databox->value == NULL
        || databox->type_indicator != SP_MP4_META_DATA_TYPE_INDICATOR_UTF8) return NULL;

    size = (long)(strlen(databox->value) + 1) * 2;
    str_in_locale = xalloc(size, unsigned char);
    
    spStrCopy((char *)str_in_locale, size, databox->value);
    spConvertKanjiToLocaleCode(str_in_locale, size, SP_KANJI_CODE_UTF8);
    
    return (char *)str_in_locale;
}

const char *spGetMp4MetaDataUTF8String(spMp4MetaBox *meta, const char *type)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", type)) == NULL) {
        return NULL;
    }

    return databox->value;
}

spBool spGetMp4MetaDataUTF8StringToLocaleCode(spMp4MetaBox *meta, const char *type, unsigned char *str_in_locale, long size)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", type)) == NULL) {
        return SP_FALSE;
    }

    return spParseMp4MetaDataUTF8StringToLocaleCode(databox, str_in_locale, size);
}

char *xspGetMp4MetaDataUTF8StringToLocaleCode(spMp4MetaBox *meta, const char *type)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", type)) == NULL) {
        return NULL;
    }

    return xspParseMp4MetaDataUTF8StringToLocaleCode(databox);
}

unsigned long spParseMp4MetaDataBESignedInt(spMp4MetaDataValueBox *databox, long *number)
{
    spMp4BoxSize i;
    spMp4BoxSize size;
    long shift;
    
    if (databox == NULL || databox->value == NULL || number == NULL) return 0;
    
    if ((size = spGetMp4BoxHeaderContentSize(&databox->header)) <= 8) {
        return 0;
    }
    size -= 8;

    shift = 0;
    *number = 0;
    for (i = size - 1; i >= 0; i--) {
        *number |= ((long)databox->value[i]) << shift;

        shift += 8;
    }
    
    return (unsigned long)size;
}

unsigned long spParseMp4MetaDataBEUnsignedInt(spMp4MetaDataValueBox *databox, unsigned long *number)
{
    spMp4BoxSize i;
    spMp4BoxSize size;
    unsigned long shift;
    
    if (databox == NULL || databox->value == NULL || number == NULL) return 0;
    
    if ((size = spGetMp4BoxHeaderContentSize(&databox->header)) <= 8) {
        return 0;
    }
    size -= 8;

    shift = 0;
    *number = 0;
    for (i = size - 1; i >= 0; i--) {
        *number |= ((unsigned long)databox->value[i]) << shift;

        shift += 8;
    }
    
    return (unsigned long)size;
}

spBool spGetMp4MetaDataTempo(spMp4MetaBox *meta, long *tempo)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", SP_MP4_META_DATA_TYPE_TEMPO)) == NULL) {
        return SP_FALSE;
    }

    if (spParseMp4MetaDataBESignedInt(databox, tempo) > 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spGetMp4MetaDataBoolean(spMp4MetaBox *meta, const char *type, spBool *flag)
{
    spMp4MetaDataValueBox *databox;
    long number;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", type)) == NULL) {
        return SP_FALSE;
    }

    number = 0;
    if (spParseMp4MetaDataBESignedInt(databox, &number) > 0) {
        *flag = (spBool)(!!number);
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spGetMp4MetaDataGapless(spMp4MetaBox *meta, spBool *flag)
{
    return spGetMp4MetaDataBoolean(meta, SP_MP4_META_DATA_TYPE_GAPLESS, flag);
}

spBool spGetMp4MetaDataCompilation(spMp4MetaBox *meta, spBool *flag)
{
    return spGetMp4MetaDataBoolean(meta, SP_MP4_META_DATA_TYPE_COMPILATION, flag);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataBESignedInt(spMp4MetaBox *meta, const char *type, long number, unsigned long size)
{
    unsigned long i;
    char *value;
    spMp4MetaDataValueBox *databox;
    long shift;

    if (meta == NULL) return NULL;

    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return NULL;
        }
    }

    value = xalloc(size + 1, char);
    memset(value, 0, size + 1);

    shift = 0;
    for (i = size - 1;; i--) {
        value[i] = (char)(0xff & (number >> shift));

        shift += 8;
        
        if (i == 0) {
            break;
        }
    }

    databox = appendMp4MetaDataValueBox(meta->ilst, type, SP_MP4_META_DATA_TYPE_INDICATOR_BE_SIGNED_INT, 0, 0,
                                        value, size, SP_FALSE);
    if (databox == NULL) {
        xfree(value);
    }

    return databox;
}

spMp4MetaDataValueBox *spAppendMp4MetaDataBEUnsignedInt(spMp4MetaBox *meta, const char *type, unsigned long number, unsigned long size)
{
    unsigned long i;
    char *value;
    spMp4MetaDataValueBox *databox;
    unsigned long shift;

    if (meta == NULL) return NULL;

    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return NULL;
        }
    }

    value = xalloc(size + 1, char);
    memset(value, 0, size + 1);

    shift = 0;
    for (i = size - 1;; i--) {
        value[i] = (char)(0xff & (number >> shift));

        shift += 8;

        if (i == 0) {
            break;
        }
    }

    databox = appendMp4MetaDataValueBox(meta->ilst, type, SP_MP4_META_DATA_TYPE_INDICATOR_BE_UNSIGNED_INT, 0, 0,
                                        value, size, SP_FALSE);
    if (databox == NULL) {
        xfree(value);
    }

    return databox;
}

spMp4MetaDataValueBox *spAppendMp4MetaDataTempo(spMp4MetaBox *meta, long tempo)
{
    return spAppendMp4MetaDataBESignedInt(meta, SP_MP4_META_DATA_TYPE_TEMPO, tempo, 2);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataBoolean(spMp4MetaBox *meta, const char *type, spBool flag)
{
    char *value;
    spMp4BoxSize value_size;
    spMp4MetaDataValueBox *databox;

    if (meta == NULL) return NULL;

    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return NULL;
        }
    }

    value_size = 1;
    value = xalloc((long)value_size + 1, char);
    memset(value, 0, (long)value_size + 1);
    value[0] = (char)(!!flag);

    databox = appendMp4MetaDataValueBox(meta->ilst, type, SP_MP4_META_DATA_TYPE_INDICATOR_BE_SIGNED_INT, 0, 0,
                                        value, value_size, SP_FALSE);
    if (databox == NULL) {
        xfree(value);
    }

    return databox;
}

spMp4MetaDataValueBox *spAppendMp4MetaDataGapless(spMp4MetaBox *meta, spBool flag)
{
    return spAppendMp4MetaDataBoolean(meta, SP_MP4_META_DATA_TYPE_GAPLESS, flag);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataCompilation(spMp4MetaBox *meta, spBool flag)
{
    return spAppendMp4MetaDataBoolean(meta, SP_MP4_META_DATA_TYPE_COMPILATION, flag);
}

spBool spParseMp4MetaDataTrackNumber(spMp4MetaDataValueBox *databox, unsigned short *number, unsigned short *total)
{
    if (databox == NULL || databox->value == NULL) return SP_FALSE;
    
    if (number != NULL) {
        *number = (((unsigned short)databox->value[2]) << 8) | (unsigned short)databox->value[3];
    }
    if (total != NULL) {
        *total = (((unsigned short)databox->value[4]) << 8) | (unsigned short)databox->value[5];
    }
    
    return SP_TRUE;
}

spBool spParseMp4MetaDataDiscNumber(spMp4MetaDataValueBox *databox, unsigned short *number, unsigned short *total)
{
    return spParseMp4MetaDataTrackNumber(databox, number, total);
}

spBool spGetMp4MetaDataTrackNumber(spMp4MetaBox *meta, unsigned short *number, unsigned short *total)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", SP_MP4_META_DATA_TYPE_TRACK)) == NULL) {
        return SP_FALSE;
    }

    return spParseMp4MetaDataTrackNumber(databox, number, total);
}

spBool spGetMp4MetaDataDiscNumber(spMp4MetaBox *meta, unsigned short *number, unsigned short *total)
{
    spMp4MetaDataValueBox *databox;

    if ((databox = (spMp4MetaDataValueBox *)spFindMp4Box((spMp4Box *)meta, "data", SP_MP4_META_DATA_TYPE_DISC)) == NULL) {
        return SP_FALSE;
    }

    return spParseMp4MetaDataDiscNumber(databox, number, total);
}

static spMp4MetaDataValueBox *appendMp4MetaDataNumber(spMp4MetaBox *meta, const char *type, unsigned short number, unsigned short total)
{
    char *value;
    spMp4BoxSize value_size;
    spMp4MetaDataValueBox *databox;

    if (meta == NULL) return NULL;

    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return NULL;
        }
    }

    if (strneq("trkn", type, 4)) {
        value_size = 8;
    } else if (strneq("disc", type, 4)) {
        value_size = 6;
    } else {
        return NULL;
    }
    value = xalloc((long)value_size + 1, char);
    memset(value, 0, (long)value_size + 1);
    value[2] = (char)((0xff00 & number) >> 8);
    value[3] = (char)(0xff & number);
    value[4] = (char)((0xff00 & total) >> 8);
    value[5] = (char)(0xff & total);

    databox = appendMp4MetaDataValueBox(meta->ilst, type, 0, 0, 0,
                                        value, value_size, SP_FALSE);
    if (databox == NULL) {
        xfree(value);
    }

    return databox;
}

spMp4MetaDataValueBox *spAppendMp4MetaDataTrackNumber(spMp4MetaBox *meta, unsigned short number, unsigned short total)
{
    return appendMp4MetaDataNumber(meta, SP_MP4_META_DATA_TYPE_TRACK, number, total);
}

spMp4MetaDataValueBox *spAppendMp4MetaDataDiscNumber(spMp4MetaBox *meta, unsigned short number, unsigned short total)
{
    return appendMp4MetaDataNumber(meta, SP_MP4_META_DATA_TYPE_DISC, number, total);
}

spMp4BoxSize spWriteMp4MetaDataValueBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4BoxSize content_size;
    spMp4BoxSize value_size;
    spMp4MetaDataValueBox *databox = (spMp4MetaDataValueBox *)box;

    spDebug(50, "spWriteMp4MetaDataValueBox", "type_indicator = %lu\n", databox->type_indicator);
    if ((nwrite = fwriteulong32(&databox->type_indicator, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    spDebug(50, "spWriteMp4MetaDataValueBox", "country = %d\n", databox->country);
    if ((nwrite = fwriteshort((short *)&databox->country, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    spDebug(50, "spWriteMp4MetaDataValueBox", "language = %d\n", databox->language);
    if ((nwrite = fwriteshort((short *)&databox->language, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((content_size = spGetMp4BoxHeaderContentSize(&databox->header)) > total_nwrite) {
        value_size = content_size - total_nwrite;
        spDebug(50, "spWriteMp4MetaDataValueBox", "value_size = %lu\n", (unsigned long)value_size);
        if ((nwrite = (long)fwrite(databox->value, 1, (long)value_size, fp)) != value_size) {
            return nwrite;
        }
        total_nwrite += nwrite;
    }

    spDebug(50, "spWriteMp4MetaDataValueBox", "done: total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, (unsigned long)content_size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4MetaDataValueBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    spMp4BoxSize total_ncopy;
    spMp4BoxSize content_size;
    spMp4BoxSize value_size;
    spMp4MetaDataValueBox *dest_databox = (spMp4MetaDataValueBox *)dest_box;
    spMp4MetaDataValueBox *src_databox = (spMp4MetaDataValueBox *)src_box;

    dest_databox->type_indicator = src_databox->type_indicator;
    dest_databox->country = src_databox->country;
    dest_databox->language = src_databox->language;
    total_ncopy = sizeof(dest_databox->type_indicator) + sizeof(dest_databox->country)
        + sizeof(dest_databox->language);

    if ((content_size = spGetMp4BoxHeaderContentSize(&src_databox->header)) > 8) {
        value_size = content_size - 8;
        
        dest_databox->value = xalloc((long)value_size + 1, char);
        memcpy(dest_databox->value, src_databox->value, (long)value_size);
        dest_databox->value[value_size] = NUL;

        total_ncopy += value_size;
    } else {
        dest_databox->value = NULL;
    }

    return total_ncopy;
}

/* 'stsh' */
spMp4BoxSize spReadMp4ShadowSyncSampleBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4ShadowSyncSampleBox *stsh = (spMp4ShadowSyncSampleBox *)box;

    if ((nread = freadulong32(&stsh->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4ShadowSyncSampleBox", "entry_count = %ld\n", stsh->entry_count);
    
    if (stsh->entry_count > 0) {
        stsh->alloc_count = (1 + stsh->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stsh->entries = xalloc(stsh->alloc_count, spMp4ShadowSyncSampleEntry);
        for (i = 0; i < stsh->entry_count; i++) {
            if ((nread = freadulong32(&stsh->entries[i].shadowed_sample_number, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            if ((nread = freadulong32(&stsh->entries[i].sync_sample_number, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            spDebug(80, "spReadMp4ShadowSyncSampleBox",
                    "%ld: shadowed_sample_number = %ld, sync_sample_number = %ld\n",
                    stsh->entries[i].shadowed_sample_number, stsh->entries[i].sync_sample_number);
        }
    } else {
        stsh->alloc_count = 0;
        stsh->entries = NULL;
    }
        
    return total_nread;
}

spBool spFreeMp4ShadowSyncSampleBox(void *box)
{
    spMp4ShadowSyncSampleBox *stsh = (spMp4ShadowSyncSampleBox *)box;

    if (stsh->entries != NULL) {
        xfree(stsh->entries);
        stsh->entries = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spAppendMp4ShadowSyncSample(spMp4ShadowSyncSampleBox *stsh,
                                          unsigned long shadowed_sample_number, unsigned long sync_sample_number)
{
    if (stsh == NULL) return 0;

    incrementEntryCount((spMp4Box *)stsh, 8, (char **)&stsh->entries, sizeof(spMp4ShadowSyncSampleEntry), &stsh->alloc_count, &stsh->entry_count);
    stsh->entries[stsh->entry_count - 1].shadowed_sample_number = shadowed_sample_number;
    stsh->entries[stsh->entry_count - 1].sync_sample_number = sync_sample_number;
    
    return stsh->entry_count;
}

spMp4BoxSize spWriteMp4ShadowSyncSampleBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4ShadowSyncSampleBox *stsh = (spMp4ShadowSyncSampleBox *)box;

    spDebug(50, "spWriteMp4ShadowSyncSampleBox", "entry_count = %ld\n", stsh->entry_count);
    if ((nwrite = fwriteulong32(&stsh->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (stsh->entry_count > 0) {
        for (i = 0; i < stsh->entry_count; i++) {
            spDebug(80, "spWriteMp4ShadowSyncSampleBox",
                    "%ld: shadowed_sample_number = %ld, sync_sample_number = %ld\n",
                    stsh->entries[i].shadowed_sample_number, stsh->entries[i].sync_sample_number);
            
            if ((nwrite = fwriteulong32(&stsh->entries[i].shadowed_sample_number, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
            
            if ((nwrite = fwriteulong32(&stsh->entries[i].sync_sample_number, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
        }
    }
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4ShadowSyncSampleBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4ShadowSyncSampleBox *dest_stsh = (spMp4ShadowSyncSampleBox *)dest_box;
    spMp4ShadowSyncSampleBox *src_stsh = (spMp4ShadowSyncSampleBox *)src_box;

    dest_stsh->entry_count = src_stsh->entry_count;
    total_ncopy = sizeof(dest_stsh->entry_count);

    total_ncopy += sizeof(dest_stsh->alloc_count) + sizeof(dest_stsh->entries);
    
    if (dest_stsh->entry_count > 0) {
        dest_stsh->alloc_count = src_stsh->alloc_count;
        dest_stsh->entries = xalloc(dest_stsh->alloc_count, spMp4ShadowSyncSampleEntry);
        for (i = 0; i < dest_stsh->entry_count; i++) {
            dest_stsh->entries[i].shadowed_sample_number = src_stsh->entries[i].shadowed_sample_number;
            total_ncopy += sizeof(dest_stsh->entries[i].shadowed_sample_number);
            
            dest_stsh->entries[i].sync_sample_number = src_stsh->entries[i].sync_sample_number;
            total_ncopy += sizeof(dest_stsh->entries[i].sync_sample_number);
        }
    } else {
        dest_stsh->alloc_count = 0;
        dest_stsh->entries = NULL;
    }
    
    return total_ncopy;
}

/* 'stss' */
spMp4BoxSize spReadMp4SyncSampleBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4SyncSampleBox *stss = (spMp4SyncSampleBox *)box;

    if ((nread = freadulong32(&stss->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4SyncSampleBox", "entry_count = %ld\n", stss->entry_count);
    
    if (stss->entry_count > 0) {
        stss->alloc_count = (1 + stss->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stss->sample_number = xalloc(stss->alloc_count, unsigned long);
        for (i = 0; i < stss->entry_count; i++) {
            if ((nread = freadulong32(&stss->sample_number[i], 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            spDebug(80, "spReadMp4SyncSampleBox", "sample_number[%ld] = %ld\n",
                    i, stss->sample_number[i]);
        }
    } else {
        stss->alloc_count = 0;
        stss->sample_number = NULL;
    }
        
    return total_nread;
}

spBool spFreeMp4SyncSampleBox(void *box)
{
    spMp4SyncSampleBox *stss = (spMp4SyncSampleBox *)box;

    if (stss->sample_number != NULL) {
        xfree(stss->sample_number);
        stss->sample_number = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spAppendMp4SyncSample(spMp4SyncSampleBox *stss, unsigned long sample)
{
    if (stss == NULL) return 0;

    incrementEntryCount((spMp4Box *)stss, 4, (char **)&stss->sample_number, sizeof(unsigned long), &stss->alloc_count, &stss->entry_count);
    stss->sample_number[stss->entry_count - 1] = sample;
    
    return stss->entry_count;
}

spMp4BoxSize spWriteMp4SyncSampleBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SyncSampleBox *stss = (spMp4SyncSampleBox *)box;

    spDebug(50, "spWriteMp4SyncSampleBox", "entry_count = %ld\n", stss->entry_count);
    if ((nwrite = fwriteulong32(&stss->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (stss->entry_count > 0) {
        for (i = 0; i < stss->entry_count; i++) {
            spDebug(80, "spWriteMp4SyncSampleBox", "sample_number[%ld] = %ld\n",
                    i, stss->sample_number[i]);
            if ((nwrite = fwriteulong32(&stss->sample_number[i], 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
        }
    }
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SyncSampleBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4SyncSampleBox *dest_stss = (spMp4SyncSampleBox *)dest_box;
    spMp4SyncSampleBox *src_stss = (spMp4SyncSampleBox *)src_box;

    dest_stss->entry_count = src_stss->entry_count;
    total_ncopy = sizeof(dest_stss->entry_count);
    
    total_ncopy += sizeof(dest_stss->alloc_count) + sizeof(dest_stss->sample_number);
    
    if (dest_stss->entry_count > 0) {
        dest_stss->alloc_count = src_stss->alloc_count;
        dest_stss->sample_number = xalloc(dest_stss->alloc_count, unsigned long);
        for (i = 0; i < dest_stss->entry_count; i++) {
            dest_stss->sample_number[i] = src_stss->sample_number[i];
            total_ncopy += sizeof(dest_stss->sample_number[i]);
        }
    } else {
        dest_stss->alloc_count = 0;
        dest_stss->sample_number = NULL;
    }
        
    return total_ncopy;
}

/* 'stco' */
spMp4BoxSize spReadMp4ChunkOffsetBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    unsigned long offset;
    char largesize[8];
    spMp4BoxSize total_nread;
    spBool co64_flag;
    spMp4ChunkOffsetBox *stco = (spMp4ChunkOffsetBox *)box;

    if ((nread = freadulong32(&stco->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4ChunkOffsetBox", "entry_count = %ld\n", stco->entry_count);
    
    if (stco->entry_count > 0) {
        if (strneq("co64", stco->header.type, 4)) {
            co64_flag = SP_TRUE;
        } else {
            co64_flag = SP_FALSE;
        }
        
        stco->alloc_count = (1 + stco->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stco->chunk_offset = xalloc(stco->alloc_count, spFileOffset);
        for (i = 0; i < stco->entry_count; i++) {
            if (co64_flag == SP_TRUE) {
                if ((nread = (long)fread(largesize, 1, 8, fp)) != 8) {
                    return nread;
                }
                total_nread += 8;
#ifdef SP_SUPPORT_64BIT_INT
                stco->chunk_offset[i] = spC64ToLong64(largesize, SP_AIFF_NEED_SWAP);
#else
                stco->chunk_offset[i] = spC64ToULong(largesize, SP_AIFF_NEED_SWAP);
#endif
            } else {
                if ((nread = freadulong32(&offset, 1, swap, fp)) != 1) {
                    return nread;
                }
                total_nread += 4;
                stco->chunk_offset[i] = offset;
            }
            spDebug(80, "spReadMp4ChunkOffsetBox", "chunk_offset[%ld] = %lu\n",
                    i, (unsigned long)stco->chunk_offset[i]);
        }
    } else {
        stco->alloc_count = 0;
        stco->chunk_offset = NULL;
    }
        
    spDebug(50, "spReadMp4ChunkOffsetBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stco->header.size);
        
    return total_nread;
}

spBool spFreeMp4ChunkOffsetBox(void *box)
{
    spMp4ChunkOffsetBox *stco = (spMp4ChunkOffsetBox *)box;

    if (stco->chunk_offset != NULL) {
        xfree(stco->chunk_offset);
        stco->chunk_offset = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spAppendMp4ChunkOffset(spMp4ChunkOffsetBox *stco, spFileOffset offset)
{
    long size_in_file;
    
    if (stco == NULL) return 0;

    spDebug(50, "spAppendMp4ChunkOffset", "in: entry_count = %ld\n", stco->entry_count);
    
    if (strneq("co64", stco->header.type, 4)) {
        size_in_file = 8;
    } else {
        size_in_file = 4;
    }

    incrementEntryCount((spMp4Box *)stco, size_in_file, (char **)&stco->chunk_offset, sizeof(spFileOffset), &stco->alloc_count, &stco->entry_count);
    stco->chunk_offset[stco->entry_count - 1] = offset;
    spDebug(50, "spAppendMp4ChunkOffset", "entry_count incremented: entry_count = %ld, offset = %lu\n",
            stco->entry_count, (unsigned long)offset);
    
    return stco->entry_count;
}

spMp4BoxSize spWriteMp4ChunkOffsetBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    unsigned long offset;
    char largesize[8];
    spMp4BoxSize total_nwrite;
    spBool co64_flag;
    spMp4ChunkOffsetBox *stco = (spMp4ChunkOffsetBox *)box;

    spDebug(50, "spWriteMp4ChunkOffsetBox", "entry_count = %ld\n", stco->entry_count);
    if ((nwrite = fwriteulong32(&stco->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (stco->entry_count > 0) {
        if (strneq("co64", stco->header.type, 4)) {
            co64_flag = SP_TRUE;
        } else {
            co64_flag = SP_FALSE;
        }
        
        for (i = 0; i < stco->entry_count; i++) {
            spDebug(80, "spWriteMp4ChunkOffsetBox", "chunk_offset[%ld] = %lu\n", i, (unsigned long)stco->chunk_offset[i]);
            
            if (co64_flag == SP_TRUE) {
#ifdef SP_SUPPORT_64BIT_INT
                spLong64ToC64(stco->chunk_offset[i], largesize, SP_AIFF_NEED_SWAP);
#else
                spULongToC64(stco->chunk_offset[i], largesize, SP_AIFF_NEED_SWAP);
#endif
                if ((nwrite = (long)fwrite(largesize, 1, 8, fp)) != 8) {
                    return nwrite;
                }
                total_nwrite += 8;
            } else {
                offset = (unsigned long)stco->chunk_offset[i];
                if ((nwrite = fwriteulong32(&offset, 1, swap, fp)) != 1) {
                    return nwrite;
                }
                total_nwrite += 4;
            }
        }
    }
        
    spDebug(50, "spWriteMp4ChunkOffsetBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stco->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4ChunkOffsetBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4ChunkOffsetBox *dest_stco = (spMp4ChunkOffsetBox *)dest_box;
    spMp4ChunkOffsetBox *src_stco = (spMp4ChunkOffsetBox *)src_box;

    dest_stco->entry_count = src_stco->entry_count;
    total_ncopy = sizeof(dest_stco->entry_count);

    total_ncopy += sizeof(dest_stco->alloc_count) + sizeof(dest_stco->chunk_offset);
    
    if (dest_stco->entry_count > 0) {
        dest_stco->alloc_count = src_stco->alloc_count;
        dest_stco->chunk_offset = xalloc(dest_stco->alloc_count, spFileOffset);
        for (i = 0; i < dest_stco->entry_count; i++) {
            dest_stco->chunk_offset[i] = src_stco->chunk_offset[i];
            total_ncopy += sizeof(dest_stco->chunk_offset[i]);
        }
    } else {
        dest_stco->alloc_count = 0;
        dest_stco->chunk_offset = NULL;
    }
        
    return total_ncopy;
}

/* 'stz2' */
spMp4BoxSize spReadMp4CompactSampleSizeBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    char cvalue;
    short svalue;
    char buf[4];
    spMp4SampleSizeBox *stz2 = (spMp4SampleSizeBox *)box;
    
    if ((nread = (long)fread(buf, 1, 3, fp)) != 3) {
        return nread;
    }
    if ((nread = (long)fread(&stz2->field_size, 1, 1, fp)) != 1
        || !(stz2->field_size == 4 || stz2->field_size == 8 || stz2->field_size == 16
             || stz2->field_size == 32)) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4CompactSampleSizeBox", "field_size = %d\n", stz2->field_size);
    
    if ((nread = freadulong32(&stz2->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4CompactSampleSizeBox", "entry_count = %ld\n", stz2->entry_count);
    
    if (stz2->entry_count > 0) {
        cvalue = 0;
        stz2->entries = xalloc(stz2->entry_count, unsigned long);
        for (i = 0; i < stz2->entry_count; i++) {
            if (stz2->field_size == 4) {
                if (i % 2 == 1) {
                    if ((nread = (long)fread(&cvalue, 1, 1, fp)) != 1) {
                        return nread;
                    }
                    total_nread += 1;
                    
                    stz2->entries[i - 1] = (0xf0 & cvalue) >> 4;
                    stz2->entries[i] = 0xf & cvalue;
                }
            } else if (stz2->field_size == 8) {
                if ((nread = (long)fread(&cvalue, 1, 1, fp)) != 1) {
                    return nread;
                }
                total_nread += 1;
                
                stz2->entries[i] = cvalue;
            } else if (stz2->field_size == 16) {
                if ((nread = freadshort(&svalue, 1, swap, fp)) != 1) {
                    return nread;
                }
                total_nread += 2;
                
                stz2->entries[i] = svalue;
            } else {
                if ((nread = freadulong32(&stz2->entries[i], 1, swap, fp)) != 1) {
                    return nread;
                }
                total_nread += 4;
            }
            spDebug(80, "spReadMp4CompactSampleSizeBox", "entries[%ld] = %ld\n", i, stz2->entries[i]);
        }
    } else {
        stz2->entries = NULL;
    }
        
    spDebug(50, "spReadMp4CompactSampleSizeBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stz2->header.size);
        
    return total_nread;
}

spBool spFreeMp4CompactSampleSizeBox(void *box)
{
    spMp4SampleSizeBox *stz2 = (spMp4SampleSizeBox *)box;

    if (stz2->entries != NULL) {
        xfree(stz2->entries);
        stz2->entries = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spWriteMp4CompactSampleSizeBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    char cvalue;
    short svalue;
    spMp4SampleSizeBox *stz2 = (spMp4SampleSizeBox *)box;
    static char buf[3] = {0, 0, 0};
    
    if ((nwrite = (long)fwrite(buf, 1, 3, fp)) != 3) {
        return nwrite;
    }
    spDebug(50, "spWriteMp4CompactSampleSizeBox", "field_size = %d\n", stz2->field_size);
    if ((nwrite = (long)fwrite(&stz2->field_size, 1, 1, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    spDebug(50, "spWriteMp4CompactSampleSizeBox", "entry_count = %ld\n", stz2->entry_count);
    if ((nwrite = fwriteulong32(&stz2->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if (stz2->entry_count > 0) {
        cvalue = 0;
        for (i = 0; i < stz2->entry_count; i++) {
            spDebug(80, "spWriteMp4CompactSampleSizeBox", "entries[%ld] = %ld\n", i, stz2->entries[i]);
            
            if (stz2->field_size == 4) {
                if (i % 2 == 1) {
                    cvalue = (0xf & stz2->entries[i - 1]) << 4;
                    cvalue |= (0xf & stz2->entries[i]);
                    if ((nwrite = (long)fwrite(&cvalue, 1, 1, fp)) != 1) {
                        return nwrite;
                    }
                    total_nwrite += 1;
                }
            } else if (stz2->field_size == 8) {
                cvalue = (char)(0xff & stz2->entries[i]);
                
                if ((nwrite = (long)fwrite(&cvalue, 1, 1, fp)) != 1) {
                    return nwrite;
                }
                total_nwrite += 1;
            } else if (stz2->field_size == 16) {
                svalue = (short)stz2->entries[i];
                
                if ((nwrite = fwriteshort(&svalue, 1, swap, fp)) != 1) {
                    return nwrite;
                }
                total_nwrite += 2;
            } else {
                if ((nwrite = fwriteulong32(&stz2->entries[i], 1, swap, fp)) != 1) {
                    return nwrite;
                }
                total_nwrite += 4;
            }
        }
    }
        
    spDebug(50, "spWriteMp4CompactSampleSizeBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stz2->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4CompactSampleSizeBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4SampleSizeBox *dest_stz2 = (spMp4SampleSizeBox *)dest_box;
    spMp4SampleSizeBox *src_stz2 = (spMp4SampleSizeBox *)src_box;

    dest_stz2->field_size = src_stz2->field_size;
    total_ncopy = sizeof(dest_stz2->field_size);

    dest_stz2->entry_count = src_stz2->entry_count;
    total_ncopy += sizeof(dest_stz2->entry_count);

    total_ncopy += sizeof(dest_stz2->entries);
    
    if (dest_stz2->entry_count > 0) {
        dest_stz2->entries = xalloc(dest_stz2->entry_count, unsigned long);
        for (i = 0; i < dest_stz2->entry_count; i++) {
            dest_stz2->entries[i] = src_stz2->entries[i];
            total_ncopy += sizeof(dest_stz2->entries[i]);
        }
    } else {
        dest_stz2->entries = NULL;
    }
        
    return total_ncopy;
}

/* 'stsz' */
spMp4BoxSize spReadMp4SampleSizeBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4SampleSizeBox *stsz = (spMp4SampleSizeBox *)box;

    if ((nread = freadulong32(&stsz->sample_size, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4SampleSizeBox", "sample_size = %ld\n", stsz->sample_size);
    
    if ((nread = freadulong32(&stsz->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4SampleSizeBox", "entry_count = %ld\n", stsz->entry_count);
    
    if (stsz->sample_size == 0 && stsz->entry_count > 0) {
        stsz->entries = xalloc(stsz->entry_count, unsigned long);
        for (i = 0; i < stsz->entry_count; i++) {
            if ((nread = freadulong32(&stsz->entries[i], 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            spDebug(80, "spReadMp4SampleSizeBox", "entries[%ld] = %ld\n", i, stsz->entries[i]);
        }
    } else {
        stsz->entry_count = 0;
        stsz->entries = NULL;
    }
        
    spDebug(50, "spReadMp4SampleSizeBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stsz->header.size);
        
    return total_nread;
}

spBool spFreeMp4SampleSizeBox(void *box)
{
    spMp4SampleSizeBox *stsz = (spMp4SampleSizeBox *)box;

    if (stsz->entries != NULL) {
        xfree(stsz->entries);
        stsz->entries = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spGetMp4SampleSize(spMp4SampleSizeBox *stsz, unsigned long sample) /* 'stsz' or 'stz2' */
{
    if (stsz == NULL) return 0;
    
    if (strneq(stsz->header.type, "stz2", 4)) {
        if (sample < stsz->entry_count) {
            return stsz->entries[sample];
        }
    } else if (strneq(stsz->header.type, "stsz", 4)) {
        if (stsz->entry_count <= 0) {
            return stsz->sample_size;
        } else if (sample < stsz->entry_count) {
            return stsz->entries[sample];
        }
    }
    
    return 0;
}

spBool spSetMp4SampleSize(spMp4SampleSizeBox *stsz, unsigned long sample, unsigned long size)
{
    long size_in_file;
    
    if (stsz == NULL || sample > stsz->entry_count) return SP_FALSE;

    stsz->sample_size = 0;
    
    spDebug(50, "spSetMp4SampleSize", "sample = %lu, size = %lu\n", sample, size);
    
    if (sample == stsz->entry_count) {
        if (strneq("stz2", stsz->header.type, 4)) {
            if (stsz->field_size == 4) {
                if (sample % 2 == 1) {
                    size_in_file = 1;
                } else {
                    size_in_file = 0;
                }
            } else {
                size_in_file = stsz->field_size / 8;
            }
        } else {
            size_in_file = 4;
        }
        incrementEntryCount((spMp4Box *)stsz, size_in_file, (char **)&stsz->entries, sizeof(unsigned long), &stsz->alloc_count, &stsz->entry_count);
        spDebug(50, "spSetMp4SampleSize", "entry_count incremented: entry_count = %ld, size = %lu\n",
                stsz->entry_count, size);
    }
    
    stsz->entries[sample] = size;

    return SP_TRUE;
}

unsigned long spAppendMp4SampleSize(spMp4SampleSizeBox *stsz, unsigned long size)
{
    if (stsz == NULL) return 0;

    spSetMp4SampleSize(stsz, stsz->entry_count, size);

    return stsz->entry_count;
}

spMp4BoxSize spWriteMp4SampleSizeBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SampleSizeBox *stsz = (spMp4SampleSizeBox *)box;

    spDebug(50, "spWriteMp4SampleSizeBox", "sample_size = %lu\n", stsz->sample_size);
    if ((nwrite = fwriteulong32(&stsz->sample_size, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    spDebug(50, "spWriteMp4SampleSizeBox", "entry_count = %ld\n", stsz->entry_count);
    if ((nwrite = fwriteulong32(&stsz->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if (stsz->sample_size == 0 && stsz->entry_count > 0) {
        for (i = 0; i < stsz->entry_count; i++) {
            spDebug(80, "spWriteMp4SampleSizeBox", "entries[%ld] = %ld\n", i, stsz->entries[i]);
            if ((nwrite = fwriteulong32(&stsz->entries[i], 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
        }
    }
        
    spDebug(50, "spWriteMp4SampleSizeBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stsz->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SampleSizeBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4SampleSizeBox *dest_stsz = (spMp4SampleSizeBox *)dest_box;
    spMp4SampleSizeBox *src_stsz = (spMp4SampleSizeBox *)src_box;

    dest_stsz->sample_size = src_stsz->sample_size;
    total_ncopy = sizeof(dest_stsz->sample_size);

    dest_stsz->entry_count = src_stsz->entry_count;
    total_ncopy += sizeof(dest_stsz->entry_count);
    
    total_ncopy += sizeof(dest_stsz->entries);
    
    if (dest_stsz->sample_size == 0 && dest_stsz->entry_count > 0) {
        dest_stsz->entries = xalloc(dest_stsz->entry_count, unsigned long);
        for (i = 0; i < dest_stsz->entry_count; i++) {
            dest_stsz->entries[i] = src_stsz->entries[i];
            total_ncopy += sizeof(dest_stsz->entries[i]);
        }
    } else {
        dest_stsz->entry_count = 0;
        dest_stsz->entries = NULL;
    }
        
    return total_ncopy;
}

/* 'stsc' */
spMp4BoxSize spReadMp4SampleToChunkBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4SampleToChunkBox *stsc = (spMp4SampleToChunkBox *)box;

    if ((nread = freadulong32(&stsc->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4SampleToChunkBox", "entry_count = %ld\n", stsc->entry_count);
    
    if (stsc->entry_count > 0) {
        stsc->alloc_count = (1 + stsc->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stsc->entries = xalloc(stsc->alloc_count, spMp4SampleToChunkEntry);
        for (i = 0; i < stsc->entry_count; i++) {
            if ((nread = freadulong32(&stsc->entries[i].first_chunk, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            if ((nread = freadulong32(&stsc->entries[i].samples_per_chunk, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            if ((nread = freadulong32(&stsc->entries[i].sample_description_index, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            spDebug(80, "spReadMp4SampleToChunkBox",
                    "%ld: first_chunk = %ld, samples_per_chunk = %ld, sample_description_index = %ld\n",
                    i, stsc->entries[i].first_chunk, stsc->entries[i].samples_per_chunk, stsc->entries[i].sample_description_index);
        }
    } else {
        stsc->alloc_count = 0;
        stsc->entries = NULL;
    }

    spDebug(50, "spReadMp4SampleToChunkBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stsc->header.size);
        
    return total_nread;
}

spBool spFreeMp4SampleToChunkBox(void *box)
{
    spMp4SampleToChunkBox *stsc = (spMp4SampleToChunkBox *)box;

    if (stsc->entries != NULL) {
        xfree(stsc->entries);
        stsc->entries = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spAppendMp4SampleToChunk(spMp4SampleToChunkBox *stsc,
                                       unsigned long chunk, unsigned long samples_per_chunk, unsigned long sample_description_index)
{
    unsigned long i;
    
    if (stsc == NULL) return 0;

    if (stsc->entry_count > 0) {
        i = stsc->entry_count - 1;
        if (stsc->entries[i].first_chunk >= chunk) {
            spDebug(10, "spAppendMp4SampleToChunk", "error: stsc->entries[%ld].first_chunk (%ld) >= chunk (%ld)\n",
                    i, stsc->entries[i].first_chunk, chunk);
            return 0;
        }

        if (stsc->entries[i].samples_per_chunk == samples_per_chunk
            && stsc->entries[i].sample_description_index == sample_description_index) {
            spDebug(10, "spAppendMp4SampleToChunk",
                    "same samples_per_chunk (%ld): stsc->entries[%ld].first_chunk (%ld) == chunk (%ld)\n",
                    samples_per_chunk, i, stsc->entries[i].first_chunk, chunk);
            return stsc->entry_count;
        }
        
    }

    i = stsc->entry_count;
    incrementEntryCount((spMp4Box *)stsc, 12, (char **)&stsc->entries, sizeof(spMp4SampleToChunkEntry), &stsc->alloc_count, &stsc->entry_count);
    stsc->entries[i].first_chunk = chunk;
    stsc->entries[i].samples_per_chunk = samples_per_chunk;
    stsc->entries[i].sample_description_index = sample_description_index;
    
    spDebug(10, "spAppendMp4SampleToChunk", "entry_count incremented: entry_count = %ld, chunk = %ld, samples_per_chunk = %ld\n",
            stsc->entry_count, chunk, samples_per_chunk);
    
    return stsc->entry_count;
}

spMp4BoxSize spWriteMp4SampleToChunkBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SampleToChunkBox *stsc = (spMp4SampleToChunkBox *)box;

    spDebug(50, "spWriteMp4SampleToChunkBox", "entry_count = %ld\n", stsc->entry_count);
    
    if ((nwrite = fwriteulong32(&stsc->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (stsc->entry_count > 0) {
        for (i = 0; i < stsc->entry_count; i++) {
            spDebug(80, "spWriteMp4SampleToChunkBox",
                    "%ld: first_chunk = %ld, samples_per_chunk = %ld, sample_description_index = %ld\n",
                    i, stsc->entries[i].first_chunk, stsc->entries[i].samples_per_chunk, stsc->entries[i].sample_description_index);
            
            if ((nwrite = fwriteulong32(&stsc->entries[i].first_chunk, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
            
            if ((nwrite = fwriteulong32(&stsc->entries[i].samples_per_chunk, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
            
            if ((nwrite = fwriteulong32(&stsc->entries[i].sample_description_index, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
        }
    }

    spDebug(50, "spWriteMp4SampleToChunkBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stsc->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SampleToChunkBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4SampleToChunkBox *dest_stsc = (spMp4SampleToChunkBox *)dest_box;
    spMp4SampleToChunkBox *src_stsc = (spMp4SampleToChunkBox *)src_box;

    dest_stsc->entry_count = src_stsc->entry_count;
    total_ncopy = sizeof(dest_stsc->entry_count);
    
    total_ncopy += sizeof(dest_stsc->alloc_count) + sizeof(dest_stsc->entries);
    
    if (dest_stsc->entry_count > 0) {
        dest_stsc->alloc_count = src_stsc->alloc_count;
        dest_stsc->entries = xalloc(src_stsc->alloc_count, spMp4SampleToChunkEntry);
        for (i = 0; i < dest_stsc->entry_count; i++) {
            dest_stsc->entries[i].first_chunk = src_stsc->entries[i].first_chunk;
            total_ncopy += sizeof(dest_stsc->entries[i].first_chunk);

            dest_stsc->entries[i].samples_per_chunk = src_stsc->entries[i].samples_per_chunk;
            total_ncopy += sizeof(dest_stsc->entries[i].samples_per_chunk);

            dest_stsc->entries[i].sample_description_index = src_stsc->entries[i].sample_description_index;
            total_ncopy += sizeof(dest_stsc->entries[i].sample_description_index);
        }
    } else {
        dest_stsc->alloc_count = 0;
        dest_stsc->entries = NULL;
    }

    return total_ncopy;
}

/* 'ctts' */
spMp4BoxSize spReadMp4CompositionOffsetBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4CompositionOffsetBox *ctts = (spMp4CompositionOffsetBox *)box;

    if ((nread = freadulong32(&ctts->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4CompositionOffsetBox", "entry_count = %ld\n", ctts->entry_count);
    
    if (ctts->entry_count > 0) {
        ctts->alloc_count = (1 + ctts->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        ctts->entries = xalloc(ctts->alloc_count, spMp4CompositionOffsetEntry);
        for (i = 0; i < ctts->entry_count; i++) {
            if ((nread = freadulong32(&ctts->entries[i].sample_count, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            if ((nread = freadulong32(&ctts->entries[i].sample_offset, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            spDebug(80, "spReadMp4CompositionOffsetBox",
                    "%ld: samples_count = %lu, sample_offset = %lu\n",
                    i, ctts->entries[i].sample_count, ctts->entries[i].sample_offset);
        }
    } else {
        ctts->alloc_count = 0;
        ctts->entries = NULL;
    }
        
    spDebug(50, "spReadMp4CompositionOffsetBox", "total_nread = %ld / %ld\n",
            (long)total_nread, ctts->header.size);
    
    return total_nread;
}

spBool spFreeMp4CompositionOffsetBox(void *box)
{
    spMp4CompositionOffsetBox *ctts = (spMp4CompositionOffsetBox *)box;

    if (ctts->entries != NULL) {
        xfree(ctts->entries);
        ctts->entries = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spWriteMp4CompositionOffsetBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4CompositionOffsetBox *ctts = (spMp4CompositionOffsetBox *)box;

    spDebug(50, "spWriteMp4CompositionOffsetBox", "entry_count = %ld\n", ctts->entry_count);
    
    if ((nwrite = fwriteulong32(&ctts->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (ctts->entry_count > 0) {
        for (i = 0; i < ctts->entry_count; i++) {
            spDebug(80, "spWriteMp4CompositionOffsetBox",
                    "%ld: samples_count = %lu, sample_offset = %lu\n",
                    i, ctts->entries[i].sample_count, ctts->entries[i].sample_offset);
            
            if ((nwrite = fwriteulong32(&ctts->entries[i].sample_count, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
            
            if ((nwrite = fwriteulong32(&ctts->entries[i].sample_offset, 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 4;
        }
    }
        
    spDebug(50, "spWriteMp4CompositionOffsetBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, ctts->header.size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4CompositionOffsetBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4CompositionOffsetBox *dest_ctts = (spMp4CompositionOffsetBox *)dest_box;
    spMp4CompositionOffsetBox *src_ctts = (spMp4CompositionOffsetBox *)src_box;

    dest_ctts->entry_count = src_ctts->entry_count;
    total_ncopy = sizeof(dest_ctts->entry_count);
    
    total_ncopy += sizeof(dest_ctts->alloc_count) + sizeof(dest_ctts->entries);
    
    if (dest_ctts->entry_count > 0) {
        dest_ctts->alloc_count = src_ctts->alloc_count;
        dest_ctts->entries = xalloc(dest_ctts->alloc_count, spMp4CompositionOffsetEntry);
        for (i = 0; i < dest_ctts->entry_count; i++) {
            dest_ctts->entries[i].sample_count = src_ctts->entries[i].sample_count;
            total_ncopy += sizeof(dest_ctts->entries[i].sample_count);

            dest_ctts->entries[i].sample_offset = src_ctts->entries[i].sample_offset;
            total_ncopy += sizeof(dest_ctts->entries[i].sample_offset);
        }
    } else {
        dest_ctts->alloc_count = 0;
        dest_ctts->entries = NULL;
    }
        
    return total_ncopy;
}

/* 'stts' */
spMp4BoxSize spReadMp4TimeToSampleBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4TimeToSampleBox *stts = (spMp4TimeToSampleBox *)box;

    stts->total_sample = 0;
#ifdef SP_SUPPORT_64BIT_INT
    stts->total_duration = 0;
#else
    memset(stts->total_duration, 0, sizeof(stts->total_duration));
#endif

    if ((nread = freadulong32(&stts->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(50, "spReadMp4TimeToSampleBox", "entry_count = %ld\n", stts->entry_count);
    
    if (stts->entry_count > 0) {
        stts->alloc_count = (1 + stts->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stts->entries = xalloc(stts->alloc_count, spMp4TimeToSampleEntry);
        for (i = 0; i < stts->entry_count; i++) {
            if ((nread = freadulong32(&stts->entries[i].sample_count, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;
            
            if ((nread = freadulong32(&stts->entries[i].sample_delta, 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 4;

            spDebug(80, "spReadMp4TimeToSampleBox",
                    "%ld: samples_count = %lu, sample_delta = %lu\n",
                    i, stts->entries[i].sample_count, stts->entries[i].sample_delta);
            
            stts->total_sample += stts->entries[i].sample_count;
#ifdef SP_SUPPORT_64BIT_INT
            stts->total_duration += (spLong64)stts->entries[i].sample_count * (spLong64)stts->entries[i].sample_delta;
            spDebug(80, "spReadMp4TimeToSampleBox", "%ld: total_duration = %lu\n", i, (unsigned long)stts->total_duration);
#else
            spAddC64(stts->total_duration, stts->entries[i].sample_count * stts->entries[i].sample_delta, 0);
#endif
        }
    } else {
        stts->alloc_count = 0;
        stts->entries = NULL;
    }
        
    spDebug(50, "spReadMp4TimeToSampleBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stts->header.size);
    
    return total_nread;
}

spBool spFreeMp4TimeToSampleBox(void *box)
{
    spMp4TimeToSampleBox *stts = (spMp4TimeToSampleBox *)box;

    if (stts->entries != NULL) {
        xfree(stts->entries);
        stts->entries = NULL;
    }
    
    return SP_TRUE;
}

unsigned long spGetMp4SampleDuration(spMp4TimeToSampleBox *stts, unsigned long sample)
{
    unsigned long i;
    unsigned long cum_sample, next_cum_sample;
    
    if (stts == NULL || stts->entry_count <= 0) return 0;

    cum_sample = 0;
        
    for (i = 0; i < stts->entry_count; i++) {
        next_cum_sample = cum_sample + stts->entries[i].sample_count;
        
        if (next_cum_sample > sample) {
            return stts->entries[i].sample_delta;
        }

        cum_sample = next_cum_sample;
    }

    return 0;
}

unsigned long spGetMp4MaxSampleDuration(spMp4TimeToSampleBox *stts)
{
    unsigned long i;
    unsigned long max_duration;
    
    if (stts == NULL || stts->entry_count <= 0) return 0;

    for (i = 0, max_duration = 0; i < stts->entry_count; i++) {
        max_duration = MAX(stts->entries[i].sample_delta, max_duration);
    }

    return max_duration;
}

spBool spUpdateMp4TrackDuration(spMp4TrackBox *trak, spMp4TimeToSampleBox *stts)
{
    spMp4MovieHeaderBox *mvhd;
    unsigned long timescale;
    double mdhd_duration;
    double mvhd_duration;
    double mvhd_timescale;
#ifdef SP_SUPPORT_64BIT_INT
    spLong64 duration;
#else
    unsigned long duration;
#endif

#ifdef SP_SUPPORT_64BIT_INT
    duration = stts->total_duration;
#else
    duration = spC64ToULong(stts->total_duration, 0);
#endif
    spDebug(80, "spUpdateMp4TrackDuration", "duration = %lu\n", (unsigned long)duration);


    if (trak->tkhd->header.version == 1) {
#ifdef SP_SUPPORT_64BIT_INT
        spLong64ToC64(duration, trak->tkhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#else
        spULongToC64(duration, trak->tkhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#endif
    } else {
        trak->tkhd->u.version0.duration = (unsigned long)duration;
    }
    
    if (trak->mdia->mdhd->header.version == 1) {
#ifdef SP_SUPPORT_64BIT_INT
        spLong64ToC64(duration, trak->mdia->mdhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#else
        spULongToC64(duration, trak->mdia->mdhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#endif
        timescale = trak->mdia->mdhd->u.version1.timescale;
    } else {
        trak->mdia->mdhd->u.version0.duration = (unsigned long)duration;
        timescale = trak->mdia->mdhd->u.version0.timescale;
    }
    mdhd_duration = (double)duration / (double)timescale;
    spDebug(80, "spUpdateMp4TrackDuration", "mdhd_duration = %f, timescale = %lu\n", mdhd_duration, timescale);

    mvhd = trak->parent->mvhd;

    if (mvhd->header.version == 1) {
#ifdef SP_SUPPORT_64BIT_INT
        mvhd_duration = (double)spC64ToLong64(mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#else
        mvhd_duration = (double)spC64ToULong(mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#endif
        mvhd_timescale = (double)mvhd->u.version1.timescale;
    } else {
        mvhd_duration = (double)mvhd->u.version0.duration;
        mvhd_timescale = (double)mvhd->u.version0.timescale;
    }
    mvhd_duration /= mvhd_timescale;
    spDebug(80, "spUpdateMp4TrackDuration", "mvhd_duration = %f, mvhd_timescale = %f\n",
            mvhd_duration, mvhd_timescale);

    if (mdhd_duration > mvhd_duration) {
        mvhd_duration = mdhd_duration * mvhd_timescale; /* convert mdhd duration to mvhd duration */
        spDebug(80, "spUpdateMp4TrackDuration", "mvhd_duration updated: %f\n", mvhd_duration);
        
        if (mvhd->header.version == 1) {
#ifdef SP_SUPPORT_64BIT_INT
            spLong64ToC64((spLong64)spRound(mvhd_duration), mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#else
            spULongToC64((unsigned long)spRound(mvhd_duration), mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#endif
        } else {
            mvhd->u.version0.duration = (unsigned long)spRound(mvhd_duration);
        }
    }

    return SP_TRUE;
}

unsigned long spAppendMp4SampleDuration(spMp4TimeToSampleBox *stts, unsigned long sample, unsigned long sample_delta)
{
    unsigned long i;
    unsigned long cum_sample, next_cum_sample;
    spMp4TrackBox *trak;
    
    if (stts == NULL) return 0;

    spDebug(50, "spAppendMp4SampleDuration", "sample = %lu, sample_delta = %lu, entry_count = %lu\n",
            sample, sample_delta, stts->entry_count);

    /* stts-stbl-minf-mdia-trak */
    trak = (spMp4TrackBox *)stts->parent->parent->parent->parent;
    
    cum_sample = 0;
        
    for (i = 0; i < stts->entry_count; i++) {
        next_cum_sample = cum_sample + stts->entries[i].sample_count;
        spDebug(100, "spAppendMp4SampleDuration",
                "stts->entries[%ld].sample_count = %lu, cum_sample = %lu, next_cum_sample = %lu\n",
                i, stts->entries[i].sample_count, cum_sample, next_cum_sample);
        
        if (next_cum_sample >= sample) {
            stts->entries[i].sample_count = MIN(stts->entries[i].sample_count, sample - cum_sample);
            
            if (sample_delta == stts->entries[i].sample_delta) {
                stts->entries[i].sample_count++;
                spDebug(50, "spAppendMp4SampleDuration", "sample_count incremented: stts->entries[%ld].sample_count = %lu\n",
                        i, stts->entries[i].sample_count);
                
                stts->total_sample++;
#ifdef SP_SUPPORT_64BIT_INT
                stts->total_duration += (spLong64)stts->entries[i].sample_delta;
                spDebug(80, "spAppendMp4SampleDuration", "%ld: total_duration = %lu\n", i, (unsigned long)stts->total_duration);
#else
                spAddC64(stts->total_duration, stts->entries[i].sample_delta, 0);
#endif
                spUpdateMp4TrackDuration(trak, stts);
            
                return i;
            }
            break;
        }

        cum_sample = next_cum_sample;
    }

    i = stts->entry_count;
    incrementEntryCount((spMp4Box *)stts, 8, (char **)&stts->entries, sizeof(spMp4TimeToSampleEntry), &stts->alloc_count, &stts->entry_count);
    stts->entries[i].sample_count = 1;
    stts->entries[i].sample_delta = sample_delta;
    
    stts->total_sample++;
#ifdef SP_SUPPORT_64BIT_INT
    stts->total_duration += (spLong64)stts->entries[i].sample_delta;
    spDebug(80, "spAppendMp4SampleDuration", "%ld: total_duration = %lu\n", i, (unsigned long)stts->total_duration);
#else
    spAddC64(stts->total_duration, stts->entries[i].sample_delta, 0);
#endif
    spUpdateMp4TrackDuration(trak, stts);
    
    spDebug(50, "spAppendMp4SampleDuration", "entry_count incremented: entry_count = %lu, sample_delta = %lu\n",
            stts->entry_count, stts->entries[i].sample_delta);
    
    return stts->entry_count;
}

spMp4BoxSize spWriteMp4TimeToSampleBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4TimeToSampleBox *stts = (spMp4TimeToSampleBox *)box;

    spDebug(50, "spWriteMp4TimeToSampleBox", "entry_count = %lu, alloc_count = %lu\n", stts->entry_count, stts->alloc_count);
    if ((nwrite = fwriteulong32(&stts->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (stts->entry_count > 0) {
        for (i = 0; i < stts->entry_count; i++) {
            spDebug(80, "spWriteMp4TimeToSampleBox",
                    "%ld: samples_count = %lu, sample_delta = %lu\n",
                    i, stts->entries[i].sample_count, stts->entries[i].sample_delta);
            
            if ((nwrite = fwriteulong32(&stts->entries[i].sample_count, 1, swap, fp)) != 1) {
                spDebug(50, "spWriteMp4TimeToSampleBox", "fwriteulong32 for sample_count failed: %ld\n", nwrite);
                return nwrite;
            }
            total_nwrite += 4;
            spDebug(80, "spWriteMp4TimeToSampleBox", "fwriteulong32 for sample_count done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
            
            if ((nwrite = fwriteulong32(&stts->entries[i].sample_delta, 1, swap, fp)) != 1) {
                spDebug(50, "spWriteMp4TimeToSampleBox", "fwriteulong32 for sample_delta failed: %ld\n", nwrite);
                return nwrite;
            }
            total_nwrite += 4;
            spDebug(80, "spWriteMp4TimeToSampleBox", "fwriteulong32 for sample_delta done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
        }
    }
        
    spDebug(50, "spWriteMp4TimeToSampleBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stts->header.size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4TimeToSampleBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4TimeToSampleBox *dest_stts = (spMp4TimeToSampleBox *)dest_box;
    spMp4TimeToSampleBox *src_stts = (spMp4TimeToSampleBox *)src_box;

    dest_stts->total_sample = src_stts->total_sample;
    total_ncopy = sizeof(dest_stts->total_sample);
#ifdef SP_SUPPORT_64BIT_INT
    dest_stts->total_duration = src_stts->total_duration;
#else
    memcpy(dest_stts->total_duration, src_stts->total_duration, sizeof(src_stts->total_duration));
#endif
    total_ncopy += sizeof(dest_stts->total_duration);

    dest_stts->entry_count = src_stts->entry_count;
    total_ncopy += sizeof(dest_stts->entry_count);

    total_ncopy += sizeof(dest_stts->alloc_count) + sizeof(dest_stts->entries);
    
    if (dest_stts->entry_count > 0) {
        dest_stts->alloc_count = src_stts->alloc_count;
        dest_stts->entries = xalloc(dest_stts->alloc_count, spMp4TimeToSampleEntry);
        for (i = 0; i < dest_stts->entry_count; i++) {
            dest_stts->entries[i].sample_count = src_stts->entries[i].sample_count;
            total_ncopy += sizeof(dest_stts->entries[i].sample_count);

            dest_stts->entries[i].sample_delta = src_stts->entries[i].sample_delta;
            total_ncopy += sizeof(dest_stts->entries[i].sample_delta);
        }
    } else {
        dest_stts->alloc_count = 0;
        dest_stts->entries = NULL;
    }
        
    return total_ncopy;
}

/* 'stdp' */
spMp4BoxSize spReadMp4DegradationPriorityBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4SampleSizeBox *stsz;
    spMp4DegradationPriorityBox *stdp = (spMp4DegradationPriorityBox *)box;

    if ((stsz = (spMp4SampleSizeBox *)spFindMp4Box((spMp4Box *)parent, "stsz", "stbl")) == NULL) {
        if ((stsz = (spMp4SampleSizeBox *)spFindMp4Box((spMp4Box *)parent, "stz2", "stbl")) == NULL) {
        }
    }

    if (stsz != NULL) {
        stdp->entry_count = stsz->entry_count;
    } else {
        stdp->entry_count = (unsigned long)(spGetMp4BoxHeaderContentSize(&stdp->header) / 2);
    }

    total_nread = 0;

    if (stdp->entry_count > 0) {
        stdp->priority = xalloc(stdp->entry_count, unsigned short);
        for (i = 0; i < stdp->entry_count; i++) {
            if ((nread = freadshort((short *)&stdp->priority[i], 1, swap, fp)) != 1) {
                return nread;
            }
            total_nread += 2;
            
            spDebug(80, "spReadMp4DegradationPriorityBox", "priority[%ld] = %d\n", i, stdp->priority[i]);
        }
    } else {
        stdp->entry_count = 0;
        stdp->priority = NULL;
    }
        
    spDebug(50, "spReadMp4DegradationPriorityBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stdp->header.size);
        
    return total_nread;
}

spBool spFreeMp4DegradationPriorityBox(void *box)
{
    spMp4DegradationPriorityBox *stdp = (spMp4DegradationPriorityBox *)box;

    if (stdp->priority != NULL) {
        xfree(stdp->priority);
        stdp->priority = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spWriteMp4DegradationPriorityBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4DegradationPriorityBox *stdp = (spMp4DegradationPriorityBox *)box;

    total_nwrite = 0;
    
    if (stdp->entry_count > 0) {
        for (i = 0; i < stdp->entry_count; i++) {
            spDebug(80, "spWriteMp4DegradationPriorityBox", "priority[%ld] = %d\n", i, stdp->priority[i]);
            
            if ((nwrite = fwriteshort((short *)&stdp->priority[i], 1, swap, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 2;
        }
    }
        
    spDebug(50, "spWriteMp4DegradationPriorityBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stdp->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4DegradationPriorityBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4DegradationPriorityBox *dest_stdp = (spMp4DegradationPriorityBox *)dest_box;
    spMp4DegradationPriorityBox *src_stdp = (spMp4DegradationPriorityBox *)src_box;

    dest_stdp->entry_count = src_stdp->entry_count;
    total_ncopy = sizeof(dest_stdp->entry_count);

    total_ncopy += sizeof(dest_stdp->priority);
    
    if (dest_stdp->entry_count > 0) {
        dest_stdp->priority = xalloc(dest_stdp->entry_count, unsigned short);
        for (i = 0; i < dest_stdp->entry_count; i++) {
            dest_stdp->priority[i] = src_stdp->priority[i];
            total_ncopy += sizeof(dest_stdp->priority[i]);
        }
    } else {
        dest_stdp->entry_count = 0;
        dest_stdp->priority = NULL;
    }
        
    return total_ncopy;
}

/* 'sdtp' */
spMp4BoxSize spReadMp4SampleDependencyTypeBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    long nread;
    spMp4BoxSize total_nread;
    spMp4SampleSizeBox *stsz;
    spMp4SampleDependencyTypeBox *sdtp = (spMp4SampleDependencyTypeBox *)box;

    if ((stsz = (spMp4SampleSizeBox *)spFindMp4Box((spMp4Box *)parent, "stsz", "stbl")) == NULL) {
        if ((stsz = (spMp4SampleSizeBox *)spFindMp4Box((spMp4Box *)parent, "stz2", "stbl")) == NULL) {
            spDebug(80, "spReadMp4SampleDependencyTypeBox", "cannot find stsz box\n");
        }
    }

    if (stsz != NULL) {
        sdtp->entry_count = stsz->entry_count;
    } else {
        sdtp->entry_count = (unsigned long)spGetMp4BoxHeaderContentSize(&sdtp->header);
    }

    total_nread = 0;

    if (sdtp->entry_count > 0) {
        sdtp->entries = xalloc(sdtp->entry_count, unsigned char);
        for (i = 0; i < sdtp->entry_count; i++) {
            if ((nread = (long)fread(&sdtp->entries[i], 1, 1, fp)) != 1) {
                return nread;
            }
            total_nread += 1;
            
            spDebug(80, "spReadMp4SampleDependencyTypeBox", "entries[%ld] = %x\n", i, sdtp->entries[i]);
        }
    } else {
        sdtp->entry_count = 0;
        sdtp->entries = NULL;
    }
        
    spDebug(50, "spReadMp4SampleDependencyTypeBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, sdtp->header.size);
        
    return total_nread;
}

spBool spFreeMp4SampleDependencyTypeBox(void *box)
{
    spMp4SampleDependencyTypeBox *sdtp = (spMp4SampleDependencyTypeBox *)box;

    if (sdtp->entries != NULL) {
        xfree(sdtp->entries);
        sdtp->entries = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spWriteMp4SampleDependencyTypeBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SampleDependencyTypeBox *sdtp = (spMp4SampleDependencyTypeBox *)box;

    total_nwrite = 0;
    
    if (sdtp->entry_count > 0) {
        for (i = 0; i < sdtp->entry_count; i++) {
            spDebug(80, "spWriteMp4SampleDependencyTypeBox", "entries[%ld] = %d\n", i, sdtp->entries[i]);
            
            if ((nwrite = (long)fwrite(&sdtp->entries[i], 1, 1, fp)) != 1) {
                return nwrite;
            }
            total_nwrite += 1;
        }
    }
        
    spDebug(50, "spWriteMp4SampleDependencyTypeBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, sdtp->header.size);
        
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SampleDependencyTypeBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize total_ncopy;
    spMp4SampleDependencyTypeBox *dest_sdtp = (spMp4SampleDependencyTypeBox *)dest_box;
    spMp4SampleDependencyTypeBox *src_sdtp = (spMp4SampleDependencyTypeBox *)src_box;

    dest_sdtp->entry_count = src_sdtp->entry_count;
    total_ncopy = sizeof(dest_sdtp->entry_count);

    total_ncopy += sizeof(dest_sdtp->entries);
    
    if (dest_sdtp->entry_count > 0) {
        dest_sdtp->entries = xalloc(dest_sdtp->entry_count, unsigned char);
        for (i = 0; i < dest_sdtp->entry_count; i++) {
            dest_sdtp->entries[i] = src_sdtp->entries[i];
            total_ncopy += sizeof(dest_sdtp->entries[i]);
        }
    } else {
        dest_sdtp->entry_count = 0;
        dest_sdtp->entries = NULL;
    }
        
    return total_ncopy;
}

/* 'clap' */
spMp4BoxSize spReadMp4CleanApertureBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4CleanApertureBox *clap = (spMp4CleanApertureBox *)box;
    
    if ((nread = freadulong32(&clap->cleanApertureWidthN, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
        
    if ((nread = freadulong32(&clap->cleanApertureWidthD, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->cleanApertureHeightN, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->cleanApertureHeightD, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->horizOffN, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->horizOffD, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->vertOffN, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&clap->vertOffD, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    spDebug(50, "spReadMp4CleanApertureBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, clap->header.size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4CleanApertureBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4CleanApertureBox *clap = (spMp4CleanApertureBox *)box;
    
    if ((nwrite = fwriteulong32(&clap->cleanApertureWidthN, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
        
    if ((nwrite = fwriteulong32(&clap->cleanApertureWidthD, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->cleanApertureHeightN, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->cleanApertureHeightD, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->horizOffN, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->horizOffD, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->vertOffN, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&clap->vertOffD, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    spDebug(50, "spWriteMp4CleanApertureBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, clap->header.size);
    
    return total_nwrite;
}

/* 'colr' */
spMp4BoxSize spReadMp4ColorParameterBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize data_size;
    spMp4BoxSize total_nread;
    spMp4ColorParameterBox *colr = (spMp4ColorParameterBox *)box;

    if ((nread = (long)fread(colr->type, 1, 4, fp)) != 4) {
        return nread;
    }
    spDebug(50, "spReadMp4ColorParameterBox", "type = %c%c%c%c\n", colr->type[0], colr->type[1], colr->type[2], colr->type[3]);
    total_nread = 4;

    if (!strneq(colr->type, "nclc", 4)) {
        /* unknown color parameter type */
        data_size = colr->header.size - 12;
        colr->data = xalloc((long)data_size + 1, char);
        if ((nread = (long)fread(colr->data, 1, (long)data_size, fp)) != data_size) {
            xfree(colr->data);
            colr->data = NULL;
            return nread;
        }
        colr->data[data_size] = NUL;
        spDebug(50, "spReadMp4ColorParameterBox", "data = %s\n", colr->data);
        total_nread += nread;
    } else {
        colr->data= NULL;
        
        if ((nread = freadshort(&colr->primaries_index, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 2;
    
        if ((nread = freadshort(&colr->transfer_function_index, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 2;
    
        if ((nread = freadshort(&colr->matrix_index, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 2;
    }
    
    spDebug(50, "spReadMp4ColorParameterBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, colr->header.size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4ColorParameterBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize data_size;
    spMp4BoxSize total_nwrite;
    spMp4ColorParameterBox *colr = (spMp4ColorParameterBox *)box;

    if ((nwrite = (long)fwrite(colr->type, 1, 4, fp)) != 4) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if (!strneq(colr->type, "nclc", 4)) {
        if (colr->data != NULL) {
            /* unknown color parameter type */
            data_size = colr->header.size - 12;
            if ((nwrite = (long)fwrite(colr->data, 1, (long)data_size, fp)) != data_size) {
                return nwrite;
            }
            total_nwrite += nwrite;
        }
    } else {
        if ((nwrite = fwriteshort(&colr->primaries_index, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 2;
    
        if ((nwrite = fwriteshort(&colr->transfer_function_index, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 2;
    
        if ((nwrite = fwriteshort(&colr->matrix_index, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 2;
    }
    
    spDebug(50, "spWriteMp4ColorParameterBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, colr->header.size);
    
    return total_nwrite;
}

/* 'stsl' */
spMp4BoxSize spReadMp4SampleScaleBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4SampleScaleBox *stsl = (spMp4SampleScaleBox *)box;
    
    if ((nread = (long)fread(&stsl->flag, 1, 1, fp)) != 1) {
        return nread;
    }
    total_nread = 1;
    
    if ((nread = (long)fread(&stsl->scale_method, 1, 1, fp)) != 1) {
        return nread;
    }
    total_nread += 1;

    if ((nread = freadshort(&stsl->display_center_x, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;

    if ((nread = freadshort(&stsl->display_center_y, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;

    spDebug(50, "spReadMp4SampleScaleBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, stsl->header.size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4SampleScaleBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SampleScaleBox *stsl = (spMp4SampleScaleBox *)box;
    
    if ((nwrite = (long)fwrite(&stsl->flag, 1, 1, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 1;
    
    if ((nwrite = (long)fwrite(&stsl->scale_method, 1, 1, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 1;

    if ((nwrite = fwriteshort(&stsl->display_center_x, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = fwriteshort(&stsl->display_center_y, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    spDebug(50, "spWriteMp4SampleScaleBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, stsl->header.size);
    
    return total_nwrite;
}

/* 'pasp' */
spMp4BoxSize spReadMp4PixelAspectRatioBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4PixelAspectRatioBox *pasp = (spMp4PixelAspectRatioBox *)box;
    
    if ((nread = freadulong32(&pasp->hSpacing, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;

    if ((nread = freadulong32(&pasp->vSpacing, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    spDebug(50, "spReadMp4PixelAspectRatioBox", "total_nread = %lu / %lu\n",
            (unsigned long)total_nread, pasp->header.size);
    
    return total_nread;
}
        
spMp4BoxSize spWriteMp4PixelAspectRatioBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4PixelAspectRatioBox *pasp = (spMp4PixelAspectRatioBox *)box;
    
    if ((nwrite = fwriteulong32(&pasp->hSpacing, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;

    if ((nwrite = fwriteulong32(&pasp->vSpacing, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    spDebug(50, "spWriteMp4PixelAspectRatioBox", "total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, pasp->header.size);
    
    return total_nwrite;
}

spMp4BoxSize spReadMp4VisualSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, int swap, FILE *fp)
{
    spMp4BoxSize nread;
    spMp4BoxSize total_nread;
    spMp4VisualSampleEntry *vide;

    spDebug(50, "spReadMp4VisualSampleEntry", "in: remain_size = %lu, depth = %ld\n",
            (unsigned long)remain_size, depth);
    
    vide = &box->u.vide;

    if ((nread = freadshort((short *)&vide->version, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;
    spDebug(50, "spReadMp4VisualSampleEntry", "version = %d\n", vide->version);

    if ((nread = freadshort((short *)&vide->revision, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4VisualSampleEntry", "revision = %d\n", vide->revision);
    
    if ((nread = freadulong32(&vide->vendor, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    if ((nread = freadulong32(&vide->temporal_quality, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    if ((nread = freadulong32(&vide->spatial_quality, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4VisualSampleEntry", "vendor = %lu, temporal_quality = %lu, spatial_quality = %lu\n",
            vide->vendor, vide->temporal_quality, vide->spatial_quality);
    
    if ((nread = freadshort((short *)&vide->width, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    
    if ((nread = freadshort((short *)&vide->height, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4VisualSampleEntry", "width = %d, height = %d\n", vide->width, vide->height);

    if ((nread = freadulong32(&vide->horizresolution, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&vide->vertresolution, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4VisualSampleEntry", "horizresolution = %lu, vertresolution = %lu\n",
            vide->horizresolution, vide->vertresolution);
    
    if ((nread = freadulong32(&vide->reserved2, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadshort((short *)&vide->frame_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;

    if ((nread = (long)fread(vide->compressorname, 1, 32, fp)) != 32) {
        return nread;
    }
    total_nread += 32;
    
    if ((nread = freadshort((short *)&vide->depth, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4VisualSampleEntry", "frame_count = %d, depth = %d\n", vide->frame_count, vide->depth);
    
    if ((nread = freadshort((short *)&vide->color_table_id, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4VisualSampleEntry", "vide->color_table_id = %x\n", vide->color_table_id);
    
    remain_size -= total_nread;
    spDebug(50, "spReadMp4VisualSampleEntry", "remain_size = %lu, total_nread = %lu\n", (unsigned long)remain_size, (unsigned long)total_nread);

    if (remain_size > 0) {
        spDebug(50, "spReadMp4VisualSampleEntry", "type = %c%c%c%c, size = %lu, remain_size = %lu\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3],
                (unsigned long)box->header.size, (unsigned long)remain_size);

#if 0
        if (0 && strneq(box->header.type, "avc1", 4)) {
            int c;
            long k;

            for (k = 0; k < remain_size; k++) {
                c = fgetc(fp);
                fputc(c, stdout);
            }
            exit(0);
        }
#endif
        
        if ((nread = spReadMp4ChildBox((spChunk *)box, 0, remain_size, depth + 1, swap, fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
        remain_size -= nread;
    }
    
    spDebug(10, "spReadMp4VisualSampleEntry", "done: total_nread = %lu / %lu, remain_size = %lu\n",
            (unsigned long)total_nread, box->header.size, (unsigned long)remain_size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4VisualSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4VisualSampleEntry *vide;

    spDebug(50, "spWriteMp4VisualSampleEntry", "in: remain_size = %lu, depth = %ld\n",
            (unsigned long)remain_size, depth);
    
    vide = &box->u.vide;

    spDebug(50, "spWriteMp4VisualSampleEntry", "version = %d\n", vide->version);
    if ((nwrite = fwriteshort((short *)&vide->version, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;

    spDebug(50, "spWriteMp4VisualSampleEntry", "reserved = %d\n", vide->revision);
    if ((nwrite = fwriteshort((short *)&vide->revision, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    if ((nwrite = fwriteulong32(&vide->vendor, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    if ((nwrite = fwriteulong32(&vide->temporal_quality, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    if ((nwrite = fwriteulong32(&vide->spatial_quality, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    spDebug(50, "spWriteMp4VisualSampleEntry", "width = %d, height = %d\n", vide->width, vide->height);
    if ((nwrite = fwriteshort((short *)&vide->width, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    if ((nwrite = fwriteshort((short *)&vide->height, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    spDebug(50, "spWriteMp4VisualSampleEntry", "horizresolution = %lu, vertresolution = %lu\n",
            vide->horizresolution, vide->vertresolution);
    if ((nwrite = fwriteulong32(&vide->horizresolution, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&vide->vertresolution, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(&vide->reserved2, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    spDebug(50, "spWriteMp4VisualSampleEntry", "frame_count = %d, depth = %d\n", vide->frame_count, vide->depth);
    if ((nwrite = fwriteshort((short *)&vide->frame_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = (long)fwrite(vide->compressorname, 1, 32, fp)) != 32) {
        return nwrite;
    }
    total_nwrite += 32;
    
    if ((nwrite = fwriteshort((short *)&vide->depth, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    if ((nwrite = fwriteshort((short *)&vide->color_table_id, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    remain_size -= total_nwrite;
    spDebug(50, "spWriteMp4VisualSampleEntry", "remain_size = %lu, total_nwrite = %lu\n", (unsigned long)remain_size, (unsigned long)total_nwrite);

    if (remain_size >= 8) {
        spDebug(50, "spWriteMp4VisualSampleEntry", "type = %c%c%c%c\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
        if ((nwrite = (long)spWriteMp4ChildOnlyBox((spChunk *)box, depth + 1, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
        remain_size -= nwrite;
    }
    
    spDebug(10, "spWriteMp4VisualSampleEntry", "done: total_nwrite = %lu / %lu, remain_size = %lu\n",
            (unsigned long)total_nwrite, box->header.size, (unsigned long)remain_size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4VisualSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box, long depth)
{
    spMp4BoxSize ncopy;
    spMp4BoxSize total_ncopy;
    spMp4VisualSampleEntry *dest_vide;
    spMp4VisualSampleEntry *src_vide;

    dest_vide = &dest_box->u.vide;
    src_vide = &src_box->u.vide;
    total_ncopy = spOffsetOf(spMp4VisualSampleEntry, esds);
    memcpy(dest_vide, src_vide, (long)total_ncopy);

    spDebug(50, "spCopyMp4VisualSampleEntry", "src version = %d, dest version = %d\n",
            src_vide->version, dest_vide->version);
    
    if ((ncopy = spCopyMp4ChildBox((spChunk *)dest_box, (spChunk *)src_box, depth + 1)) > 0) {
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

long spReadMp4AudioSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, int swap, FILE *fp)
{
    long nread;
    long total_nread;
    spMp4AudioSampleEntry *soun;

    soun = &box->u.soun;
    
    spDebug(10, "spReadMp4AudioSampleEntry", "in: box type = %c%c%c%c\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    if ((nread = freadshort((short *)&soun->version, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "version = %d\n", soun->version);
    
    if ((nread = freadshort((short *)&soun->revision, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "revision = %d\n", soun->revision);
    
    if ((nread = freadulong32(&soun->vendor, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(10, "spReadMp4AudioSampleEntry", "vendor = %ld\n", soun->vendor);
    
    if ((nread = freadshort((short *)&soun->channelcount, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "channelcount = %d\n", soun->channelcount);
    
    if ((nread = freadshort((short *)&soun->samplesize, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "samplesize = %d\n", soun->samplesize);
    
    if ((nread = freadshort(&soun->compression_id, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "compression_id = %d\n", soun->compression_id);
    
    if ((nread = freadshort((short *)&soun->reserved, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4AudioSampleEntry", "reserved = %d\n", soun->reserved);

    if ((nread = freadulong32(&soun->samplerate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    soun->samples_per_packet = 0;
    soun->bytes_per_packet = 0;
    soun->bytes_per_frame = 0;
    soun->bytes_per_sample = 0;
    
    if (soun->version >= 1) {
        if ((nread = freadulong32(&soun->samples_per_packet, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 4;
        spDebug(10, "spReadMp4AudioSampleEntry", "samples_per_packet = %lu\n", soun->samples_per_packet);
        
        if ((nread = freadulong32(&soun->bytes_per_packet, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 4;
        spDebug(10, "spReadMp4AudioSampleEntry", "bytes_per_packet = %lu\n", soun->bytes_per_packet);

            if ((nread = freadulong32(&soun->bytes_per_frame, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 4;
        spDebug(10, "spReadMp4AudioSampleEntry", "bytes_per_frame = %lu\n", soun->bytes_per_frame);

            if ((nread = freadulong32(&soun->bytes_per_sample, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread += 4;
        spDebug(10, "spReadMp4AudioSampleEntry", "bytes_per_sample = %lu\n", soun->bytes_per_sample);
    }

    remain_size -= total_nread;
    spDebug(50, "spReadMp4AudioSampleEntry", "remain_size = %lu, total_nread = %lu\n", (unsigned long)remain_size, (unsigned long)total_nread);

    if (remain_size >= 8) {
        if ((nread = (long)spReadMp4ChildBox((spChunk *)box, 0, remain_size, depth + 1, swap, fp)) <= 0) {
            return nread;
        }
        spDebug(10, "spReadMp4AudioSampleEntry", "spReadMp4ChildBox result = %ld\n", nread);
        total_nread += nread;
    }

    spDebug(10, "spReadMp4AudioSampleEntry", "done: total_nread = %lu, version = %d, channelcount = %d, samplesize = %d, samplerate = %f\n",
            (unsigned long)total_nread, soun->version, soun->channelcount, soun->samplesize, (double)soun->samplerate / 65536.0);
    
    return total_nread;
}

long spWriteMp4AudioSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    long total_nwrite;
    spMp4AudioSampleEntry *soun;

    soun = &box->u.soun;

    spDebug(10, "spWriteMp4AudioSampleEntry", "in: box type = %c%c%c%c\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    if ((nwrite = fwriteshort((short *)&soun->version, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;
    spDebug(10, "spWriteMp4AudioSampleEntry", "version = %d\n", soun->version);
    
    if ((nwrite = fwriteshort((short *)&soun->revision, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    spDebug(10, "spWriteMp4AudioSampleEntry", "revision = %d\n", soun->revision);
    
    if ((nwrite = fwriteulong32(&soun->vendor, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    spDebug(10, "spWriteMp4AudioSampleEntry", "vendor = %ld\n", soun->vendor);
    
    if ((nwrite = fwriteshort((short *)&soun->channelcount, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    spDebug(10, "spWriteMp4AudioSampleEntry", "channelcount = %d\n", soun->channelcount);
    
    if ((nwrite = fwriteshort((short *)&soun->samplesize, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    spDebug(10, "spWriteMp4AudioSampleEntry", "samplesize = %d\n", soun->samplesize);
    
    if ((nwrite = fwriteshort(&soun->compression_id, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    spDebug(10, "spWriteMp4AudioSampleEntry", "compression_id = %d\n", soun->compression_id);
    
    if ((nwrite = fwriteshort((short *)&soun->reserved, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = fwriteulong32(&soun->samplerate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if (soun->version >= 1) {
        if ((nwrite = fwriteulong32(&soun->samples_per_packet, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 4;
        spDebug(10, "spWriteMp4AudioSampleEntry", "samples_per_packet = %lu\n", soun->samples_per_packet);
        
        if ((nwrite = fwriteulong32(&soun->bytes_per_packet, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 4;
        spDebug(10, "spWriteMp4AudioSampleEntry", "bytes_per_packet = %lu\n", soun->bytes_per_packet);

            if ((nwrite = fwriteulong32(&soun->bytes_per_frame, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 4;
        spDebug(10, "spWriteMp4AudioSampleEntry", "bytes_per_frame = %lu\n", soun->bytes_per_frame);

            if ((nwrite = fwriteulong32(&soun->bytes_per_sample, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite += 4;
        spDebug(10, "spWriteMp4AudioSampleEntry", "bytes_per_sample = %lu\n", soun->bytes_per_sample);
    }

    remain_size -= total_nwrite;
    spDebug(50, "spWriteMp4AudioSampleEntry", "remain_size = %lu, total_nwrite = %lu\n", (unsigned long)remain_size, (unsigned long)total_nwrite);

    if (remain_size >= 8) {
        if ((nwrite = (long)spWriteMp4ChildOnlyBox((spChunk *)box, depth + 1, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        spDebug(10, "spWriteMp4AudioSampleEntry", "spWriteMp4ChildOnlyBox result = %ld\n", nwrite);
        total_nwrite += nwrite;
    }

    spDebug(10, "spWriteMp4AudioSampleEntry", "done: total_nwrite = %lu, version = %d, channelcount = %d, samplesize = %d, samplerate = %f\n",
            (unsigned long)total_nwrite, soun->version, soun->channelcount, soun->samplesize, (double)soun->samplerate / 65536.0);
    
    return total_nwrite;
}

long spCopyMp4AudioSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box, long depth)
{
    long ncopy;
    long total_ncopy;
    spMp4AudioSampleEntry *dest_soun;
    spMp4AudioSampleEntry *src_soun;

    dest_soun = &dest_box->u.soun;
    src_soun = &src_box->u.soun;

    total_ncopy = spOffsetOf(spMp4AudioSampleEntry, frma);
    memcpy(dest_soun, src_soun, total_ncopy);
    
    if ((ncopy = (long)spCopyMp4ChildBox((spChunk *)dest_box, (spChunk *)src_box, depth + 1)) > 0) {
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

spMp4BoxSize spReadMp4RtpHintSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4RtpHintSampleEntry *rtp;

    spDebug(50, "spReadMp4RtpHintSampleEntry", "in: remain_size = %lu, depth = %ld\n",
            (unsigned long)remain_size, depth);
    
    rtp = &box->u.rtp;

    if ((nread = freadshort((short *)&rtp->hint_track_version, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;

    if ((nread = freadshort((short *)&rtp->highest_compatible_version, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    
    if ((nread = freadulong32(&rtp->max_packet_size, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    remain_size -= total_nread;
    spDebug(50, "spReadMp4RtpHintSampleEntry", "remain_size = %lu, total_nread = %lu\n", (unsigned long)remain_size, (unsigned long)total_nread);

    if (remain_size >= 8) {
        spDebug(50, "spReadMp4RtpHintSampleEntry", "type = %c%c%c%c\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
        if ((nread = (long)spReadMp4ChildBox((spChunk *)box, 0, remain_size, depth + 1, swap, fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
        remain_size -= nread;
    }
    
    spDebug(10, "spReadMp4RtpHintSampleEntry", "done: total_nread = %lu / %lu, remain_size = %lu\n",
            (unsigned long)total_nread, box->header.size, (unsigned long)remain_size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4RtpHintSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4RtpHintSampleEntry *rtp;

    spDebug(50, "spWriteMp4RtpHintSampleEntry", "in: remain_size = %lu, depth = %ld\n",
            (unsigned long)remain_size, depth);
    
    rtp = &box->u.rtp;

    if ((nwrite = fwriteshort((short *)&rtp->hint_track_version, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;

    if ((nwrite = fwriteshort((short *)&rtp->highest_compatible_version, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = fwriteulong32(&rtp->max_packet_size, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    remain_size -= total_nwrite;
    spDebug(50, "spWriteMp4RtpHintSampleEntry", "remain_size = %lu, total_nwrite = %lu\n", (unsigned long)remain_size, (unsigned long)total_nwrite);

    if (remain_size >= 8) {
        spDebug(50, "spWriteMp4RtpHintSampleEntry", "type = %c%c%c%c\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
        if ((nwrite = (long)spWriteMp4ChildOnlyBox((spChunk *)box, depth + 1, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
        remain_size -= nwrite;
    }
    
    spDebug(10, "spWriteMp4RtpHintSampleEntry", "done: total_nwrite = %lu / %lu, remain_size = %lu\n",
            (unsigned long)total_nwrite, box->header.size, (unsigned long)remain_size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4RtpHintSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box, long depth)
{
    long ncopy;
    spMp4BoxSize total_ncopy;
    spMp4RtpHintSampleEntry *dest_rtp;
    spMp4RtpHintSampleEntry *src_rtp;

    dest_rtp = &dest_box->u.rtp;
    src_rtp = &src_box->u.rtp;
    total_ncopy = spOffsetOf(spMp4RtpHintSampleEntry, tims);
    memcpy(dest_rtp, src_rtp, (long)total_ncopy);

    if ((ncopy = (long)spCopyMp4ChildBox((spChunk *)dest_box, (spChunk *)src_box, depth + 1)) > 0) {
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

long spReadMp4BitRateBox(spMp4BitRateBox *box, int swap, FILE *fp)
{
    long nread;
    long total_nread = 0;

    if ((nread = (long)spReadMp4BoxHeader((spMp4Box *)box, &box->btrt, swap, fp)) <= 0) {
        return nread;
    }
    total_nread = nread;
    
    if ((nread = freadulong32(&box->bufferSizeDB, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    if ((nread = freadulong32(&box->maxBitRate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    if ((nread = freadulong32(&box->avgBitRate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    return total_nread;
}

long spWriteMp4BitRateBox(spMp4BitRateBox *box, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    long total_nwrite = 0;

    if ((nwrite = (long)spWriteMp4BoxHeader((spMp4Box *)box, &box->btrt, swap, fp)) <= 0) {
        return nwrite;
    }
    total_nwrite = nwrite;
    
    if ((nwrite = fwriteulong32(&box->bufferSizeDB, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(&box->maxBitRate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(&box->avgBitRate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    return total_nwrite;
}

long spCopyMp4BitRateBox(spMp4BitRateBox *dest_box, spMp4BitRateBox *src_box)
{
    long ncopy;

    ncopy = sizeof(spMp4BitRateBox);
    memcpy(dest_box, src_box, ncopy);
    
    return ncopy;
}

spMp4BoxSize spReadMp4DataSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize content_size;
    spMp4BoxSize total_nread = 0;

    spDebug(50, "spReadMp4DataSampleEntry", "type = %c%c%c%c\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);

    content_size = spGetMp4BoxHeaderContentSize(&box->header) - 8;
    
    if (content_size > 0) {
        box->u.data.data = xalloc((long)content_size + 1, char);
        if ((nread = (long)fread(box->u.data.data, 1, (long)content_size, fp)) != content_size) {
            xfree(box->u.data.data);
            box->u.data.data = NULL;
            return nread;
        }
        box->u.data.data[content_size] = NUL;
#if 0
        showDataInHex(box->u.data.data, content_size);
#endif
        spDebug(50, "spReadMp4DataSampleEntry", "data = %s\n", box->u.data.data);
        total_nread += nread;
    } else {
        box->u.data.data = NULL;
    }
    
    return total_nread;
}
    
spMp4BoxSize spWriteMp4DataSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize content_size;
    spMp4BoxSize total_nwrite = 0;

    spDebug(50, "spWriteMp4DataSampleEntry", "type = %c%c%c%c\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    content_size = spGetMp4BoxHeaderContentSize(&box->header) - 8;
    
    if (content_size > 0) {
        if ((nwrite = (long)fwrite(box->u.data.data, 1, (long)content_size, fp)) != content_size) {
            return nwrite;
        }
        total_nwrite += nwrite;
    }
    
    spDebug(50, "spWriteMp4DataSampleEntry", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);

    return total_nwrite;
}

spMp4BoxSize spCopyMp4DataSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box)
{
    spMp4BoxSize content_size;

    content_size = spGetMp4BoxHeaderContentSize(&src_box->header) - 8;
    
    if (content_size > 0) {
        dest_box->u.data.data = xalloc((long)content_size + 1, char);
        memcpy(dest_box->u.data.data, src_box->u.data.data, (long)content_size);
        dest_box->u.data.data[content_size] = NUL;
    } else {
        dest_box->u.data.data = NULL;
    }

    return content_size;
}

long spReadMp4MetaSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, int swap, FILE *fp)
{
    long nread;
    long total_nread = 0;
    spMp4BitRateBox *bitRate = NULL;

    spDebug(50, "spReadMp4MetaSampleEntry", "type = %c%c%c%c\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    
    if (strneq(box->header.type, "metx", 4)) {
        if ((nread = readMp4String(box->u.metx.content_encoding,
                                   sizeof(box->u.metx.content_encoding), fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
        
        if ((nread = readMp4String(box->u.metx.name_space,
                                   sizeof(box->u.metx.name_space), fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
        
        if ((nread = readMp4String(box->u.metx.schema_location,
                                   sizeof(box->u.metx.schema_location), fp)) <= 0) {
            return nread;
        }
        total_nread += nread;

        bitRate = &box->u.metx.bitRate;
    } else if (strneq(box->header.type, "mett", 4)) {
        if ((nread = readMp4String(box->u.mett.content_encoding,
                                   sizeof(box->u.mett.content_encoding), fp)) <= 0) {
            return nread;
        }
        total_nread += nread;

        if ((nread = readMp4String(box->u.mett.mime_format,
                                   sizeof(box->u.mett.mime_format), fp)) <= 0) {
            return nread;
        }
        total_nread += nread;

        bitRate = &box->u.mett.bitRate;
    }

    remain_size -= total_nread;
    spDebug(50, "spReadMp4MetaSampleEntry", "remain_size = %lu, total_nread = %lu\n", (unsigned long)remain_size, (unsigned long)total_nread);

    if (bitRate != NULL && remain_size >= 20) {
        if ((nread = spReadMp4BitRateBox(bitRate, swap, fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
    }
    
    return total_nread;
}
    
long spWriteMp4MetaSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    long total_nwrite = 0;
    spMp4BitRateBox *bitRate = NULL;

    if (strneq(box->header.type, "metx", 4)) {
        if ((nwrite = writeMp4String(box->u.metx.content_encoding,
                                     sizeof(box->u.metx.content_encoding), fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
        
        if ((nwrite = writeMp4String(box->u.metx.name_space,
                                   sizeof(box->u.metx.name_space), fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
        
        if ((nwrite = writeMp4String(box->u.metx.schema_location,
                                   sizeof(box->u.metx.schema_location), fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;

        bitRate = &box->u.metx.bitRate;
    } else if (strneq(box->header.type, "mett", 4)) {
        if ((nwrite = writeMp4String(box->u.mett.content_encoding,
                                   sizeof(box->u.mett.content_encoding), fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;

        if ((nwrite = writeMp4String(box->u.mett.mime_format,
                                   sizeof(box->u.mett.mime_format), fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;

        bitRate = &box->u.mett.bitRate;
    }

    remain_size -= total_nwrite;
    
    if (bitRate != NULL && remain_size >= 20) {
        if ((nwrite = spWriteMp4BitRateBox(bitRate, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
    }

    return total_nwrite;
}

long spCopyMp4MetaSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box)
{
    long ncopy = 0;

    if (strneq(src_box->header.type, "metx", 4)) {
        ncopy = sizeof(spMp4XMLMetaDataSampleEntry);
        memcpy(&dest_box->u.metx, &src_box->u.metx, ncopy);
    } else if (strneq(src_box->header.type, "mett", 4)) {
        ncopy = sizeof(spMp4TextMetaDataSampleEntry);
        memcpy(&dest_box->u.mett, &src_box->u.mett, ncopy);
    }

    return ncopy;
}

spMp4BoxSize spReadMp4TimecodeSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, int swap, FILE *fp)
{
    long nread;
    long total_nread = 0;

    if ((nread = freadulong32(&box->u.tmcd.reserved, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&box->u.tmcd.flags, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    if ((nread = freadulong32(&box->u.tmcd.timescale, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    if ((nread = freadulong32(&box->u.tmcd.frame_duration, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    if ((nread = (long)fread(&box->u.tmcd.number_of_frames, 1, 1, fp)) != 1) {
        return nread;
    }
    total_nread += 1;
    
    if ((nread = (long)fread(&box->u.tmcd.reserved2, 1, 1, fp)) != 1) {
        return nread;
    }
    total_nread += 1;

    spDebug(50, "spReadMp4TimecodeSampleEntry", "flags = %lx, timescale = %lu, frame_duration = %lu, number_of_frames = %d\n",
            box->u.tmcd.flags, box->u.tmcd.timescale, box->u.tmcd.frame_duration, box->u.tmcd.number_of_frames);
    
    remain_size -= total_nread;
    spDebug(50, "spReadMp4TimecodeSampleEntry", "remain_size = %lu, total_nread = %lu\n", (unsigned long)remain_size, (unsigned long)total_nread);

    if (remain_size >= 8) {
        spDebug(50, "spReadMp4TimecodeSampleEntry", "type = %c%c%c%c\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
        if ((nread = (long)spReadMp4ChildBox((spChunk *)box, 0, remain_size, depth + 1, swap, fp)) <= 0) {
            return nread;
        }
        total_nread += nread;
        remain_size -= nread;
    }
    
    return total_nread;
}

spMp4BoxSize spWriteMp4TimecodeSampleEntry(spMp4SampleEntryBox *box, spMp4BoxSize remain_size, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4TimecodeSampleEntry *tmcd;

    spDebug(50, "spWriteMp4TimecodeSampleEntry", "in: remain_size = %lu, depth = %ld\n",
            (unsigned long)remain_size, depth);
    
    tmcd = &box->u.tmcd;
    
    if ((nwrite = fwriteulong32(&tmcd->reserved, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if ((nwrite = fwriteulong32(&tmcd->flags, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(&tmcd->timescale, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(&tmcd->frame_duration, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = (long)fwrite(&tmcd->number_of_frames, 1, 1, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 1;

    if ((nwrite = (long)fwrite(&tmcd->reserved2, 1, 1, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 1;

    remain_size -= total_nwrite;
    spDebug(50, "spWriteMp4TimecodeSampleEntry", "remain_size = %lu, total_nwrite = %lu\n", (unsigned long)remain_size, (unsigned long)total_nwrite);

    if (remain_size >= 8) {
        spDebug(50, "spWriteMp4TimecodeSampleEntry", "type = %c%c%c%c\n",
                box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
        if ((nwrite = (long)spWriteMp4ChildOnlyBox((spChunk *)box, depth + 1, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        total_nwrite += nwrite;
        remain_size -= nwrite;
    }
    
    spDebug(10, "spWriteMp4TimecodeSampleEntry", "done: total_nwrite = %lu / %lu, remain_size = %lu\n",
            (unsigned long)total_nwrite, box->header.size, (unsigned long)remain_size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4TimecodeSampleEntry(spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box, long depth)
{
    long ncopy;
    spMp4BoxSize total_ncopy;
    spMp4TimecodeSampleEntry *dest_tmcd;
    spMp4TimecodeSampleEntry *src_tmcd;
    
    dest_tmcd = &dest_box->u.tmcd;
    src_tmcd = &src_box->u.tmcd;
    total_ncopy = spOffsetOf(spMp4TimecodeSampleEntry, name);
    memcpy(dest_tmcd, src_tmcd, (long)total_ncopy);

    if ((ncopy = (long)spCopyMp4ChildBox((spChunk *)dest_box, (spChunk *)src_box, depth + 1)) > 0) {
        total_ncopy += ncopy;
    }
    
    return ncopy;
}

spMp4BoxSize spReadMp4SampleEntryBox(spMp4SampleDescriptionBox *stsd, char handler_type[4],
                                     spMp4SampleEntryBox *box, long depth, int swap, FILE *fp)
{
    spMp4BoxSize nread;
    spMp4BoxSize remain_size;
    spMp4BoxSize total_nread;

    spDebug(50, "spReadMp4SampleEntryBox", "handler_type = %c%c%c%c\n",
            handler_type[0], handler_type[1], handler_type[2], handler_type[3]);
    
    spDebug(50, "spReadMp4SampleEntryBox", "box type = %c%c%c%c, depth = %ld\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3], depth);
    
    box->parent = (spMp4Box *)stsd;

    if ((nread = spReadMp4BoxHeader((spMp4Box *)box, &box->header, swap, fp)) <= 0) {
        return nread;
    }
    total_nread = nread;
    
    if ((nread = (long)fread(box->reserved, 1, 6, fp)) != 6) {
        return nread;
    }
    total_nread += 6;
    spDebug(50, "spReadMp4SampleEntryBox", "reserved = %d-%d-%d-%d-%d-%d\n",
            box->reserved[0], box->reserved[1], box->reserved[2],
            box->reserved[3], box->reserved[4], box->reserved[5]);

    if ((nread = freadshort((short *)&box->data_reference_index, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4SampleEntryBox", "data_reference_index = %d\n", box->data_reference_index);

    remain_size = spGetMp4BoxSize((spMp4Box *)box, NULL) - total_nread;
    spDebug(50, "spReadMp4SampleEntryBox", "current total_nread = %lu, remain_size = %lu\n",
            (unsigned long)total_nread, (unsigned long)remain_size);
    
    if (strneq(handler_type, "vide", 4)) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'vide'\n");
        if ((nread = (long)spReadMp4VisualSampleEntry(box, remain_size, depth, swap, fp)) <= 0) {
            return nread;
        }
    } else if (strneq(handler_type, "soun", 4)) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'soun'\n");
        if ((nread = spReadMp4AudioSampleEntry(box, remain_size, depth, swap, fp)) <= 0) {
            return nread;
        }
    } else if (strneq(handler_type, "hint", 4)) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'rtp '\n");
        if ((nread = (long)spReadMp4RtpHintSampleEntry(box, remain_size, depth, swap, fp)) <= 0) {
            return nread;
        }
#if 0
    } else if (strneq(handler_type, "hint", 4)) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'hint'\n");
        if (remain_size > 0) {
            box->u.hint.data = xalloc(remain_size, unsigned char);
            if ((nread = (long)fread(box->u.hint.data, 1, remain_size, fp)) != remain_size) {
                return nread;
            }
        } else {
            box->u.hint.data = NULL;
            nread = 0;
        }
#endif
    } else if (strneq(handler_type, "meta", 4)
               && (strneq(box->header.type, "metx", 4) || strneq(box->header.type, "mett", 4))) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'meta'\n");
        if ((nread = spReadMp4MetaSampleEntry(box, remain_size, swap, fp)) <= 0) {
            return nread;
        }
    } else if (strneq(handler_type, "tmcd", 4)) {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is 'tmcd'\n");
        if ((nread = spReadMp4TimecodeSampleEntry(box, remain_size, depth, swap, fp)) <= 0) {
            return nread;
        }
    } else {
        spDebug(50, "spReadMp4SampleEntryBox", "handler_type is unknown data\n");
        if ((nread = spReadMp4DataSampleEntry(box, remain_size, swap, fp)) <= 0) {
            return nread;
        }
    }
    total_nread += nread;

    spDebug(10, "spReadMp4SampleEntryBox", "done: total_nread = %lu / %lu\n",
            (unsigned long)total_nread, box->header.size);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4SampleEntryBox(spMp4SampleDescriptionBox *stsd, char handler_type[4],
                                      spMp4SampleEntryBox *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spMp4BoxSize nwrite;
    spMp4BoxSize remain_size;
    spMp4BoxSize total_nwrite;

    spDebug(50, "spWriteMp4SampleEntryBox", "box type = %c%c%c%c, depth = %ld\n",
            box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3], depth);
    
    if ((nwrite = spWriteMp4BoxHeader((spMp4Box *)box, &box->header, swap, fp)) <= 0) {
        return nwrite;
    }
    total_nwrite = nwrite;
    
    if ((nwrite = (long)fwrite(box->reserved, 1, 6, fp)) != 6) {
        return nwrite;
    }
    total_nwrite += 6;

    spDebug(50, "spWriteMp4SampleEntryBox", "data_reference_index = %d\n", box->data_reference_index);
    
    if ((nwrite = fwriteshort((short *)&box->data_reference_index, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    spDebug(50, "spWriteMp4SampleEntryBox", "current total_nwrite = %lu\n", (unsigned long)total_nwrite);

    remain_size = spGetMp4BoxSize((spMp4Box *)box, NULL) - total_nwrite;
    
    if (strneq(handler_type, "vide", 4)) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'vide'\n");
        if ((nwrite = spWriteMp4VisualSampleEntry(box, remain_size, depth, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
    } else if (strneq(handler_type, "soun", 4)) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'soun'\n");
        if ((nwrite = spWriteMp4AudioSampleEntry(box, remain_size, depth, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
    } else if (strneq(handler_type, "hint", 4)) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'rtp '\n");
        if ((nwrite = spWriteMp4RtpHintSampleEntry(box, remain_size, depth, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
#if 0
    } else if (strneq(handler_type, "hint", 4)) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'hint'\n");
        if (remain_size > 0) {
            if ((nwrite = (long)fwrite(box->u.hint.data, 1, remain_size, fp)) != remain_size) {
                return nwrite;
            }
        } else {
            nwrite = 0;
        }
#endif
    } else if (strneq(handler_type, "meta", 4)
               && (strneq(box->header.type, "metx", 4) || strneq(box->header.type, "mett", 4))) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'meta'\n");
        if ((nwrite = spWriteMp4MetaSampleEntry(box, remain_size, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
    } else if (strneq(handler_type, "tmcd", 4)) {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is 'meta'\n");
        if ((nwrite = spWriteMp4TimecodeSampleEntry(box, remain_size, depth, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
    } else {
        spDebug(50, "spWriteMp4SampleEntryBox", "handler_type is unknown data\n");
        if ((nwrite = spWriteMp4DataSampleEntry(box, remain_size, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
    }
    total_nwrite += nwrite;

    spDebug(10, "spWriteMp4SampleEntryBox", "done: total_nwrite = %lu / %lu\n",
            (unsigned long)total_nwrite, box->header.size);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SampleEntryBox(spMp4SampleDescriptionBox *stsd, char handler_type[4],
                                     spMp4SampleEntryBox *dest_box, spMp4SampleEntryBox *src_box, long depth)
{
    spMp4BoxSize ncopy;
    spMp4BoxSize total_ncopy;

    dest_box->parent = (spMp4Box *)stsd;

    if ((ncopy = spCopyMp4BoxHeader(&dest_box->header, &src_box->header)) <= 0) {
        return ncopy;
    }
    total_ncopy = ncopy;

    ncopy = sizeof(dest_box->reserved);
    memcpy(dest_box->reserved, src_box->reserved, (long)ncopy);
    total_ncopy += ncopy;

    dest_box->data_reference_index = src_box->data_reference_index;
    total_ncopy += sizeof(dest_box->data_reference_index);

    if (strneq(handler_type, "vide", 4)) {
        ncopy = spCopyMp4VisualSampleEntry(dest_box, src_box, depth);
    } else if (strneq(handler_type, "soun", 4)) {
        ncopy = spCopyMp4AudioSampleEntry(dest_box, src_box, depth);
    } else if (strneq(handler_type, "hint", 4)) {
        ncopy = spCopyMp4RtpHintSampleEntry(dest_box, src_box, depth);
    } else if (strneq(handler_type, "meta", 4)
               && (strneq(src_box->header.type, "metx", 4) || strneq(src_box->header.type, "mett", 4))) {
        ncopy = spCopyMp4MetaSampleEntry(dest_box, src_box);
    } else if (strneq(handler_type, "tmcd", 4)) {
        ncopy = spCopyMp4TimecodeSampleEntry(dest_box, src_box, depth);
    } else {
        ncopy = spCopyMp4DataSampleEntry(dest_box, src_box);
    }
    total_ncopy += ncopy;

    return total_ncopy;
}

/* 'stsd' */
spMp4BoxSize spReadMp4SampleDescriptionBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long i;
    spMp4BoxSize nread;
    spMp4BoxSize remain_size;
    spMp4BoxSize total_nread = 0;
    char handler_type[4];
    spMp4SampleDescriptionBox *stsd = (spMp4SampleDescriptionBox *)box;
    
    if (spGetMp4BoxHandleType((spMp4Box *)stsd, handler_type) == SP_FALSE) {
        spDebug(10, "spReadMp4SampleDescriptionBox", "spGetMp4BoxHandleType failed\n");
        return 0;
    }
    spDebug(10, "spReadMp4SampleDescriptionBox", "handler_type = %c%c%c%c\n",
            handler_type[0], handler_type[1], handler_type[2], handler_type[3]);
    
    if ((nread = freadulong32(&stsd->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(10, "spReadMp4SampleDescriptionBox", "depth = %ld, entry_count = %lu\n", depth, stsd->entry_count);

    remain_size = spGetMp4BoxHeaderContentSize(&stsd->header) - total_nread;

    stsd->end_marker_size = 0;
    
    if (stsd->entry_count > 0) {
        stsd->alloc_count = (1 + stsd->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        stsd->entries = xalloc(stsd->entry_count, spMp4SampleEntryBox);
        for (i = 0; i < stsd->entry_count; i++) {
            memset(&stsd->entries[i], 0, sizeof(spMp4SampleEntryBox));
            
            if ((nread = spReadMp4SampleEntryBox(stsd, handler_type, &stsd->entries[i], depth + 1, swap, fp)) <= 0) {
                return nread;
            }

            if (i == 0) {
                stsd->child = (spMp4Box *)&stsd->entries[0];
            } else {
                stsd->entries[i].prev = (spMp4Box *)&stsd->entries[i - 1];
                stsd->entries[i - 1].next = (spMp4Box *)&stsd->entries[i].prev;
            }
            
            total_nread += nread;
            remain_size -= nread;

            if (remain_size > 0 && remain_size <= 4) {
                /* maybe optional end of box marker */
                spDebug(10, "spReadMp4SampleDescriptionBox", "optional end of entry found: remain_size = %lu\n", (unsigned long)remain_size);
                spSeekFile(fp, remain_size, SEEK_CUR);
                total_nread += remain_size;
                stsd->end_marker_size = (unsigned long)remain_size;
                break;
            }
        }
    } else {
        stsd->alloc_count = 0;
        stsd->entries = NULL;
    }

    spDebug(50, "spReadMp4SampleDescriptionBox", "total_nread = %lu / %lu (handler_type = %c%c%c%c), remain_size = %lu\n",
            (unsigned long)total_nread, stsd->header.size + 12, handler_type[0], handler_type[1], handler_type[2], handler_type[3], (unsigned long)remain_size);
    
    return total_nread;
}

spBool spFreeMp4SampleDescriptionBox(void *box)
{
    unsigned long i;
    char handler_type[4];
    spMp4SampleDescriptionBox *stsd = (spMp4SampleDescriptionBox *)box;

    if (stsd->entries != NULL) {
        if (spGetMp4BoxHandleType((spMp4Box *)stsd, handler_type) == SP_TRUE) {
            if (strneq(handler_type, "hint", 4)) {
                for (i = 0; i < stsd->entry_count; i++) {
                    if (stsd->entries[i].u.hint.data != NULL) {
                        xfree(stsd->entries[i].u.hint.data);
                    }
                }
            }
        }
        xfree(stsd->entries);
        stsd->entries = NULL;
    }
    
    return SP_TRUE;
}

spMp4SampleEntryBox *spAppendMp4SampleDescriptionEntry(spMp4SampleDescriptionBox *stsd, const char *codingname, unsigned short data_reference_index)
{
    unsigned long i;
    unsigned long alloc_count;

    spDebug(80, "spAppendMp4SampleDescriptionEntry", "old alloc_count = %lu, entry_count = %lu\n",
            stsd->alloc_count, stsd->entry_count);
    
    if (stsd->alloc_count <= 0) {
        stsd->entries = NULL;
    }
    
    stsd->entry_count++;
    alloc_count = (1 + stsd->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;

    if (alloc_count > stsd->alloc_count) {
        stsd->alloc_count = alloc_count;
        stsd->entries = xrealloc(stsd->entries, stsd->alloc_count, spMp4SampleEntryBox);
    }
    spDebug(80, "spAppendMp4SampleDescriptionEntry", "updated alloc_count = %lu, entry_count = %lu\n",
            stsd->alloc_count, stsd->entry_count);

    i = stsd->entry_count - 1;
    memset(&stsd->entries[i], 0, sizeof(spMp4SampleEntryBox));
    memcpy(stsd->entries[i].header.type, codingname, 4);
    stsd->entries[i].header.size = 16; /* 8 + 8 */
    stsd->entries[i].data_reference_index = data_reference_index;

    if (i == 0) {
        stsd->child = (spMp4Box *)&stsd->entries[0];
    } else {
        stsd->entries[i].prev = (spMp4Box *)&stsd->entries[i - 1];
        stsd->entries[i - 1].next = (spMp4Box *)&stsd->entries[i].prev;
    }
    stsd->entries[i].parent = (spMp4Box *)stsd;

    spPropagateMp4BoxContentSize((spMp4Box *)stsd, 16);
        
    return &stsd->entries[i];
}

spMp4BoxSize spWriteMp4SampleDescriptionBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long i;
    spMp4BoxSize nwrite;
    spMp4BoxSize total_nwrite = 0;
    char handler_type[4];
    spMp4SampleDescriptionBox *stsd = (spMp4SampleDescriptionBox *)box;
    
    if (spGetMp4BoxHandleType((spMp4Box *)stsd, handler_type) == SP_FALSE) {
        spDebug(10, "spWriteMp4SampleDescriptionBox", "spGetMp4BoxHandleType failed\n");
        return 0;
    }
    spDebug(10, "spWriteMp4SampleDescriptionBox", "handler_type = %c%c%c%c\n",
            handler_type[0], handler_type[1], handler_type[2], handler_type[3]);
    
    spDebug(10, "spWriteMp4SampleDescriptionBox", "depth = %ld, entry_count = %lu\n", depth, stsd->entry_count);
    if ((nwrite = fwriteulong32(&stsd->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;

    if (stsd->entry_count > 0) {
        for (i = 0; i < stsd->entry_count; i++) {
            if ((nwrite = spWriteMp4SampleEntryBox(stsd, handler_type, &stsd->entries[i], depth + 1, rewrite_flag, swap, fp)) <= 0) {
                return nwrite;
            }
            
            total_nwrite += nwrite;
        }
        if (stsd->end_marker_size > 0) {
            char c = 0;
            for (i = 0; i < stsd->end_marker_size; i++) {
                if ((nwrite = fwrite(&c, 1, 1, fp)) != 1) {
                    return nwrite;
                }
                ++total_nwrite;
            }
        }
    }

    spDebug(50, "spWriteMp4SampleDescriptionBox", "total_nwrite = %lu / %lu (handler_type = %c%c%c%c)\n",
            (unsigned long)total_nwrite, stsd->header.size + 12, handler_type[0], handler_type[1], handler_type[2], handler_type[3]);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4SampleDescriptionBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long i;
    spMp4BoxSize ncopy;
    spMp4BoxSize total_ncopy = 0;
    char handler_type[4];
    spMp4SampleDescriptionBox *dest_stsd = (spMp4SampleDescriptionBox *)dest_box;
    spMp4SampleDescriptionBox *src_stsd = (spMp4SampleDescriptionBox *)src_box;
    
    if (spGetMp4BoxHandleType((spMp4Box *)src_stsd, handler_type) == SP_FALSE) {
        spDebug(10, "spCopyMp4SampleDescriptionBox", "spGetMp4BoxHandleType failed\n");
        return 0;
    }

    dest_stsd->entry_count = src_stsd->entry_count;
    total_ncopy = sizeof(dest_stsd->entry_count);

    total_ncopy += sizeof(dest_stsd->alloc_count) + sizeof(dest_stsd->entries);
    
    if (dest_stsd->entry_count > 0) {
        dest_stsd->alloc_count = src_stsd->alloc_count;
        dest_stsd->entries = xalloc(dest_stsd->entry_count, spMp4SampleEntryBox);
        for (i = 0; i < dest_stsd->entry_count; i++) {
            memset(&dest_stsd->entries[i], 0, sizeof(spMp4SampleEntryBox));

            ncopy = spCopyMp4SampleEntryBox(dest_stsd, handler_type, &dest_stsd->entries[i], &src_stsd->entries[i], depth + 1);

            if (i == 0) {
                dest_stsd->child = (spMp4Box *)&dest_stsd->entries[0];
            } else {
                dest_stsd->entries[i].prev = (spMp4Box *)&dest_stsd->entries[i - 1];
                dest_stsd->entries[i - 1].next = (spMp4Box *)&dest_stsd->entries[i].prev;
            }

            total_ncopy += ncopy;
        }
    } else {
        dest_stsd->alloc_count = 0;
        dest_stsd->entries = NULL;
    }

    return total_ncopy;
}

/* 'vmhd' */
spMp4BoxSize spReadMp4VideoMediaHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4VideoMediaHeaderBox *vmhd = (spMp4VideoMediaHeaderBox *)box;

    if ((nread = freadshort((short *)&vmhd->graphicsmode, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;

    if ((nread = freadshort((short *)vmhd->opcolor, 3, swap, fp)) != 3) {
        return nread;
    }
    total_nread += 2 * 3;
    
    return total_nread;
}

spMp4BoxSize spUpdateMp4VideoMediaHeaderBox(spMp4VideoMediaHeaderBox *vmhd, unsigned short graphicsmode, unsigned short opcolor[3])
{
    spMp4BoxSize size;

    vmhd->graphicsmode = graphicsmode;
    memcpy(vmhd->opcolor, opcolor, sizeof(vmhd->opcolor));
    size = 8;
    spSetMp4BoxContentSize((spMp4Box *)vmhd, size, SP_FALSE);

    return size;
}

spMp4BoxSize spWriteMp4VideoMediaHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4VideoMediaHeaderBox *vmhd = (spMp4VideoMediaHeaderBox *)box;

    if ((nwrite = fwriteshort((short *)&vmhd->graphicsmode, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;

    if ((nwrite = fwriteshort((short *)vmhd->opcolor, 3, swap, fp)) != 3) {
        return nwrite;
    }
    total_nwrite += 2 * 3;
    
    return total_nwrite;
}

/* 'smhd' */
spMp4BoxSize spReadMp4SoundMediaHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4SoundMediaHeaderBox *smhd = (spMp4SoundMediaHeaderBox *)box;

    if ((nread = freadshort((short *)&smhd->balance, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;
    spDebug(10, "spReadMp4SoundMediaHeaderBox", "balance = %d\n", smhd->balance);

    if ((nread = freadshort((short *)&smhd->reserved, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(10, "spReadMp4SoundMediaHeaderBox", "reserved = %d\n", smhd->reserved);
    
    return total_nread;
}

spMp4BoxSize spUpdateMp4SoundMediaHeaderBox(spMp4SoundMediaHeaderBox *smhd, unsigned short balance)
{
    spMp4BoxSize size;

    smhd->balance = balance;
    smhd->reserved = 0;
    
    size = 4;
    spSetMp4BoxContentSize((spMp4Box *)smhd, size, SP_FALSE);

    return size;
}

spMp4BoxSize spWriteMp4SoundMediaHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4SoundMediaHeaderBox *smhd = (spMp4SoundMediaHeaderBox *)box;

    if ((nwrite = fwriteshort((short *)&smhd->balance, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;

    if ((nwrite = fwriteshort((short *)&smhd->reserved, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    return total_nwrite;
}

/* 'hmhd' */
spMp4BoxSize spReadMp4HintMediaHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4HintMediaHeaderBox *hmhd = (spMp4HintMediaHeaderBox *)box;

    if ((nread = freadshort((short *)&hmhd->maxPDUSize, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 2;
    if ((nread = freadshort((short *)&hmhd->avgPDUSize, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;

    if ((nread = freadulong32(&hmhd->maxBitRate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    if ((nread = freadulong32(&hmhd->avgBitRate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&hmhd->reserved, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    return total_nread;
}

spMp4BoxSize spUpdateMp4HintMediaHeaderBox(spMp4HintMediaHeaderBox *hmhd, unsigned short maxPDUSize, unsigned short avgPDUSize,
                                           unsigned long maxBitRate, unsigned long avgBitRate)
{
    spMp4BoxSize size;

    hmhd->maxPDUSize = maxPDUSize;
    hmhd->avgPDUSize = avgPDUSize;
    hmhd->maxBitRate = maxBitRate;
    hmhd->avgBitRate = avgBitRate;
    hmhd->reserved = 0;
    
    size = 16;
    spSetMp4BoxContentSize((spMp4Box *)hmhd, size, SP_FALSE);

    return size;
}

spMp4BoxSize spWriteMp4HintMediaHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4HintMediaHeaderBox *hmhd = (spMp4HintMediaHeaderBox *)box;

    if ((nwrite = fwriteshort((short *)&hmhd->maxPDUSize, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 2;
    if ((nwrite = fwriteshort((short *)&hmhd->avgPDUSize, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = fwriteulong32(&hmhd->maxBitRate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    if ((nwrite = fwriteulong32(&hmhd->avgBitRate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&hmhd->reserved, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    return total_nwrite;
}

/* 'keys' */
long spReadMp4KeyEntry(spMp4KeyEntry *key, int swap, FILE *fp)
{
    long len;
    long nread;
    long total_nread;
    
    if ((nread = freadulong32(&key->size, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    
    if ((nread = (long)fread(key->name_space, 1, 4, fp)) != 4) {
        return nread;
    }
    total_nread += 4;

    len = key->size - 8;
    key->value = xalloc(len + 1, char);
    
    if ((nread = (long)fread(key->value, 1, len, fp)) != len) {
        return nread;
    }
    key->value[len] = NUL;
    total_nread += nread;

    return total_nread;
}

spMp4BoxSize spReadMp4KeysBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long k;
    long nread;
    spMp4BoxSize total_nread;
    spMp4KeysBox *keys = (spMp4KeysBox *)box;

    if ((nread = freadulong32(&keys->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(10, "spReadMp4KeysBox", "entry_count = %lu\n", keys->entry_count);

    if (keys->entry_count > 0) {
        keys->alloc_count = (1 + keys->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        keys->keys = xalloc(keys->alloc_count, spMp4KeyEntry);
    
        for (k = 0; k < keys->entry_count; k++) {
            if ((nread = spReadMp4KeyEntry(&keys->keys[k], swap, fp)) < 8) {
                spDebug(10, "spReadMp4KeysBox", "spReadMp4KeyEntry failed: nread = %ld\n", nread);
                return nread;
            }
            total_nread += nread;
        }
    }

    spDebug(10, "spReadMp4KeysBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spFreeMp4KeysBox(void *box)
{
    unsigned long k;
    spMp4KeysBox *keys = (spMp4KeysBox *)box;

    if (keys->keys != NULL) {
        for (k = 0; k < keys->entry_count; k++) {
            xfree(keys->keys[k].value);
        }
        xfree(keys->keys);
        keys->keys = NULL;
    }
    
    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4KeysBox(spMp4KeysBox *keys, spMp4KeyEntry *entries, long count, spBool propagate_size)
{
    unsigned long k;
    long len;
    spMp4BoxSize size;
    
    spFreeMp4KeysBox(keys);

    keys->entry_count = count;
    size = 4;
    
    if (keys->entry_count > 0) {
        keys->alloc_count = (1 + keys->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
        keys->keys = xalloc(keys->alloc_count, spMp4KeyEntry);
        
        for (k = 0; k < keys->entry_count; k++) {
            size += entries[k].size;
            
            keys->keys[k].size = entries[k].size;
            memcpy(keys->keys[k].name_space, entries[k].name_space, 4);
            
            len = entries[k].size - 8;
            keys->keys[k].value = xalloc(len + 1, char);
            memcpy(keys->keys[k].value, entries[k].name_space, len);
            keys->keys[k].value[len] = NUL;
        }
    }

    spSetMp4BoxContentSize((spMp4Box *)keys, size, propagate_size);
    
    return size;
}

long spWriteMp4KeyEntry(spMp4KeyEntry *key, int swap, FILE *fp)
{
    long len;
    long nwrite;
    long total_nwrite;
    
    if ((nwrite = fwriteulong32(&key->size, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if ((nwrite = (long)fwrite(key->name_space, 1, 4, fp)) != 4) {
        return nwrite;
    }
    total_nwrite += 4;

    len = key->size - 8;
    
    if ((nwrite = (long)fwrite(key->value, 1, len, fp)) != len) {
        return nwrite;
    }
    total_nwrite += nwrite;

    return total_nwrite;
}

spMp4BoxSize spWriteMp4KeysBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long k;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4KeysBox *keys = (spMp4KeysBox *)box;

    spDebug(10, "spWriteMp4KeysBox", "entry_count = %lu\n", keys->entry_count);
    if ((nwrite = fwriteulong32(&keys->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;

    for (k = 0; k < keys->entry_count; k++) {
        if ((nwrite = spWriteMp4KeyEntry(&keys->keys[k], swap, fp)) < 8) {
            spDebug(10, "spWriteMp4KeysBox", "spWriteMp4KeyEntry failed: nwrite = %ld\n", nwrite);
            return nwrite;
        }
        total_nwrite += nwrite;
    }

    spDebug(10, "spWriteMp4KeysBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

long spCopyMp4KeyEntry(spMp4KeyEntry *dest_key, spMp4KeyEntry *src_key)
{
    long len;
    long total_ncopy;

    total_ncopy = spOffsetOf(spMp4KeyEntry, value);
    memcpy(dest_key, src_key, total_ncopy);

    len = src_key->size - 8;
    dest_key->value = xalloc(len + 1, char);

    memcpy(dest_key->value, src_key->value, len);
    dest_key->value[len] = NUL;
    total_ncopy += len;

    return total_ncopy;
}

spMp4BoxSize spCopyMp4KeysBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long k;
    long ncopy;
    spMp4BoxSize total_ncopy;
    spMp4KeysBox *dest_keys = (spMp4KeysBox *)dest_box;
    spMp4KeysBox *src_keys = (spMp4KeysBox *)src_box;

    dest_keys->entry_count = src_keys->entry_count;
    total_ncopy = sizeof(dest_keys->entry_count);

    total_ncopy += sizeof(dest_keys->alloc_count) + sizeof(dest_keys->keys);
    
    if (dest_keys->entry_count > 0) {
        dest_keys->alloc_count = src_keys->alloc_count;
        dest_keys->keys = xalloc(dest_keys->alloc_count, spMp4KeyEntry);
    
        for (k = 0; k < dest_keys->entry_count; k++) {
            ncopy = spCopyMp4KeyEntry(&dest_keys->keys[k], &src_keys->keys[k]);
            total_ncopy += ncopy;
        }
    } else {
        dest_keys->alloc_count = 0;
        dest_keys->keys = NULL;
    }

    return total_ncopy;
}

/* 'hdlr' */
spMp4BoxSize spReadMp4HandlerReferenceBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    long name_len;
    spMp4HandlerReferenceBox *hdlr = (spMp4HandlerReferenceBox *)box;

    if ((nread = freadulong32(&hdlr->pre_defined, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    
    if ((nread = (long)fread(hdlr->handler_type, 1, 4, fp)) != 4) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4HandlerReferenceBox", "handler_type = %c%c%c%c\n",
            hdlr->handler_type[0], hdlr->handler_type[1], hdlr->handler_type[2], hdlr->handler_type[3]);
    
    if ((nread = freadulong32(hdlr->reserved, 3, swap, fp)) != 3) {
        return nread;
    }
    total_nread += 4 * 3;
    spDebug(50, "spReadMp4HandlerReferenceBox", "reserved = %lu-%lu-%lu\n",
            hdlr->reserved[0], hdlr->reserved[1], hdlr->reserved[2]);

    name_len = (long)(spGetMp4BoxHeaderContentSize(&hdlr->header) - total_nread);
    hdlr->name = xalloc(name_len + 1, char); /* +1: just in case */
    spDebug(50, "spReadMp4HandlerReferenceBox", "name_len = %ld\n", name_len);

    if ((nread = (long)fread(hdlr->name, 1, name_len, fp)) != name_len) {
        return nread;
    }
    hdlr->name[nread] = NUL;        /* just in case */
    spDebug(50, "spReadMp4HandlerReferenceBox", "name = '%s'\n", hdlr->name);
    total_nread += nread;

    spDebug(50, "spReadMp4HandlerReferenceBox", "total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}
    
spBool spFreeMp4HandlerReferenceBox(void *box)
{
    spMp4HandlerReferenceBox *hdlr = (spMp4HandlerReferenceBox *)box;

    if (hdlr->name != NULL) {
        xfree(hdlr->name);
        hdlr->name = NULL;
    }
    
    return SP_TRUE;
}
    
spMp4BoxSize spUpdateMp4HandlerReferenceBox(spMp4HandlerReferenceBox *hdlr, const char *handler_type, const char *name, spBool propagate_size)
{
    long name_len;
    
    memset(&hdlr->pre_defined, 0, sizeof(hdlr->pre_defined));
    memcpy(hdlr->handler_type, handler_type, 4);

    memset(hdlr->reserved, 0, sizeof(hdlr->reserved));

    if (hdlr->name != NULL) xfree(hdlr->name);
    
    if (name == NULL) {
        name_len = 2;
        hdlr->name = xalloc(2, char);
        hdlr->name[0] = NUL;
        hdlr->name[1] = NUL;
    } else {
        name_len = (long)strlen(name);
        hdlr->name = strclone(name);
    }
    
    spDebug(50, "spUpdateMp4HandlerReferenceBox", "name_len = %ld\n", name_len);

    spSetMp4BoxContentSize((spMp4Box *)hdlr, name_len + 20, propagate_size);
    
    return name_len + 20;
}

spMp4BoxSize spWriteMp4HandlerReferenceBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    long name_len;
    spMp4HandlerReferenceBox *hdlr = (spMp4HandlerReferenceBox *)box;

    if ((nwrite = fwriteulong32(&hdlr->pre_defined, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    spDebug(50, "spWriteMp4HandlerReferenceBox", "handler_type = %c%c%c%c\n",
            hdlr->handler_type[0], hdlr->handler_type[1], hdlr->handler_type[2], hdlr->handler_type[3]);
    if ((nwrite = (long)fwrite(hdlr->handler_type, 1, 4, fp)) != 4) {
        return nwrite;
    }
    total_nwrite += 4;
    
    if ((nwrite = fwriteulong32(hdlr->reserved, 3, swap, fp)) != 3) {
        return nwrite;
    }
    total_nwrite += 4 * 3;

    name_len = (long)(spGetMp4BoxHeaderContentSize(&hdlr->header) - total_nwrite);
    spDebug(50, "spWriteMp4HandlerReferenceBox", "name_len = %ld\n", name_len);

    if ((nwrite = (long)fwrite(hdlr->name, 1, name_len, fp)) != name_len) {
        return nwrite;
    }
    total_nwrite += nwrite;

    spDebug(50, "spWriteMp4HandlerReferenceBox", "total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4HandlerReferenceBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    spMp4BoxSize total_ncopy;
    long name_len;
    spMp4HandlerReferenceBox *dest_hdlr = (spMp4HandlerReferenceBox *)dest_box;
    spMp4HandlerReferenceBox *src_hdlr = (spMp4HandlerReferenceBox *)src_box;

    dest_hdlr->pre_defined = src_hdlr->pre_defined;
    total_ncopy = sizeof(dest_hdlr->pre_defined);
    
    memcpy(dest_hdlr->handler_type, src_hdlr->handler_type, sizeof(dest_hdlr->handler_type));
    total_ncopy += sizeof(dest_hdlr->handler_type);
    
    memcpy(dest_hdlr->reserved, src_hdlr->reserved, sizeof(dest_hdlr->reserved));
    total_ncopy += sizeof(dest_hdlr->reserved);

    name_len = (long)spGetMp4BoxHeaderContentSize(&src_hdlr->header) - (4 + 4 + 4 * 3);
    dest_hdlr->name = xalloc(name_len + 1, char); /* +1: just in case */

    memcpy(dest_hdlr->name, src_hdlr->name, name_len);
    dest_hdlr->name[name_len] = NUL;        /* just in case */
    total_ncopy += name_len;

    return total_ncopy;
}

/* 'mdhd' */
spMp4BoxSize spReadMp4MediaHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4MediaHeaderBox *mdhd = (spMp4MediaHeaderBox *)box;

    if (mdhd->header.version == 1) {
        if ((nread = (long)fread(mdhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = (long)fread(mdhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = freadulong32(&mdhd->u.version1.timescale, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = (long)fread(mdhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nread;
        }
        total_nread = 28;
    } else {
        if ((nread = freadulong32(&mdhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mdhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mdhd->u.version0.timescale, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mdhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread = 16;
        spDebug(50, "spReadMp4MediaHeaderBox", "creation_time = %lu, modification_time = %lu, timescale = %lu, duration = %lu\n",
                mdhd->u.version0.creation_time, mdhd->u.version0.modification_time, mdhd->u.version0.timescale, mdhd->u.version0.duration);
    }

    if ((nread = freadshort((short *)&mdhd->language, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
#if 0
    {
        char lang[4];
        spGetMp4LangStr(mdhd->language, lang, 4);
        spDebug(50, "spReadMp4MediaHeaderBox", "language = %s\n", lang);
    }
#endif
    
    if ((nread = freadshort((short *)&mdhd->quality, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    
    spDebug(50, "spReadMp4MediaHeaderBox", "total_nread = %lu, language = %d, quality = %d\n",
            (unsigned long)total_nread, mdhd->language, mdhd->quality);
    
    return total_nread;
}

spMp4BoxSize spWriteMp4MediaHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4MediaHeaderBox *mdhd = (spMp4MediaHeaderBox *)box;

    if (mdhd->header.version == 1) {
        if ((nwrite = (long)fwrite(mdhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(mdhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mdhd->u.version1.timescale, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(mdhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nwrite;
        }
        total_nwrite = 28;
    } else {
        if ((nwrite = fwriteulong32(&mdhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mdhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mdhd->u.version0.timescale, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mdhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite = 16;
    }

    if ((nwrite = fwriteshort((short *)&mdhd->language, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    if ((nwrite = fwriteshort((short *)&mdhd->quality, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    return total_nwrite;
}

spBool spInitMp4MediaHeaderBox(spChunk *parent, void *box, unsigned long options)
{
    spMp4MediaHeaderBox *mdhd = (spMp4MediaHeaderBox *)box;

    if (mdhd->header.version == 1) {
        spGetMp4TimeC64(time(NULL), mdhd->u.version1.creation_time);
        memcpy(mdhd->u.version1.creation_time, mdhd->u.version1.modification_time, 8);
    } else {
        mdhd->u.version0.creation_time = spGetMp4Time(time(NULL));
        mdhd->u.version0.modification_time = mdhd->u.version0.creation_time;
    }
    
    mdhd->language = spGetMp4LangCode("und"); /* undefined */
    
    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4MediaHeaderBoxSize(void *box)
{
    spMp4MediaHeaderBox *mdhd = (spMp4MediaHeaderBox *)box;
    
    if (mdhd->header.version == 1) {
        return 32;
    } else {
        return 20;
    }
}

spMp4BoxSize spUpdateMp4MediaHeaderBoxV0(spMp4MediaHeaderBox *mdhd,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long timescale, unsigned long duration,
                                         unsigned short language, unsigned short quality)
{
    mdhd->header.version = 0;
    if (creation_time > 0) {
        mdhd->u.version0.creation_time = (unsigned long)creation_time;
    }
    if (modification_time > 0) {
        mdhd->u.version0.modification_time = (unsigned long)modification_time;
    }
    if (timescale > 0) {
        mdhd->u.version0.timescale = timescale;
    }
    if (duration > 0) {
        mdhd->u.version0.duration = duration;
    }
    
    if (language > 0) {
        mdhd->language = language;
    }
    if (quality > 0) {
        mdhd->quality = quality;
    }
    
    return 20;
}

#ifdef SP_SUPPORT_64BIT_INT
spMp4BoxSize spUpdateMp4MediaHeaderBoxV1(spMp4MediaHeaderBox *mdhd,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long timescale, spLong64 duration,
                                         unsigned short language, unsigned short quality)
{
    mdhd->header.version = 1;
    if (creation_time > 0) {
        spLong64ToC64(creation_time, mdhd->u.version1.creation_time, SP_AIFF_NEED_SWAP);
    }
    if (modification_time > 0) {
        spLong64ToC64(modification_time, mdhd->u.version1.modification_time, SP_AIFF_NEED_SWAP);
    }
    if (timescale > 0) {
        mdhd->u.version1.timescale = timescale;
    }
    if (duration > 0) {
        spLong64ToC64(duration, mdhd->u.version1.duration, SP_AIFF_NEED_SWAP);
    }
    
    if (language > 0) {
        mdhd->language = language;
    }
    if (quality > 0) {
        mdhd->quality = quality;
    }
    
    return 32;
}
#endif

/* 'dref' */
spMp4BoxSize spReadMp4DataReferenceBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4BoxSize content_size;
    spMp4BoxSize remain_size;
    spMp4DataReferenceBox *dref = (spMp4DataReferenceBox *)box;

    if ((nread = freadulong32(&dref->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(10, "spReadMp4DataReferenceBox", "entry_count = %lu\n", dref->entry_count);

    content_size = spGetMp4BoxContentSize((spMp4Box *)dref, NULL);
    remain_size = content_size - total_nread;
    
    if (remain_size >= 8) {
        if ((nread = (long)spReadMp4ChildBox((spChunk *)box, dref->entry_count, remain_size, depth + 1, swap, fp)) <= 0) {
            return nread;
        }
        spDebug(10, "spReadMp4DataReferenceBox", "spReadMp4ChildBox result = %ld\n", nread);
        total_nread += nread;

        remain_size = content_size - total_nread;
        if (remain_size > 0) {
            spSeekFile(fp, remain_size, SEEK_CUR);
        }
    }

    spDebug(10, "spReadMp4DataReferenceBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spMp4Box *spAppendMp4DataReferenceEntry(spMp4DataReferenceBox *dref, const char *type, const char *data, long size)
{
    spMp4DataOnlyBox *dobox;

    dref->entry_count++;
    dobox = (spMp4DataOnlyBox *)spCreateMp4Box((spMp4Box *)dref, type);
    dobox->header.flags[2] = 1;        /* self containing */
    spUpdateMp4DataOnlyBox(dobox, data, size, SP_TRUE);

    return (spMp4Box *)dobox;
}

spMp4BoxSize spWriteMp4DataReferenceBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4DataReferenceBox *dref = (spMp4DataReferenceBox *)box;

    spDebug(10, "spWriteMp4DataReferenceBox", "entry_count = %lu\n", dref->entry_count);
    if ((nwrite = fwriteulong32(&dref->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;

    if (dref->entry_count > 0) {
        if ((nwrite = (long)spWriteMp4ChildOnlyBox((spChunk *)box, depth + 1, rewrite_flag, swap, fp)) <= 0) {
            return nwrite;
        }
        spDebug(10, "spWriteMp4DataReferenceBox", "spWriteMp4ChildOnlyBox result = %ld\n", nwrite);
        total_nwrite += nwrite;
    }

    spDebug(10, "spWriteMp4DataReferenceBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4DataReferenceBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    long ncopy;
    spMp4BoxSize total_ncopy;
    spMp4DataReferenceBox *dest_dref = (spMp4DataReferenceBox *)dest_box;
    spMp4DataReferenceBox *src_dref = (spMp4DataReferenceBox *)src_box;

    dest_dref->entry_count = src_dref->entry_count;
    total_ncopy = sizeof(dest_dref->entry_count);

    if ((ncopy = (long)spCopyMp4ChildBox((spChunk *)dest_dref, (spChunk *)src_dref, depth + 1)) > 0) {
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

/* 'elst' */
long spReadMp4EditListEntryV0(spMp4EditListEntryV0 *entry, int swap, FILE *fp)
{
    long nread;
    long total_nread;
    
    if ((nread = freadulong32(&entry->segment_duration, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    
    if ((nread = freadlong32(&entry->media_time, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    if ((nread = freadulong32(&entry->media_rate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    spDebug(100, "spReadMp4EditListEntryV0", "segment_duration = %lu, media_time = %ld, media_rate = %f\n",
            entry->segment_duration, entry->media_time, (double)entry->media_rate / 65536.0);
    
    return total_nread;
}

long spReadMp4EditListEntryV1(spMp4EditListEntryV1 *entry, int swap, FILE *fp)
{
    long nread;
    long total_nread;
    
#ifdef SP_SUPPORT_64BIT_INT
    if ((nread = freadlong64((spLong64 *)&entry->segment_duration, 1, swap, fp)) != 1) {
        return nread;
    }
#else
    if ((nread = (long)fread(entry->segment_duration, 1, 8, fp)) != 8) {
        return nread;
    }
#endif
    total_nread = 8;
    
#ifdef SP_SUPPORT_64BIT_INT
    if ((nread = freadlong64(&entry->media_time, 1, swap, fp)) != 1) {
        return nread;
    }
#else
    if ((nread = (long)fread(&entry->media_time, 1, 8, fp)) != 8) {
        return nread;
    }
#endif
    total_nread += 8;

    if ((nread = freadulong32(&entry->media_rate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;

    return total_nread;
}

spMp4BoxSize spReadMp4EditListBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    unsigned long k;
    long nread;
    spMp4BoxSize total_nread;
    spMp4EditListBox *elst = (spMp4EditListBox *)box;

    if ((nread = freadulong32(&elst->entry_count, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread = 4;
    spDebug(10, "spReadMp4EditListBox", "entry_count = %lu\n", elst->entry_count);

    elst->alloc_count = (1 + elst->entry_count / SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT) * SP_MP4_BOX_MIN_ENTRY_ALLOC_COUNT;
    
    if (elst->header.version == 1) {
        elst->u.edit_list_table_v1 = xalloc(elst->alloc_count, spMp4EditListEntryV1);
    } else {
        elst->u.edit_list_table_v0 = xalloc(elst->alloc_count, spMp4EditListEntryV0);
    }
    
    for (k = 0; k < elst->entry_count; k++) {
        if (elst->header.version == 1) {
            if ((nread = spReadMp4EditListEntryV1(&elst->u.edit_list_table_v1[k], swap, fp)) < 20) {
                spDebug(10, "spReadMp4EditListBox", "spReadMp4EditListEntryV0 failed: nread = %ld\n", nread);
                return nread;
            }
        } else {
            if ((nread = spReadMp4EditListEntryV0(&elst->u.edit_list_table_v0[k], swap, fp)) < 12) {
                spDebug(10, "spReadMp4EditListBox", "spReadMp4EditListEntryV0 failed: nread = %ld\n", nread);
                return nread;
            }
        }
        total_nread += nread;
    }

    spDebug(10, "spReadMp4EditListBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spFreeMp4EditListBox(void *box)
{
    spMp4EditListBox *elst = (spMp4EditListBox *)box;

    if (elst->header.version == 1) {
        if (elst->u.edit_list_table_v1 != NULL) {
            xfree(elst->u.edit_list_table_v1);
            elst->u.edit_list_table_v1 = NULL;
        }
    } else {
        if (elst->u.edit_list_table_v0 != NULL) {
            xfree(elst->u.edit_list_table_v0);
            elst->u.edit_list_table_v0 = NULL;
        }
    }
    
    return SP_TRUE;
}

long spWriteMp4EditListEntryV0(spMp4EditListEntryV0 *entry, int swap, FILE *fp)
{
    long nwrite;
    long total_nwrite;
    
    spDebug(100, "spWriteMp4EditListEntryV0", "segment_duration = %lu, media_time = %ld, media_rate = %f\n",
            entry->segment_duration, entry->media_time, (double)entry->media_rate / 65536.0);
    
    if ((nwrite = fwriteulong32(&entry->segment_duration, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    
    if ((nwrite = fwritelong32(&entry->media_time, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    if ((nwrite = fwriteulong32(&entry->media_rate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    return total_nwrite;
}

long spWriteMp4EditListEntryV1(spMp4EditListEntryV1 *entry, int swap, FILE *fp)
{
    long nwrite;
    long total_nwrite;
    
#ifdef SP_SUPPORT_64BIT_INT
    if ((nwrite = fwritelong64((spLong64 *)&entry->segment_duration, 1, swap, fp)) != 1) {
        return nwrite;
    }
#else
    if ((nwrite = (long)fwrite(entry->segment_duration, 1, 8, fp)) != 8) {
        return nwrite;
    }
#endif
    total_nwrite = 8;
    
#ifdef SP_SUPPORT_64BIT_INT
    if ((nwrite = fwritelong64(&entry->media_time, 1, swap, fp)) != 1) {
        return nwrite;
    }
#else
    if ((nwrite = (long)fwrite(&entry->media_time, 1, 8, fp)) != 8) {
        return nwrite;
    }
#endif
    total_nwrite += 8;

    if ((nwrite = fwriteulong32(&entry->media_rate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;

    return total_nwrite;
}

spMp4BoxSize spWriteMp4EditListBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    unsigned long k;
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4EditListBox *elst = (spMp4EditListBox *)box;

    if ((nwrite = fwriteulong32(&elst->entry_count, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite = 4;
    spDebug(10, "spWriteMp4EditListBox", "entry_count = %lu\n", elst->entry_count);

    for (k = 0; k < elst->entry_count; k++) {
        if (elst->header.version == 1) {
            if ((nwrite = spWriteMp4EditListEntryV1(&elst->u.edit_list_table_v1[k], swap, fp)) < 20) {
                spDebug(10, "spWriteMp4EditListBox", "spWriteMp4EditListEntryV0 failed: nwrite = %ld\n", nwrite);
                return nwrite;
            }
        } else {
            if ((nwrite = spWriteMp4EditListEntryV0(&elst->u.edit_list_table_v0[k], swap, fp)) < 12) {
                spDebug(10, "spWriteMp4EditListBox", "spWriteMp4EditListEntryV0 failed: nwrite = %ld\n", nwrite);
                return nwrite;
            }
        }
        total_nwrite += nwrite;
    }

    spDebug(10, "spWriteMp4EditListBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spCopyMp4EditListBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    unsigned long k;
    long ncopy;
    spMp4BoxSize total_ncopy;
    spMp4EditListBox *dest_elst = (spMp4EditListBox *)dest_box;
    spMp4EditListBox *src_elst = (spMp4EditListBox *)src_box;

    dest_elst->entry_count = src_elst->entry_count;
    total_ncopy = sizeof(dest_elst->entry_count);

    dest_elst->alloc_count = src_elst->alloc_count;
    total_ncopy += sizeof(dest_elst->alloc_count);
    
    if (src_elst->header.version == 1) {
        dest_elst->u.edit_list_table_v1 = xalloc(dest_elst->alloc_count, spMp4EditListEntryV1);
        total_ncopy += sizeof(dest_elst->u.edit_list_table_v1);
        ncopy = sizeof(spMp4EditListEntryV1);
    } else {
        dest_elst->u.edit_list_table_v0 = xalloc(dest_elst->alloc_count, spMp4EditListEntryV0);
        total_ncopy += sizeof(dest_elst->u.edit_list_table_v0);
        ncopy = sizeof(spMp4EditListEntryV0);
    }
    
    for (k = 0; k < dest_elst->entry_count; k++) {
        if (dest_elst->header.version == 1) {
            memcpy(&dest_elst->u.edit_list_table_v1[k], &src_elst->u.edit_list_table_v1[k], ncopy);
        } else {
            memcpy(&dest_elst->u.edit_list_table_v0[k], &src_elst->u.edit_list_table_v0[k], ncopy);
        }
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

/* 'tref' */
spMp4BoxSize spReadMp4TrackReferenceBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    long count;
    spMp4TrackReferenceBox *tref = (spMp4TrackReferenceBox *)box;

    if ((nread = (long)spReadMp4BoxHeader((spMp4Box *)tref, &tref->reference_type, swap, fp)) <= 0) {
        return nread;
    }
    total_nread = nread;

    count = (long)spGetMp4BoxHeaderContentSize(&tref->reference_type) / 4;

    tref->track_ids = xalloc(count, unsigned long);
    if ((nread = freadulong32(tref->track_ids, count, swap, fp)) != count) {
        return nread;
    }
    total_nread += 4 * count;

    return total_nread;
}

spBool spFreeMp4TrackReferenceBox(void *box)
{
    spMp4TrackReferenceBox *tref = (spMp4TrackReferenceBox *)box;

    if (tref->track_ids != NULL) {
        xfree(tref->track_ids);
        tref->track_ids = NULL;
    }

    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4TrackReferenceBox(spMp4TrackReferenceBox *tref, const char *reference_type,
                                          unsigned long *track_ids, long count, spBool propagate_size)
{
    long size;
    
    memset(&tref->reference_type, 0, sizeof(spMp4BoxHeader));
    memcpy(tref->reference_type.type, reference_type, 4);

    if (tref->track_ids != NULL) xfree(tref->track_ids);

    size = 4;
    
    if (track_ids == NULL || count <= 0) {
        count = 0;
        tref->track_ids = NULL;
    } else {
        tref->track_ids = xalloc(count, unsigned long);
        memcpy(tref->track_ids, track_ids, sizeof(unsigned long) * count);
        size += count * 4;
    }
    spDebug(50, "spUpdateMp4TrackReferenceBox", "size = %ld\n", size);

    spSetMp4BoxHeaderContentSize(&tref->reference_type, size);
    
    spSetMp4BoxContentSize((spMp4Box *)tref, spGetMp4BoxHeaderBoxSize(&tref->reference_type), propagate_size);
    
    return spGetMp4BoxHeaderContentSize(&tref->header);
}

spMp4BoxSize spWriteMp4TrackReferenceBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    long count;
    spMp4TrackReferenceBox *tref = (spMp4TrackReferenceBox *)box;

    if ((nwrite = (long)spWriteMp4BoxHeader((spMp4Box *)tref, &tref->reference_type, swap, fp)) <= 0) {
        return nwrite;
    }
    total_nwrite = nwrite;

    count = (long)spGetMp4BoxHeaderContentSize(&tref->reference_type) / 4;

    if ((nwrite = fwriteulong32(tref->track_ids, count, swap, fp)) != count) {
        return nwrite;
    }
    total_nwrite += 4 * count;

    return total_nwrite;
}

spMp4BoxSize spCopyMp4TrackReferenceBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    long ncopy;
    spMp4BoxSize total_ncopy;
    long count;
    spMp4TrackReferenceBox *dest_tref = (spMp4TrackReferenceBox *)dest_box;
    spMp4TrackReferenceBox *src_tref = (spMp4TrackReferenceBox *)src_box;

    if ((ncopy = (long)spCopyMp4BoxHeader(&dest_tref->reference_type, &src_tref->reference_type)) <= 0) {
        return ncopy;
    }
    total_ncopy = ncopy;

    count = (long)spGetMp4BoxHeaderContentSize(&src_tref->reference_type) / 4;

    dest_tref->track_ids = xalloc(count, unsigned long);
    ncopy = sizeof(long) * count;
    memcpy(dest_tref->track_ids, src_tref->track_ids, ncopy);
    total_ncopy += ncopy;

    return total_ncopy;
}

/* 'tkhd' */
spMp4BoxSize spReadMp4TrackHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4TrackHeaderBox *tkhd = (spMp4TrackHeaderBox *)box;

    spDebug(50, "spReadMp4TrackHeaderBox", "version = %d, flag = %d\n",
            tkhd->header.version, tkhd->header.flags[2]);
    
    if (tkhd->header.version == 1) {
        if ((nread = (long)fread(tkhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = (long)fread(tkhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version1.track_id, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version1.reserved, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = (long)fread(tkhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nread;
        }
        total_nread = 32;
    } else {
        if ((nread = freadulong32(&tkhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version0.track_id, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version0.reserved, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&tkhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread = 20;
        spDebug(50, "spReadMp4TrackHeaderBox", "creation_time = %lu, modification_time = %lu, track_id = %lu, duration = %lu, reserved = %lu\n",
                tkhd->u.version0.creation_time, tkhd->u.version0.modification_time, tkhd->u.version0.track_id, tkhd->u.version0.duration, tkhd->u.version0.reserved);
    }
    
    if ((nread = freadulong32(tkhd->reserved2, 2, swap, fp)) != 2) {
        return nread;
    }
    total_nread += 4 * 2;
    if ((nread = freadshort((short *)&tkhd->layer, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    if ((nread = freadshort((short *)&tkhd->alternate_group, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    if ((nread = freadshort((short *)&tkhd->volume, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4TrackHeaderBox", "layer = %d, alternate_group = %d, volume = %d, reserved2 = %lu-%lu\n",
            tkhd->layer, tkhd->alternate_group, tkhd->volume, tkhd->reserved2[0], tkhd->reserved2[1]);
    
    if ((nread = freadshort((short *)&tkhd->reserved3, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    if ((nread = freadlong32(tkhd->matrix, 9, swap, fp)) != 9) {
        return nread;
    }
    total_nread += 4 * 9;
    if ((nread = freadulong32(&tkhd->width, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    if ((nread = freadulong32(&tkhd->height, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    
    spDebug(50, "spReadMp4TrackHeaderBox", "width = %lu, height = %lu, reserved3 = %d\n",
            tkhd->width, tkhd->height, tkhd->reserved3);
    
    spDebug(10, "spReadMp4TrackHeaderBox", "total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spInitMp4TrackHeaderBox(spChunk *parent, void *box, unsigned long options)
{
    spMp4TrackHeaderBox *tkhd = (spMp4TrackHeaderBox *)box;

    if (tkhd->header.version == 1) {
        spGetMp4TimeC64(time(NULL), tkhd->u.version1.creation_time);
        memcpy(tkhd->u.version1.creation_time, tkhd->u.version1.modification_time, 8);
    } else {
        tkhd->u.version0.creation_time = spGetMp4Time(time(NULL));
        tkhd->u.version0.modification_time = tkhd->u.version0.creation_time;
    }
    
    tkhd->volume = 0x0100;
    memcpy(tkhd->matrix, sp_mp4_default_matrix, sizeof(long) * 9);

    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4TrackHeaderBoxSize(void *box)
{
    spMp4TrackHeaderBox *tkhd = (spMp4TrackHeaderBox *)box;
    
    if (tkhd->header.version == 1) {
        return 92;
    } else {
        return 80;
    }
}

spMp4BoxSize spUpdateMp4TrackHeaderBoxV0(spMp4TrackHeaderBox *tkhd, unsigned long flags,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long track_id, unsigned long duration,
                                         short layer, short alternate_group, short volume, long matrix[9],
                                         unsigned long width, unsigned long height)
{
    tkhd->header.version = 0;
    tkhd->header.flags[2] = (char)(0xff & flags);
    
    if (creation_time > 0) {
        tkhd->u.version0.creation_time = (unsigned long)creation_time;
    }
    if (modification_time > 0) {
        tkhd->u.version0.modification_time = (unsigned long)modification_time;
    }
    if (track_id > 0) {
        tkhd->u.version0.track_id = track_id;
    }
    tkhd->u.version0.reserved = 0;
    if (duration > 0) {
        tkhd->u.version0.duration = duration;
    }
    
    if (layer > 0) {
        tkhd->layer = layer;
    }
    if (alternate_group > 0) {
        tkhd->alternate_group = alternate_group;
    }
    if (volume > 0) {
        tkhd->volume = volume;
    }
    if (matrix != NULL) {
        memcpy(tkhd->matrix, matrix, sizeof(long) * 9);
    }
    if (width > 0) {
        tkhd->width = width;
    }
    if (height > 0) {
        tkhd->height = height;
    }
    
    return 80;
}

#ifdef SP_SUPPORT_64BIT_INT
spMp4BoxSize spUpdateMp4TrackHeaderBoxV1(spMp4TrackHeaderBox *tkhd, unsigned long flags,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long track_id, spLong64 duration,
                                         short layer, short alternate_group, short volume, long matrix[9],
                                         unsigned long width, unsigned long height)
{
    tkhd->header.version = 1;
    tkhd->header.flags[2] = (char)(0xff & flags);
    
    if (creation_time > 0) {
        spLong64ToC64(creation_time, tkhd->u.version1.creation_time, SP_AIFF_NEED_SWAP);
    }
    if (modification_time > 0) {
        spLong64ToC64(modification_time, tkhd->u.version1.modification_time, SP_AIFF_NEED_SWAP);
    }
    if (track_id > 0) {
        tkhd->u.version1.track_id = track_id;
    }
    tkhd->u.version1.reserved = 0;
    if (duration > 0) {
        spLong64ToC64(duration, tkhd->u.version1.duration, SP_AIFF_NEED_SWAP);
    }
    
    if (layer > 0) {
        tkhd->layer = layer;
    }
    if (alternate_group > 0) {
        tkhd->alternate_group = alternate_group;
    }
    if (volume > 0) {
        tkhd->volume = volume;
    }
    if (matrix != NULL) {
        memcpy(tkhd->matrix, matrix, sizeof(long) * 9);
    }
    if (width > 0) {
        tkhd->width = width;
    }
    if (height > 0) {
        tkhd->height = height;
    }
    
    return 92;
}
#endif

spMp4BoxSize spWriteMp4TrackHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4TrackHeaderBox *tkhd = (spMp4TrackHeaderBox *)box;

    spDebug(50, "spWriteMp4TrackHeaderBox", "version = %d\n", tkhd->header.version);
    
    if (tkhd->header.version == 1) {
        if ((nwrite = (long)fwrite(tkhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(tkhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version1.track_id, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version1.reserved, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(tkhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nwrite;
        }
        total_nwrite = 32;
    } else {
        spDebug(50, "spWriteMp4TrackHeaderBox", "track_id = %lu, duration = %lu\n",
                tkhd->u.version0.track_id, tkhd->u.version0.duration);
        if ((nwrite = fwriteulong32(&tkhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version0.track_id, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version0.reserved, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&tkhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite = 20;
    }
    
    spDebug(50, "spWriteMp4TrackHeaderBox", "layer = %d, alternate_group = %d, volume = %d\n",
            tkhd->layer, tkhd->alternate_group, tkhd->volume);
    
    if ((nwrite = fwriteulong32(tkhd->reserved2, 2, swap, fp)) != 2) {
        return nwrite;
    }
    total_nwrite += 4 * 2;
    if ((nwrite = fwriteshort((short *)&tkhd->layer, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    if ((nwrite = fwriteshort((short *)&tkhd->alternate_group, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    if ((nwrite = fwriteshort((short *)&tkhd->volume, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    
    spDebug(50, "spWriteMp4TrackHeaderBox", "width = %lu, height = %lu\n",
            tkhd->width, tkhd->height);
    
    if ((nwrite = fwriteshort((short *)&tkhd->reserved3, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    if ((nwrite = fwritelong32(tkhd->matrix, 9, swap, fp)) != 9) {
        return nwrite;
    }
    total_nwrite += 4 * 9;
    if ((nwrite = fwriteulong32(&tkhd->width, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    if ((nwrite = fwriteulong32(&tkhd->height, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    spDebug(10, "spWriteMp4TrackHeaderBox", "total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

/* 'mvhd' */
spMp4BoxSize spReadMp4MovieHeaderBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    long nread;
    spMp4BoxSize total_nread;
    spMp4MovieHeaderBox *mvhd = (spMp4MovieHeaderBox *)box;

    spDebug(50, "spReadMp4MovieHeaderBox", "version = %d\n", mvhd->header.version);
    
    if (mvhd->header.version == 1) {
        if ((nread = (long)fread(mvhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = (long)fread(mvhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nread;
        }
        if ((nread = freadulong32(&mvhd->u.version1.timescale, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = (long)fread(mvhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nread;
        }
        total_nread = 28;
    } else {
        if ((nread = freadulong32(&mvhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mvhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mvhd->u.version0.timescale, 1, swap, fp)) != 1) {
            return nread;
        }
        if ((nread = freadulong32(&mvhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nread;
        }
        total_nread = 16;
        spDebug(50, "spReadMp4MovieHeaderBox", "cretation_time = %lu, modification_time = %lu, timescale = %lu, duration = %lu\n",
                mvhd->u.version0.creation_time, mvhd->u.version0.modification_time, mvhd->u.version0.timescale, mvhd->u.version0.duration);
    }
    
    if ((nread = freadlong32(&mvhd->rate, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    if ((nread = freadshort((short *)&mvhd->volume, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    spDebug(50, "spReadMp4MovieHeaderBox", "rate = %ld, volume = %d\n",
            mvhd->rate, mvhd->volume);

    if ((nread = freadshort((short *)&mvhd->reserved, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 2;
    if ((nread = freadulong32(mvhd->reserved2, 2, swap, fp)) != 2) {
        return nread;
    }
    total_nread += 4 * 2;
    spDebug(50, "spReadMp4MovieHeaderBox", "reserved = %d, reserved2 = %lu-%lu\n",
            mvhd->reserved, mvhd->reserved2[0], mvhd->reserved2[1]);
    if ((nread = freadlong32(mvhd->matrix, 9, swap, fp)) != 9) {
        return nread;
    }
    total_nread += 4 * 9;
    if ((nread = freadulong32(mvhd->pre_defined, 6, swap, fp)) != 6) {
        return nread;
    }
    total_nread += 4 * 6;
    spDebug(50, "spReadMp4MovieHeaderBox", "pre_defined = %lu%lu%lu%lu%lu%lu\n",
            mvhd->pre_defined[0], mvhd->pre_defined[1], mvhd->pre_defined[2],
            mvhd->pre_defined[3], mvhd->pre_defined[4], mvhd->pre_defined[5]);
    if ((nread = freadulong32(&mvhd->next_track_id, 1, swap, fp)) != 1) {
        return nread;
    }
    total_nread += 4;
    spDebug(50, "spReadMp4MovieHeaderBox", "next_track_id = %lu\n", mvhd->next_track_id);
    
    spDebug(10, "spReadMp4MovieHeaderBox", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spInitMp4MovieHeaderBox(spChunk *parent, void *box, unsigned long options)
{
    spMp4MovieHeaderBox *mvhd = (spMp4MovieHeaderBox *)box;

    if (mvhd->header.version == 1) {
        spGetMp4TimeC64(time(NULL), mvhd->u.version1.creation_time);
        memcpy(mvhd->u.version1.creation_time, mvhd->u.version1.modification_time, 8);
    } else {
        mvhd->u.version0.creation_time = spGetMp4Time(time(NULL));
        mvhd->u.version0.modification_time = mvhd->u.version0.creation_time;
    }
    
    mvhd->rate = 0x00010000;
    mvhd->volume = 0x0100;
    memcpy(mvhd->matrix, sp_mp4_default_matrix, sizeof(long) * 9);
    mvhd->next_track_id = 2;

    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4MovieHeaderBoxSize(void *box)
{
    spMp4MovieHeaderBox *mvhd = (spMp4MovieHeaderBox *)box;
    
    if (mvhd->header.version == 1) {
        return 108;
    } else {
        return 96;
    }
}

spMp4BoxSize spUpdateMp4MovieHeaderBoxV0(spMp4MovieHeaderBox *mvhd,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long timescale, unsigned long duration,
                                         long rate, short volume, long matrix[9], unsigned long next_track_id)
{
    mvhd->header.version = 0;
    if (creation_time > 0) {
        mvhd->u.version0.creation_time = (unsigned long)creation_time;
    }
    if (modification_time > 0) {
        mvhd->u.version0.modification_time = (unsigned long)modification_time;
    }
    if (timescale > 0) {
        mvhd->u.version0.timescale = timescale;
    }
    if (duration > 0) {
        mvhd->u.version0.duration = duration;
    }
    
    if (rate > 0) {
        mvhd->rate = rate;
    }
    if (volume > 0) {
        mvhd->volume = volume;
    }
    if (matrix != NULL) {
        memcpy(mvhd->matrix, matrix, sizeof(long) * 9);
    }
    if (next_track_id > 0) {
        mvhd->next_track_id = next_track_id;
    }
    
    return 96;
}

#ifdef SP_SUPPORT_64BIT_INT
spMp4BoxSize spUpdateMp4MovieHeaderBoxV1(spMp4MovieHeaderBox *mvhd,
                                         time_t creation_time, time_t modification_time,
                                         unsigned long timescale, spLong64 duration,
                                         long rate, short volume, long matrix[9], unsigned long next_track_id)
{
    mvhd->header.version = 1;
    if (creation_time > 0) {
        spLong64ToC64(creation_time, mvhd->u.version1.creation_time, SP_AIFF_NEED_SWAP);
    }
    if (modification_time > 0) {
        spLong64ToC64(modification_time, mvhd->u.version1.modification_time, SP_AIFF_NEED_SWAP);
    }
    if (timescale > 0) {
        mvhd->u.version1.timescale = timescale;
    }
    if (duration > 0) {
        spLong64ToC64(duration, mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
    }
    
    if (rate > 0) {
        mvhd->rate = rate;
    }
    if (volume > 0) {
        mvhd->volume = volume;
    }
    if (matrix != NULL) {
        memcpy(mvhd->matrix, matrix, sizeof(long) * 9);
    }
    if (next_track_id > 0) {
        mvhd->next_track_id = next_track_id;
    }
    
    return 108;
}
#endif

spMp4BoxSize spWriteMp4MovieHeaderBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long nwrite;
    spMp4BoxSize total_nwrite;
    spMp4MovieHeaderBox *mvhd = (spMp4MovieHeaderBox *)box;

    spDebug(50, "spWriteMp4MovieHeaderBox", "version = %d\n", mvhd->header.version);
    
    if (mvhd->header.version == 1) {
        if ((nwrite = (long)fwrite(mvhd->u.version1.creation_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(mvhd->u.version1.modification_time, 1, 8, fp)) != 8) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mvhd->u.version1.timescale, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = (long)fwrite(mvhd->u.version1.duration, 1, 8, fp)) != 8) {
            return nwrite;
        }
        total_nwrite = 28;
    } else {
        spDebug(50, "spWriteMp4MovieHeaderBox", "timescale = %lu, duration = %lu\n",
                mvhd->u.version0.timescale, mvhd->u.version0.duration);
        if ((nwrite = fwriteulong32(&mvhd->u.version0.creation_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mvhd->u.version0.modification_time, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mvhd->u.version0.timescale, 1, swap, fp)) != 1) {
            return nwrite;
        }
        if ((nwrite = fwriteulong32(&mvhd->u.version0.duration, 1, swap, fp)) != 1) {
            return nwrite;
        }
        total_nwrite = 16;
    }
    
    spDebug(50, "spWriteMp4MovieHeaderBox", "rate = %ld, volume = %d\n",
            mvhd->rate, mvhd->volume);
    
    if ((nwrite = fwritelong32(&mvhd->rate, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    if ((nwrite = fwriteshort((short *)&mvhd->volume, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;

    if ((nwrite = fwriteshort((short *)&mvhd->reserved, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 2;
    if ((nwrite = fwriteulong32(mvhd->reserved2, 2, swap, fp)) != 2) {
        return nwrite;
    }
    total_nwrite += 4 * 2;
    if ((nwrite = fwritelong32(mvhd->matrix, 9, swap, fp)) != 9) {
        return nwrite;
    }
    total_nwrite += 4 * 9;
    if ((nwrite = fwriteulong32(mvhd->pre_defined, 6, swap, fp)) != 6) {
        return nwrite;
    }
    total_nwrite += 4 * 6;
    spDebug(50, "spWriteMp4MovieHeaderBox", "next_track_id = %lu\n", mvhd->next_track_id);
    if ((nwrite = fwriteulong32(&mvhd->next_track_id, 1, swap, fp)) != 1) {
        return nwrite;
    }
    total_nwrite += 4;
    
    spDebug(10, "spWriteMp4MovieHeaderBox", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

/* 'ftyp' */
spMp4BoxSize spReadMp4FileTypeBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    spMp4BoxSize size;
    spMp4BoxSize nread;
    spMp4FileTypeBox *ftypbox = (spMp4FileTypeBox *)box;

    spDebug(80, "spReadMp4FileTypeBox", "in\n");
    
    if (!strneq("ftyp", ftypbox->header.type, 4)) {
        spDebug(10, "spReadMp4FileTypeBox", "Can't find 'ftyp' box.\n");
        return 0;
    }
    
    if (fread(ftypbox->major_brand, 1, 4, fp) != 4) {
        spDebug(10, "spReadMp4FileTypeBox", "Can't read major brand in 'ftyp' box.\n");
        return 0;
    }
    nread = 4;
    
    if (freadlong32(&ftypbox->minor_version, 1, swap, fp) <= 0) {
        spDebug(10, "spReadMp4FileTypeBox", "Can't read minor version in 'ftyp' box.\n");
    }
    nread += 4;

    size = spGetMp4BoxHeaderContentSize(&ftypbox->header) - 8;
    if (size > 0) {
        ftypbox->nbrands = (long)(size / 4);
        ftypbox->compatible_brands = xalloc(size + 1, char);
    
        if (fread(ftypbox->compatible_brands, 1, (long)size, fp) != size) {
            spDebug(10, "spReadMp4FileTypeBox", "Can't read combatible brand in 'ftyp' box.\n");
            return 0;
        }
        ftypbox->compatible_brands[size] = NUL;
        nread += size;
    } else {
        ftypbox->nbrands = 0;
        ftypbox->compatible_brands = NULL;
    }

    spDebug(80, "spReadMp4FileTypeBox", "done: nread = %lu\n", (unsigned long)nread);
    
    return nread;
}

spBool spFreeMp4FileTypeBox(void *box)
{
    spMp4FileTypeBox *ftyp = (spMp4FileTypeBox *)box;

    if (ftyp->compatible_brands != NULL) {
        xfree(ftyp->compatible_brands);
        ftyp->compatible_brands = NULL;
    }

    return SP_TRUE;
}

spMp4BoxSize spUpdateMp4FileTypeBox(spMp4FileTypeBox *ftypbox, const char *major_brand, long minor_version,
                                    long nbrands, const char *compatible_brands)
{
    unsigned long i, len;
    spMp4BoxSize size;
    
    memcpy(ftypbox->major_brand, major_brand, 4);
    ftypbox->minor_version = minor_version;
    ftypbox->nbrands = nbrands;

    if (ftypbox->compatible_brands != NULL) xfree(ftypbox->compatible_brands);
    len = nbrands * 4;
    ftypbox->compatible_brands = xalloc(len + 1, char);

    for (i = 0; i < len; i++) {
        if (compatible_brands[i] == NUL) {
            break;
        }
        ftypbox->compatible_brands[i] = compatible_brands[i];
    }
    for (; i <= len; i++) {
        ftypbox->compatible_brands[i] = NUL;
    }
    spDebug(80, "spUpdateMp4FileTypeBox", "ftypbox->compatible_brands = %s\n", ftypbox->compatible_brands);

    size = 8 + len;
    spSetMp4BoxContentSize((spMp4Box *)ftypbox, size, SP_TRUE);

    return size;
}

spMp4BoxSize spWriteMp4FileTypeBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    spMp4BoxSize size;
    spMp4BoxSize nwrite;
    spMp4FileTypeBox *ftypbox = (spMp4FileTypeBox *)box;

    spDebug(80, "spWriteMp4FileTypeBox", "in\n");
    
    if (fwrite(ftypbox->major_brand, 1, 4, fp) != 4) {
        spDebug(10, "spWriteMp4FileTypeBox", "Can't write major brand in 'ftyp' box.\n");
        return 0;
    }
    nwrite = 4;
    
    if (fwritelong32(&ftypbox->minor_version, 1, swap, fp) <= 0) {
        spDebug(10, "spWriteMp4FileTypeBox", "Can't write minor version in 'ftyp' box.\n");
    }
    nwrite += 4;

    if (ftypbox->nbrands > 0) {
        size = ftypbox->nbrands * 4;
        if (fwrite(ftypbox->compatible_brands, 1, (long)size, fp) != size) {
            spDebug(10, "spWriteMp4FileTypeBox", "Can't write combatible brand in 'ftyp' box.\n");
            return 0;
        }
        nwrite += size;
    }

    spDebug(80, "spWriteMp4FileTypeBox", "done: nwrite = %lu\n", (unsigned long)nwrite);
    
    return nwrite;
}

spMp4BoxSize spCopyMp4FileTypeBox(spChunk *dest_box, spChunk *src_box, long depth)
{
    int offset;
    spMp4BoxSize size;
    spMp4BoxSize ncopy;
    spMp4FileTypeBox *dest_ftypbox = (spMp4FileTypeBox *)dest_box;
    spMp4FileTypeBox *src_ftypbox = (spMp4FileTypeBox *)src_box;

    offset = spOffsetOf(spMp4FileTypeBox, major_brand);

    ncopy = spOffsetOf(spMp4FileTypeBox, nbrands) - offset;
    memcpy((char *)dest_ftypbox + offset, (char *)src_ftypbox + offset, (long)ncopy);

    ncopy += sizeof(dest_ftypbox->nbrands) + sizeof(dest_ftypbox->compatible_brands);
    
    size = spGetMp4BoxHeaderContentSize(&src_ftypbox->header) - 8;
    if (size > 0) {
        dest_ftypbox->nbrands = src_ftypbox->nbrands;
        dest_ftypbox->compatible_brands = xalloc(size + 1, char);

        memcpy(dest_ftypbox->compatible_brands, src_ftypbox->compatible_brands, (long)size);
        dest_ftypbox->compatible_brands[size] = NUL;
        ncopy += size;
    } else {
        dest_ftypbox->nbrands = 0;
        dest_ftypbox->compatible_brands = NULL;
    }

    return ncopy;
}

static spChunkInfoTable sp_mp4_box_info_list[] = {
    {"free", "", SP_FALSE, SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
    {"skip", "", SP_FALSE, SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
    {"wide", "", SP_FALSE, SP_FALSE, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
        
    {"ftyp", "", SP_TRUE, SP_FALSE, spOffsetOf(spMp4Header, ftyp), sizeof(spMp4FileTypeBox), 0, spReadMp4FileTypeBox, spWriteMp4FileTypeBox, NULL, spFreeMp4FileTypeBox, SP_TRUE, NULL, spCopyMp4FileTypeBox},
    {"moov", "", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4Header, moov), sizeof(spMp4MovieBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},

    {"udta", "moov", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4MovieBox, udta), sizeof(spMp4UserDataBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"meta", "udta", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4UserDataBox, meta), sizeof(spMp4MetaBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"meta", "", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, 0, sizeof(spMp4MetaBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"hdlr", "meta", SP_TRUE, SP_TRUE, spOffsetOf(spMp4MetaBox, hdlr), sizeof(spMp4HandlerReferenceBox), 20, spReadMp4HandlerReferenceBox, spWriteMp4HandlerReferenceBox, NULL, spFreeMp4HandlerReferenceBox, SP_TRUE, NULL, spCopyMp4HandlerReferenceBox},
    {"ilst", "meta", SP_FALSE, SP_FALSE, spOffsetOf(spMp4MetaBox, ilst), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"mhdr", "meta", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MetaBox, mhdr), sizeof(spMp4ULongOnlyBox), 4, spReadMp4ULongOnlyBox, spWriteMp4ULongOnlyBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"keys", "meta", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MetaBox, keys), sizeof(spMp4KeysBox), 0, spReadMp4KeysBox, spWriteMp4KeysBox, NULL, spFreeMp4KeysBox, SP_TRUE, NULL, spCopyMp4KeysBox},
    {SP_MP4_META_DATA_TYPE_TITLE, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ARTIST, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ALBUM_ARTIST, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_COMPOSER, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ALBUM, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_GENRE, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_RELEASE, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_SOFTWARE, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_COPYRIGHT, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_COMMENT, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_TITLE_YOMI, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ARTIST_YOMI, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ALBUM_ARTIST_YOMI, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_ALBUM_YOMI, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_COMPOSER_YOMI, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_TRACK, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_DISC, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_TEMPO, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_COMPILATION, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {SP_MP4_META_DATA_TYPE_GAPLESS, "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"covr", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"\251grp", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"\251lyr", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"desc", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"tvsh", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"tven", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"tvsn", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"tves", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"apID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"cnID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"rtng", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"atID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"cmID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"plID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"geID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"sfID", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"stik", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"purd", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"xid ", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"flvr", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"----", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"itif", "ilst", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, 0, sizeof(spMp4ULongOnlyBox), 4, spReadMp4ULongOnlyBox, spWriteMp4ULongOnlyBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"data", "", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, sizeof(spMp4MetaDataValueBox), 0, spReadMp4MetaDataValueBox, spWriteMp4MetaDataValueBox, NULL, spFreeMp4MetaDataValueBox, SP_TRUE, NULL, spCopyMp4MetaDataValueBox},

    {"name", "schi", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"user", "schi", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"cert", "schi", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"righ", "schi", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"chtb", "schi", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"mean", "", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"name", "", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"\251cpy", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251inf", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251day", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251fmt", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251nam", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251dir", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251prd", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251prf", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"\251src", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4UserDataStringBox), 0, spReadMp4UserDataStringBox, spWriteMp4UserDataStringBox, NULL, spFreeMp4UserDataStringBox, SP_TRUE, NULL, spCopyMp4UserDataStringBox},
    {"WLOC", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"LOOP", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"MCPS", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"play", "udta", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"mvhd", "moov", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4MovieBox, mvhd), sizeof(spMp4MovieHeaderBox), 0, spReadMp4MovieHeaderBox, spWriteMp4MovieHeaderBox, spInitMp4MovieHeaderBox, NULL, SP_TRUE, spUpdateMp4MovieHeaderBoxSize, NULL},
    {"trak", "moov", SP_FALSE/*SP_TRUE*/, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4MovieBox, trak), sizeof(spMp4TrackBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"iods", "moov", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"esds", "mp4a", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.soun.esds), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"chan", "mp4a", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"sbtd", "mp4a", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"pinf", "mp4a", SP_FALSE, SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"schi", "pinf", SP_FALSE, SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"sign", "pinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"schm", "pinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"frma", "pinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"tkhd", "trak", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4TrackBox, tkhd), sizeof(spMp4TrackHeaderBox), 0, spReadMp4TrackHeaderBox, spWriteMp4TrackHeaderBox, spInitMp4TrackHeaderBox, NULL, SP_TRUE, spUpdateMp4TrackHeaderBoxSize, NULL},
    {"tref", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, tref), sizeof(spMp4TrackReferenceBox), 0, spReadMp4TrackReferenceBox, spWriteMp4TrackReferenceBox, NULL, spFreeMp4TrackReferenceBox, SP_TRUE, NULL, spCopyMp4TrackReferenceBox},
    {"mdia", "trak", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4TrackBox, mdia), sizeof(spMp4MediaBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"udta", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, udta), sizeof(spMp4UserDataBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"edts", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, edts), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"elst", "edts", SP_FALSE, SP_TRUE, 0, sizeof(spMp4EditListBox), 0, spReadMp4EditListBox, spWriteMp4EditListBox, NULL, spFreeMp4EditListBox, SP_TRUE, NULL, spCopyMp4EditListBox},
    {"clip", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, clip), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"crgn", "clip", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"matt", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, matt), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"kmat", "matt", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"imap", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, imap), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"  in", "imap", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"load", "trak", SP_FALSE, SP_FALSE, spOffsetOf(spMp4TrackBox, load), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"mdhd", "mdia", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4MediaBox, mdhd), sizeof(spMp4MediaHeaderBox), 0, spReadMp4MediaHeaderBox, spWriteMp4MediaHeaderBox, spInitMp4MediaHeaderBox, NULL, SP_TRUE, spUpdateMp4MediaHeaderBoxSize, NULL},
    {"hdlr", "mdia", SP_TRUE, SP_TRUE, spOffsetOf(spMp4MediaBox, hdlr), sizeof(spMp4HandlerReferenceBox), 20, spReadMp4HandlerReferenceBox, spWriteMp4HandlerReferenceBox, NULL, spFreeMp4HandlerReferenceBox, SP_TRUE, NULL, spCopyMp4HandlerReferenceBox},
    {"minf", "mdia", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4MediaBox, minf), sizeof(spMp4MediaInformationBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"udta", "mdia", SP_FALSE, SP_FALSE, spOffsetOf(spMp4MediaBox, udta), sizeof(spMp4UserDataBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"hnti", "udta", SP_FALSE, SP_FALSE, spOffsetOf(spMp4UserDataBox, hnti), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"hinf", "udta", SP_FALSE, SP_FALSE, spOffsetOf(spMp4UserDataBox, hinf), 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},

    {"rtp ", "hnti", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"sdp ", "hnti", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"sdp ", "rtp ", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"trpY", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"trpy", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"totl", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"nump", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"npck", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tpyl", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tpaY", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tpay", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"maxr", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"dmed", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"dimm", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"drep", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tmin", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tmax", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"pmax", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"dmax", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"payt", "hinf", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"vmhd", "minf", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MediaInformationBox, u.vmhd), sizeof(spMp4VideoMediaHeaderBox), 8, spReadMp4VideoMediaHeaderBox, spWriteMp4VideoMediaHeaderBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"smhd", "minf", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MediaInformationBox, u.smhd), sizeof(spMp4SoundMediaHeaderBox), 4, spReadMp4SoundMediaHeaderBox, spWriteMp4SoundMediaHeaderBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"hmhd", "minf", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MediaInformationBox, u.hmhd), sizeof(spMp4HintMediaHeaderBox), 16, spReadMp4HintMediaHeaderBox, spWriteMp4HintMediaHeaderBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"nmhd", "minf", SP_FALSE, SP_CHUNK_INFO_OPTION_WRITE_ZEROS_MASK | SP_TRUE, spOffsetOf(spMp4MediaInformationBox, u.nmhd), sizeof(spMp4NullMediaHeaderBox), 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
    {"dinf", "minf", SP_TRUE, SP_FALSE, spOffsetOf(spMp4MediaInformationBox, dinf), sizeof(spMp4DataInformationBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"stbl", "minf", SP_TRUE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, spOffsetOf(spMp4MediaInformationBox, stbl), sizeof(spMp4SampleTableBox), 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"hdlr", "minf", SP_FALSE, SP_TRUE, spOffsetOf(spMp4MediaInformationBox, hdlr), sizeof(spMp4HandlerReferenceBox), 20, spReadMp4HandlerReferenceBox, spWriteMp4HandlerReferenceBox, NULL, spFreeMp4HandlerReferenceBox, SP_TRUE, NULL, spCopyMp4HandlerReferenceBox},
     
    {"dref", "dinf", SP_TRUE, SP_TRUE, spOffsetOf(spMp4DataInformationBox, dref), sizeof(spMp4DataReferenceBox), 4, spReadMp4DataReferenceBox, spWriteMp4DataReferenceBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4DataReferenceBox},
    {"url ", "dref", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"urn ", "dref", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"alis", "dref", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"rsrc", "dref", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"cios", "dref", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"stsd", "stbl", SP_TRUE, SP_TRUE, spOffsetOf(spMp4SampleTableBox, stsd), sizeof(spMp4SampleDescriptionBox), 4, spReadMp4SampleDescriptionBox, spWriteMp4SampleDescriptionBox, NULL, spFreeMp4SampleDescriptionBox, SP_TRUE, NULL, spCopyMp4SampleDescriptionBox},
    {"esds", "mp4v", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.vide.esds), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"clap", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.vide.clap), sizeof(spMp4CleanApertureBox), 0, spReadMp4CleanApertureBox, spWriteMp4CleanApertureBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"colr", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.vide.colr), sizeof(spMp4ColorParameterBox), 0, spReadMp4ColorParameterBox, spWriteMp4ColorParameterBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"pasp", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.vide.pasp), sizeof(spMp4PixelAspectRatioBox), 0, spReadMp4PixelAspectRatioBox, spWriteMp4PixelAspectRatioBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"gama", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.vide.gama), sizeof(spMp4ULongOnlyBox), 0, spReadMp4ULongOnlyBox, spWriteMp4ULongOnlyBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"stsl", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.vide.stsl), sizeof(spMp4SampleScaleBox), 0, spReadMp4SampleScaleBox, spWriteMp4SampleScaleBox, NULL, NULL, SP_TRUE, NULL, NULL},
    {"fiel", "", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.vide.fiel), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"avcC", "", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.vide.avcC), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"uuid", "", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.vide.uuid), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"tims", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.rtp.tims), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"tsro", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.rtp.tsro), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"snro", "", SP_FALSE, SP_FALSE, spOffsetOf(spMp4SampleEntryBox, u.rtp.snro), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
        
    {"stts", "stbl", /*SP_TRUE*/SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stts), sizeof(spMp4TimeToSampleBox), 4, spReadMp4TimeToSampleBox, spWriteMp4TimeToSampleBox, NULL, spFreeMp4TimeToSampleBox, SP_TRUE, NULL, spCopyMp4TimeToSampleBox},
    {"ctts", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, ctts), sizeof(spMp4CompositionOffsetBox), 4, spReadMp4CompositionOffsetBox, spWriteMp4CompositionOffsetBox, NULL, spFreeMp4CompositionOffsetBox, SP_TRUE, NULL, spCopyMp4CompositionOffsetBox},
    {"stsc", "stbl", /*SP_TRUE*/SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stsc), sizeof(spMp4SampleToChunkBox), 4, spReadMp4SampleToChunkBox, spWriteMp4SampleToChunkBox, NULL, spFreeMp4SampleToChunkBox, SP_TRUE, NULL, spCopyMp4SampleToChunkBox},
    {"stsz", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stsz), sizeof(spMp4SampleSizeBox), 8, spReadMp4SampleSizeBox, spWriteMp4SampleSizeBox, NULL, spFreeMp4SampleSizeBox, SP_TRUE, NULL, spCopyMp4SampleSizeBox},
    {"stz2", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stsz), sizeof(spMp4SampleSizeBox), 8, spReadMp4CompactSampleSizeBox, spWriteMp4CompactSampleSizeBox, NULL, spFreeMp4CompactSampleSizeBox, SP_TRUE, NULL, spCopyMp4CompactSampleSizeBox},
    {"stco", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stco), sizeof(spMp4ChunkOffsetBox), 4, spReadMp4ChunkOffsetBox, spWriteMp4ChunkOffsetBox, NULL, spFreeMp4ChunkOffsetBox, SP_TRUE, NULL, spCopyMp4ChunkOffsetBox},
    {"co64", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stco), sizeof(spMp4ChunkOffsetBox), 4, spReadMp4ChunkOffsetBox, spWriteMp4ChunkOffsetBox, NULL, spFreeMp4ChunkOffsetBox, SP_TRUE, NULL, spCopyMp4ChunkOffsetBox},
    {"stss", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stss), sizeof(spMp4SyncSampleBox), 4, spReadMp4SyncSampleBox, spWriteMp4SyncSampleBox, NULL, spFreeMp4SyncSampleBox, SP_TRUE, NULL, spCopyMp4SyncSampleBox},
    {"stsh", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stsh), sizeof(spMp4ShadowSyncSampleBox), 4, spReadMp4ShadowSyncSampleBox, spWriteMp4ShadowSyncSampleBox, NULL, spFreeMp4ShadowSyncSampleBox, SP_TRUE, NULL, spCopyMp4ShadowSyncSampleBox},
    {"stdp", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, stdp), sizeof(spMp4DegradationPriorityBox), 4, spReadMp4DegradationPriorityBox, spWriteMp4DegradationPriorityBox, NULL, spFreeMp4DegradationPriorityBox, SP_TRUE, NULL, spCopyMp4DegradationPriorityBox},
    {"sdtp", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, spOffsetOf(spMp4SampleTableBox, sdtp), sizeof(spMp4SampleDependencyTypeBox), 4, spReadMp4SampleDependencyTypeBox, spWriteMp4SampleDependencyTypeBox, NULL, spFreeMp4SampleDependencyTypeBox, SP_TRUE, NULL, spCopyMp4SampleDependencyTypeBox},
    {"cslg", "stbl", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"wave", "", SP_FALSE, SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE, 0, 0, 0, spReadMp4ChildOnlyBox, spWriteMp4ChildOnlyBox, NULL, NULL, SP_TRUE, NULL, spCopyMp4ChildOnlyBox},
    {"esds", "wave", SP_FALSE, SP_TRUE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"mp4a", "wave", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"frma", "wave", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"\0\0\0\0", "wave", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"QDCA", "wave", SP_FALSE, SP_FALSE, 0, sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},

    {"alac", "alac", SP_FALSE, SP_TRUE, spOffsetOf(spMp4SampleEntryBox, u.soun.alac), sizeof(spMp4DataOnlyBox), 0, spReadMp4DataOnlyBox, spWriteMp4DataOnlyBox, NULL, spFreeMp4DataOnlyBox, SP_TRUE, NULL, spCopyMp4DataOnlyBox},
    {"mdat", "", /*SP_FALSE*/SP_TRUE, SP_CHUNK_INFO_OPTION_PAUSE_BEFORE_WRITE_MASK | SP_CHUNK_INFO_OPTION_REWRITE_MASK | SP_FALSE,
         spOffsetOf(spMp4Header, mdat), 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
};

spChunkFileSpec sp_mp4_file_spec = {
    sizeof(struct _spMp4BoxHeader),
    sizeof(struct _spMp4Box),
    sp_mp4_box_info_list,
    0,

    (spChunkInitFunc)spInitMp4Box,
    (spChunkHeaderGetSizeFunc)spGetMp4BoxHeaderHeaderSize,
    (spChunkHeaderGetSizeFunc)spGetMp4BoxHeaderContentSize,
    (spChunkGetSizeFunc)spGetMp4BoxHeaderSize,
    (spChunkGetSizeFunc)spGetMp4BoxContentSizeForChunkFileSpec,
    (spChunkSetSizeFunc)spSetMp4BoxContentSizeForChunkFileSpec,
    (spChunkGetSizeFunc)spGetMp4BoxContentMarginSize,
    (spChunkSetSizeFunc)spSetMp4BoxContentMarginSize,
    (spChunkReadHeaderFunc)spReadMp4BoxHeader,
    (spChunkReadHeaderRestFunc)spReadMp4FullBoxHeaderRest,
    (spChunkSkipFunc)spSkipMp4Box,

    (spChunkWriteHeaderFunc)spWriteMp4BoxHeader,
    (spChunkWriteHeaderRestFunc)spWriteMp4FullBoxHeaderRest,
    (spChunkWriteContentRestFunc)spWriteMp4ContentRest,

    NULL,
};

void spPropagateMp4BoxContentSize(spMp4Box *parent, spChunkSize size_diff)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    
    spPropagateChunkContentSize(&sp_mp4_file_spec, (spChunk *)parent, size_diff);

    return;
}

spBool spSetMp4BoxContentSize(spMp4Box *box, spMp4BoxSize size, spBool propagate_size)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    return spSetChunkContentSize(&sp_mp4_file_spec, (spChunk *)box, size, propagate_size);
}

spMp4BoxSize spReadMp4ChildBox(spChunk *parent, long max_count, spMp4BoxSize remain_size, long depth, int swap, FILE *fp)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    return spReadChildChunk(&sp_mp4_file_spec, parent, max_count, remain_size, depth, swap, fp);
}

spMp4BoxSize spCopyMp4ChildBox(spChunk *dest_parent, spChunk *src_parent, long depth)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    return spCopyChildChunk(&sp_mp4_file_spec, dest_parent, src_parent, depth);
}

spBool spFreeMp4Box(spMp4Box *box)
{
    return spFreeChunk(&sp_mp4_file_spec, (spChunk *)box);
}

spMp4BoxSize spReadMp4ChildOnlyBox(spChunk *parent, long depth, void *box, int swap, FILE *fp)
{
    return spReadMp4ChildBox((spChunk *)box, 0, 0, depth, swap, fp);
}

spMp4BoxSize spWriteMp4ChildOnlyBox(spChunk *box, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    return spWriteChildChunk(&sp_mp4_file_spec, box, depth, SP_TRUE, rewrite_flag, swap, NULL, fp);
}

spMp4BoxSize spCopyMp4ChildOnlyBox(spChunk *dest_chunk, spChunk *src_chunk, long depth)
{
    return spCopyMp4ChildBox(dest_chunk, src_chunk, depth);
}

spMp4BoxSize spReadMp4Header(spMp4Header *header, FILE *fp)
{
    spMp4BoxSize nread;
    spMp4BoxSize total_nread;

    spDebug(80, "spReadMp4Header", "in\n");
    
    memset(header, 0, sizeof(spMp4Header));

    if ((nread = spReadMp4ChildBox((spChunk *)header, 0, 0, 0, SP_AIFF_NEED_SWAP, fp)) <= 0) {
        spDebug(10, "spReadMp4Header", "spReadMp4ChildBox returns 0\n");
        spSeekFile(fp, 0, SEEK_SET);
        return nread;
    }
    total_nread = nread;

    spDebug(80, "spReadMp4Header", "done: total_nread = %lu\n", (unsigned long)total_nread);
    
    return total_nread;
}

spBool spFreeMp4Header(spMp4Header *header)
{
    spDebug(80, "spFreeMp4Header", "in\n");
    
    if (header != NULL) {
        spFreeMp4Box((spMp4Box *)header);
        
        spDebug(80, "spFreeMp4Header", "done\n");
        return SP_TRUE;
    } else {
        spDebug(80, "spFreeMp4Header", "header == NULL\n");
        return SP_FALSE;
    }
}

spBool spInitMp4HeaderWithCB(spMp4Header *header, spMp4BoxUpdateFunc update_func, void *data)
{
    spChunkSize size;
    
    if (header == NULL) return SP_FALSE;
    
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }

    memset(header, 0, sizeof(spMp4Header));

    size = spCreateChildChunk(&sp_mp4_file_spec, (spChunk *)header, SP_FALSE, SP_TRUE, (spChunkUpdateFunc)update_func, data);

    if (header->ftyp == NULL) {
        spDebug(80, "spInitMp4HeaderWithCB", "failed: ftype == NULL\n");
        return SP_FALSE;
    } else {
        spDebug(80, "spInitMp4HeaderWithCB", "done: size = %lu\n", (unsigned long)size);
        return SP_TRUE;
    }
}

spBool spInitMp4Header(spMp4Header *header, const char *major_brand, long minor_version,
                       long nbrands, const char *compatible_brands)
{
    if (spInitMp4HeaderWithCB(header, NULL, NULL) == SP_FALSE) {
        return SP_FALSE;
    }
    
    if (spUpdateMp4FileTypeBox(header->ftyp, major_brand, minor_version, nbrands, compatible_brands) > 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spMp4BoxSize spWriteMp4Header(spMp4Header *header, spMp4BoxSize moov_size, spMp4Box **paused_box, FILE *fp)
{
    spMp4BoxSize nwrite;
    spMp4BoxSize total_nwrite;
    spBool rewrite_flag = SP_FALSE;

    if (header == NULL || fp == NULL
        || /*header->ftyp == NULL ||*/ header->moov == NULL || header->mdat == NULL) {
        spDebug(10, "spWriteMp4Header", "invalid header\n");
        return 0;
    }

    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }
    
    spDebug(80, "spWriteMp4Header", "moov_size = %lu\n", (unsigned long)moov_size);
    
    if (moov_size > 0) {
        spMp4BoxSize marginsize;
        
        marginsize = moov_size - spGetMp4BoxHeaderContentSize(&header->moov->header) + header->moov->header.marginsize;
        spSetMp4BoxContentSize((spMp4Box *)header->moov, moov_size, SP_TRUE);
        spSetMp4BoxContentMarginSize((spMp4Box *)header->moov, marginsize);
    }
    
    if (paused_box == NULL || *paused_box == NULL) {
        spDebug(80, "spWriteMp4Header", "no paused_box, seek\n");
        spSeekFile(fp, 0, SEEK_SET);
    } else if (paused_box != NULL && *paused_box != NULL) {
        spDebug(80, "spWriteMp4Header", "paused_box exists, rewrite\n");
        rewrite_flag = SP_TRUE;
    }
    
    if ((nwrite = spWriteChildChunk(&sp_mp4_file_spec, (spChunk *)header, 0, SP_TRUE,
                                    SP_FALSE, SP_AIFF_NEED_SWAP, (spChunk **)paused_box, fp)) <= 0) {
        spDebug(10, "spWriteMp4Header", "first spWriteChildChunk failed: nwrite = %lu\n", (unsigned long)nwrite);
    }
    spDebug(80, "spWriteMp4Header", "first spWriteChildChunk done: nwrite = %lu\n", (unsigned long)nwrite);
    total_nwrite = nwrite;

    if (rewrite_flag == SP_TRUE && *paused_box == NULL) {
        if (spSeekFile(fp, 0, SEEK_SET) == 0) {
            if ((nwrite = spWriteChildChunk(&sp_mp4_file_spec, (spChunk *)header, 0, SP_TRUE,
                                            SP_TRUE, SP_AIFF_NEED_SWAP, NULL, fp)) <= 0) {
                spDebug(10, "spWriteMp4Header", "second spWriteChildChunk failed: nwrite = %lu\n", (unsigned long)nwrite);
                return nwrite;
            }
            spDebug(80, "spWriteMp4Header", "second spWriteChildChunk done: nwrite = %lu\n", (unsigned long)nwrite);
            total_nwrite = nwrite;
        } else {
            spDebug(10, "spWriteMp4Header", "spSeekFile failed\n");
        }
    }

    spDebug(80, "spWriteMp4Header", "done: total_nwrite = %lu\n", (unsigned long)total_nwrite);
    
    return total_nwrite;
}

spMp4BoxSize spRewriteMp4Header(spMp4Header *header, spMp4BoxSize mdat_size, spMp4Box **paused_box, FILE *fp)
{
    spMp4BoxSize moov_size;

    if (header == NULL || fp == NULL
        || /*header->ftyp == NULL ||*/ header->moov == NULL || header->mdat == NULL) return 0;

    spSetMp4BoxContentSize(header->mdat, mdat_size, SP_TRUE);
    
    moov_size = spGetMp4BoxHeaderContentSize(&header->moov->header);
    
    return spWriteMp4Header(header, moov_size, paused_box, fp);
}

spMp4BoxSize spCopyMp4Header(spMp4Header *dest_header, spMp4Header *src_header)
{
    spMp4BoxSize ncopy;
    spMp4BoxSize total_ncopy;

    memset(dest_header, 0, sizeof(spMp4Header));

    if ((ncopy = spCopyMp4ChildBox((spChunk *)dest_header, (spChunk *)src_header, 0)) <= 0) {
        return ncopy;
    }
    total_ncopy = ncopy;

    spDebug(80, "spCopyMp4Header", "done: total_ncopy = %lu\n", (unsigned long)total_ncopy);
    
    return total_ncopy;
}

spMp4Box *spCreateMp4BoxWithCB(spMp4Box *parent, const char *type, spMp4BoxUpdateFunc update_func, void *data)
{
    spMp4Box *box;
    
    if (parent == NULL || strnone(type)) return NULL;
    
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }

    if ((box = (spMp4Box *)spCreateChunk(&sp_mp4_file_spec, (spChunk *)parent, NULL, type,
                                         SP_TRUE, SP_TRUE, (spChunkUpdateFunc)update_func, data)) == NULL) {
        return NULL;
    }

    return box;
}

spMp4Box *spCreateMp4Box(spMp4Box *parent, const char *type)
{
    return spCreateMp4BoxWithCB(parent, type, NULL, NULL);
}

spChunkInfoTable *spFindMp4BoxInfoTable(spMp4BoxHeader *parent_header, spMp4BoxHeader *header)
{
    if (sp_mp4_file_spec.num_list <= 0) {
        sp_mp4_file_spec.num_list = spArraySize(sp_mp4_box_info_list);
    }

    return spFindChunkInfoTable(sp_mp4_box_info_list, sp_mp4_file_spec.num_list, (spChunkHeader *)parent_header, (spChunkHeader *)header);
}

spMp4Box *spFindMp4BoxFromHeader(spMp4Header *header, const char *type, const char *parent_type)
{
    return (spMp4Box *)spFindChildChunk((spChunk *)header, type, parent_type);
}

spMp4Box *spFindMp4Box(spMp4Box *box, const char *type, const char *parent_type)
{
    return (spMp4Box *)spFindChunk((spChunk *)box, type, parent_type);
}

spMp4TrackBox *spFindMp4TrackBox(spMp4Header *header, long index, const char *handler_type)
{
    long count;
    char type_buf[4];
    spMp4TrackBox *trak;
    
    if (header == NULL || header->moov == NULL || header->moov->trak == NULL) {
        return NULL;
    }

    trak = header->moov->trak;
    count = 0;
    
    while (trak != NULL) {
        if (spEqMp4BoxType((spMp4Box *)trak, "trak")) {
            if (strnone(handler_type)
                || (spGetMp4BoxHandleType((spMp4Box *)trak, type_buf) == SP_TRUE
                    && strneq(type_buf, handler_type, strlen(handler_type)))) {
                if (index < 0 || count == index) {
                    return trak;
                }
                ++count;
            }
        }

        trak = (spMp4TrackBox *)trak->next;
    }
    
    return NULL;
}

unsigned long spGetMp4MaxTrackId(spMp4Header *header)
{
    spMp4TrackBox *trak;
    spMp4Box *child;
    unsigned long max_track_id = 0;

    if (header == NULL || header->moov == NULL) {
        return 0;
    }
    
    child = header->moov->child;

    while (child != NULL) {
        if (spEqMp4BoxType(child, "trak")) {
            trak = (spMp4TrackBox *)child;
            if (trak->tkhd->header.version == 1) {
                max_track_id = MAX(max_track_id, trak->tkhd->u.version1.track_id);
            } else {
                max_track_id = MAX(max_track_id, trak->tkhd->u.version0.track_id);
            }
        }
        
        child = child->next;
    }

    return max_track_id;
}

spMp4SampleTableBox *spGetMp4TrackSampleTableBox(spMp4TrackBox *trak)
{
    if (trak->mdia == NULL || trak->mdia->minf == NULL) {
        return NULL;
    }
    
    return trak->mdia->minf->stbl;
}

spMp4SampleEntryBox *spGetMp4TrackSampleEntryBox(spMp4TrackBox *trak, unsigned long index)
{
    spMp4SampleTableBox *stbl;
    
    if ((stbl = spGetMp4TrackSampleTableBox(trak)) == NULL) {
        return NULL;
    }
    
    if (stbl->stsd == NULL || stbl->stsd->entry_count <= 0 || index >= stbl->stsd->entry_count) {
        return NULL;
    }
    
    return &stbl->stsd->entries[index];
}

spMp4TrackBox *spAppendMp4SoundTrackBox(spMp4Header *header, unsigned long track_id,
                                        const char *codingname, unsigned short data_reference_index, 
                                        unsigned short channelcount, unsigned short samplesize,
                                        short compression_id, double samplerate, char stz2_field_size, spBool co64_flag,
                                        spMp4BoxUpdateFunc update_func, void *data)
{
    spMp4TrackBox *trak;
    spMp4BoxSize sebox_size;
    spMp4SampleEntryBox *sebox;
    spMp4SoundMediaHeaderBox *smhd;
    spMp4SampleTableBox *stbl;
    
    if (header == NULL || header->moov == NULL) {
        spDebug(10, "spAppendMp4SoundTrackBox", "failed: invalid MP4 header\n");
        return NULL;
    }

    spDebug(80, "spAppendMp4SoundTrackBox", "in\n");

    if (track_id <= 0) {
        track_id = header->moov->mvhd->next_track_id;
    }
    
    if ((trak = (spMp4TrackBox *)spCreateChunk(&sp_mp4_file_spec, (spChunk *)header->moov, NULL,
                                               "trak", SP_TRUE, SP_TRUE, NULL, NULL)) == NULL) {
        return NULL;
    }
    spDebug(80, "spAppendMp4SoundTrackBox", "spCreateChunk of 'trak' done\n");

    if (trak->tkhd->header.version == 1) {
        spUpdateMp4TrackHeaderBoxV1(trak->tkhd, 0x7, 0, 0, track_id, 0, 0, 0, 0, NULL, 0, 0);
    } else {
        spUpdateMp4TrackHeaderBoxV0(trak->tkhd, 0x7, 0, 0, track_id, 0, 0, 0, 0, NULL, 0, 0);
    }
    header->moov->mvhd->next_track_id = spGetMp4MaxTrackId(header) + 1;

    spUpdateMp4HandlerReferenceBox(trak->mdia->hdlr, "soun", NULL, SP_TRUE);

    if ((smhd = (spMp4SoundMediaHeaderBox *)spCreateChunk(&sp_mp4_file_spec,
                                                          (spChunk *)trak->mdia->minf, (spChunk *)trak->mdia->minf->dinf,
                                                          "smhd", SP_TRUE, SP_TRUE, NULL, NULL)) == NULL) {
        return NULL;
    }
    spDebug(80, "spAppendMp4SoundTrackBox", "spCreateChunk of 'smhd' done\n");
    
    stbl = trak->mdia->minf->stbl;
    
    if ((sebox = spAppendMp4SampleDescriptionEntry(stbl->stsd, codingname, data_reference_index)) == NULL) {
        return NULL;
    }
    spDebug(80, "spAppendMp4SoundTrackBox", "spAppendMp4SampleDescriptionEntry done\n");
    
    sebox->u.soun.version = 0;
    sebox->u.soun.channelcount = channelcount;
    sebox->u.soun.samplesize = samplesize;
    sebox->u.soun.compression_id = compression_id;
    sebox->u.soun.samplerate = (unsigned long)spRound(samplerate * 65536.0);

    if (trak->mdia->mdhd->header.version == 1) {
        spUpdateMp4MediaHeaderBoxV1(trak->mdia->mdhd, 0, 0, (unsigned long)spRound(samplerate), 0, 0, 0);
    } else {
        spUpdateMp4MediaHeaderBoxV0(trak->mdia->mdhd, 0, 0, (unsigned long)spRound(samplerate), 0, 0, 0);
    }

    if (header->moov->mvhd->header.version == 1) {
        if (header->moov->mvhd->u.version1.timescale == 0) {
            header->moov->mvhd->u.version1.timescale = trak->mdia->mdhd->u.version1.timescale;
        }
    } else {
        if (header->moov->mvhd->u.version0.timescale == 0) {
            header->moov->mvhd->u.version0.timescale = trak->mdia->mdhd->u.version0.timescale;
        }
    }

    sebox_size = 8/* after header */ + 2/* short */ * 6 + 4/* long */ * 2;
    spDebug(80, "spAppendMp4SoundTrackBox", "sebox_size = %lu\n", (unsigned long)sebox_size);
    spSetMp4BoxContentSize((spMp4Box *)sebox, sebox_size, SP_TRUE);
    
    spUpdateChildChunk(&sp_mp4_file_spec, (spChunk *)trak, (spChunkUpdateFunc)update_func, data);
    spDebug(80, "spAppendMp4SoundTrackBox", "spUpdateChildChunk done\n");

    spCreateMp4BoxWithCB((spMp4Box *)stbl, "stts", update_func, data);
    spCreateMp4BoxWithCB((spMp4Box *)stbl, "stsc", update_func, data);
    if (stz2_field_size == 4 || stz2_field_size == 8 || stz2_field_size == 16) {
        spMp4SampleSizeBox *stz2;
        stz2 = (spMp4SampleSizeBox *)spCreateMp4BoxWithCB((spMp4Box *)stbl, "stz2", update_func, data);
        stz2->field_size = stz2_field_size;
    } else {
        spCreateMp4BoxWithCB((spMp4Box *)stbl, "stsz", update_func, data);
    }
    spCreateMp4BoxWithCB((spMp4Box *)stbl, co64_flag ? "co64" : "stco", update_func, data);
    
    spDebug(80, "spAppendMp4SoundTrackBox", "done\n");
    
    return trak;
}

spFileOffset spGetMp4MdatContentOffset(spMp4Header *header)
{
    return header->mdat->offset + spGetMp4BoxHeaderSize(header->mdat);
}

unsigned long spConvertMp4TimeToSample(spMp4TimeToSampleBox *stts, spULong t, unsigned long *remainder)
{
    unsigned long i;
    spULong duration, next_duration;
    unsigned long sample, next_sample;
    unsigned long t_delta;
    unsigned long sample_shift;
    
    if (stts == NULL || stts->entry_count <= 0) return 0;

    if (remainder != NULL) {
        *remainder = 0;
    }
    
    sample = 0;
    duration = 0;
        
    for (i = 0; i < stts->entry_count; i++) {
        next_duration = duration + stts->entries[i].sample_count * stts->entries[i].sample_delta;
        next_sample = sample + stts->entries[i].sample_count;
        
        if (next_duration > t) {
            t_delta = (unsigned long)(t - duration);
            sample_shift = t_delta / stts->entries[i].sample_delta;
            
            if (remainder != NULL) {
                *remainder = t_delta - sample_shift * stts->entries[i].sample_delta;
            }

            sample += sample_shift;
            
            break;
        }

        sample = next_sample;
        duration = next_duration;
    }

    return sample;
}

unsigned long spConvertMp4ChunkToSample(spMp4SampleToChunkBox *stsc, unsigned long chunk,
                                        unsigned long *samples_per_chunk, unsigned long *description_index)
{
    unsigned long i;
    unsigned long chunk_diff;
    unsigned long sample_offset;
    
    sample_offset = 0;
    for (i = 1; i < stsc->entry_count; i++) {
        if (stsc->entries[i].first_chunk > chunk) {
            break;
        }

        chunk_diff = stsc->entries[i].first_chunk - stsc->entries[i - 1].first_chunk;
        sample_offset += stsc->entries[i - 1].samples_per_chunk * chunk_diff;
    }

    chunk_diff = chunk - stsc->entries[i - 1].first_chunk;
    sample_offset += stsc->entries[i - 1].samples_per_chunk * chunk_diff;
    
    if (samples_per_chunk != NULL) *samples_per_chunk = stsc->entries[i - 1].samples_per_chunk;
    if (description_index != NULL) *description_index = stsc->entries[i - 1].sample_description_index;

    return sample_offset;
}

/* returns 1-based index of the chunk */
unsigned long spConvertMp4SampleToChunk(spMp4SampleToChunkBox *stsc, unsigned long description_index, unsigned long sample, 
                                        unsigned long *remainder)
{
    unsigned long i, j;
    unsigned long sample_delta;
    unsigned long delta_chunk;
    unsigned long total_sample, next_total_sample;
    unsigned long chunk, next_chunk;
    unsigned long chunk_shift;
    
    if (stsc == NULL || stsc->entry_count <= 0) return 0;

    if (description_index <= 0) {
        description_index = stsc->entries[0].sample_description_index;
    }
    
    if (remainder != NULL) {
        *remainder = 0;
    }
    
    chunk = 0;
    total_sample = 0;

    for (i = 0; i < stsc->entry_count; i++) {
        spDebug(100, "spConvertMp4SampleToChunk", "stsc->entries[%ld] first_chunk = %lu, samples_per_chunk = %lu\n",
                i, stsc->entries[i].first_chunk, stsc->entries[i].samples_per_chunk);
        spDebug(100, "spConvertMp4SampleToChunk", "stsc->entries[%ld].sample_description_index = %lu / %lu\n",
                i, stsc->entries[i].sample_description_index, description_index);
        if (stsc->entries[i].sample_description_index == description_index) {
            next_chunk = chunk;
            for (j = i; j < stsc->entry_count; j++) {
                if (stsc->entries[j].sample_description_index == description_index) {
                    if (chunk == 0) {
                        chunk = stsc->entries[j].first_chunk;
                    } else {
                        next_chunk = stsc->entries[j].first_chunk;
                        break;
                    }
                }
            }

            delta_chunk = MAX(next_chunk - chunk, 1);
            spDebug(100, "spConvertMp4SampleToChunk", "delta_chunk = %lu (%lu - %lu)\n", delta_chunk, next_chunk, chunk);
            next_total_sample = total_sample + delta_chunk * stsc->entries[i].samples_per_chunk;
            spDebug(100, "spConvertMp4SampleToChunk", "sample = %lu, total_sample = %lu-%lu, samples_per_chunk = %lu\n",
                    sample, total_sample, next_total_sample, stsc->entries[i].samples_per_chunk);

            if (next_total_sample > sample || next_chunk == chunk) {
                sample_delta = sample - total_sample;
                chunk_shift = sample_delta / stsc->entries[i].samples_per_chunk;
                spDebug(60, "spConvertMp4SampleToChunk", "found: sample_delta = %lu, chunk_shift = %lu\n",
                        sample_delta, chunk_shift);
                
                if (remainder != NULL) {
                    *remainder = sample_delta - chunk_shift * stsc->entries[i].samples_per_chunk;
                    spDebug(60, "spConvertMp4SampleToChunk", "remainder = %lu\n", *remainder);
                }
                
                chunk += chunk_shift;
            
                break;
            }

            chunk = next_chunk;
            total_sample = next_total_sample;
        }
    }

    spDebug(60, "spConvertMp4SampleToChunk", "done: chunk = %lu\n", chunk);
    
    return chunk;
}

spFileOffset spConvertMp4SampleToFileOffset(spMp4SampleToChunkBox *stsc, unsigned long description_index, unsigned long sample, 
                                            unsigned long *chunk, unsigned long *remainder, unsigned long *sample_size)
{
    spFileOffset foffset;
    unsigned long tchunk, tremainder;
    unsigned long current_sample;
    unsigned long tsample_size;
    spMp4Box *box;
    spMp4ChunkOffsetBox *stco;
    spMp4SampleSizeBox *stsz;

    if (stsc == NULL || stsc->parent == NULL) return 0;

    stco = NULL;
    stsz = NULL;
    box = stsc->parent->child;
    
    while (box != NULL) {
        if (strneq(box->header.type, "stco", 4)
            || strneq(box->header.type, "co64", 4)) {
            stco = (spMp4ChunkOffsetBox *)box;
        } else if (strneq(box->header.type, "stsz", 4)
                   || strneq(box->header.type, "stz2", 4)) {
            stsz = (spMp4SampleSizeBox *)box;
        }

        if (stco != NULL && stsz != NULL) {
            break;
        }
        
        box = box->next;
    }

    if (stco == NULL || stsz == NULL) {
        spDebug(10, "spConvertMp4SampleToFileOffset", "can't find stco or stsz.\n");
        return 0;
    }

    tsample_size = 0;
    
    if ((tchunk = spConvertMp4SampleToChunk(stsc, description_index, sample, &tremainder)) == 0
        || stco->entry_count < tchunk) {
        spDebug(10, "spConvertMp4SampleToFileOffset",
                "spConvertMp4SampleToChunk failed: tchunk = %lu, stco->entry_count = %lu\n", tchunk, stco->entry_count);
        foffset = 0;
        tchunk = 0;
        tremainder = 0;
    } else {
        foffset = stco->chunk_offset[tchunk - 1];
        spDebug(80, "spConvertMp4SampleToFileOffset", "foffset = %lu, tchunk = %lu, sample = %lu, tremainder = %lu\n",
                (unsigned long)foffset, tchunk, sample, tremainder);

        for (current_sample = sample - tremainder;; ++current_sample) {
            if ((tsample_size = spGetMp4SampleSize(stsz, current_sample)) <= 0) {
                spDebug(10, "spConvertMp4SampleToFileOffset", "spGetMp4SampleSize failed.\n");
                foffset = 0;
                tchunk = 0;
                tremainder = 0;
                break;
            }

            if (current_sample >= sample) {
                break;
            }

            foffset += tsample_size;
        }
    }

    if (chunk != NULL) *chunk = tchunk;
    if (remainder != NULL) *remainder = tremainder;
    if (sample_size != NULL) *sample_size = tsample_size;

    spDebug(80, "spConvertMp4SampleToFileOffset", "done: foffset = %lu, tsample_size = %lu\n",
            (unsigned long)foffset, tsample_size);
    
    return foffset;
}

spMp4TrackBox *spFindMp4NextChunkFromFileOffset(spMp4Header *header, spFileOffset offset,
                                                spFileOffset *o_offset, unsigned long *chunk /* 1-based */)
{
    unsigned long i, j;
    spMp4TrackBox *trak;
    spMp4TrackBox *opt_trak;
    spMp4SampleTableBox *stbl;
    spFileOffset diff;
    spFileOffset min_diff;
    spFileOffset opt_offset;
    unsigned long opt_chunk;

    if (header == NULL) return NULL;

    opt_chunk = 0;
    opt_trak = NULL;
    opt_offset = offset;
    min_diff = 0x7fffffff;
    
    for (i = 0;; i++) {
        if ((trak = spFindMp4TrackBox(header, i, NULL)) == NULL) {
            break;
        }
        spDebug(80, "spFindMp4NextChunkFromFileOffset", "i = %lu, offset = %lu\n", i, (unsigned long)offset);

        if ((stbl = spGetMp4TrackSampleTableBox(trak)) != NULL) {
            for (j = 0; j < stbl->stco->entry_count; j++) {
                spDebug(100, "spFindMp4NextChunkFromFileOffset", "chunk_offset[%lu] = %lu, offset = %lu\n",
                        j, (long)stbl->stco->chunk_offset[j], (unsigned long)offset);
                if (stbl->stco->chunk_offset[j] >= offset) {
                    diff = stbl->stco->chunk_offset[j] - offset;
                    spDebug(100, "spFindMp4NextChunkFromFileOffset", "diff = %lu\n", (unsigned long)diff);
                    if (diff < min_diff) {
                        min_diff = diff;
                        opt_trak = trak;
                        opt_chunk = j + 1;
                        opt_offset = stbl->stco->chunk_offset[j];
                    }
                    break;
                }
            }
        }
        
        if (min_diff == 0) {
            break;
        }
    }

    spDebug(100, "spFindMp4NextChunkFromFileOffset", "done: opt_offset = %lu, opt_chunk = %lu\n",
            (unsigned long)opt_offset, opt_chunk);
    
    if (o_offset != NULL) *o_offset = opt_offset;
    if (chunk != NULL) *chunk = opt_chunk;

    return opt_trak;
}

spBool spParseMp4SongInfo(spMp4MetaBox *meta, spSongInfoV2 *song_info)
{
    unsigned short number, total;
    long tempo;
    
    if (meta == NULL || meta->ilst == NULL || song_info == NULL) return SP_FALSE;

    spInitSongInfoV2(song_info);

    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_TITLE, (unsigned char *)song_info->title, sizeof(song_info->title))) {
        song_info->info_mask |= SP_SONG_TITLE_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_ARTIST, (unsigned char *)song_info->artist, sizeof(song_info->artist))) {
        song_info->info_mask |= SP_SONG_ARTIST_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_ALBUM, (unsigned char *)song_info->album, sizeof(song_info->album))) {
        song_info->info_mask |= SP_SONG_ALBUM_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_GENRE, (unsigned char *)song_info->genre, sizeof(song_info->genre))) {
        song_info->info_mask |= SP_SONG_GENRE_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_ALBUM_ARTIST, (unsigned char *)song_info->album_artist, sizeof(song_info->album_artist))) {
        song_info->info_mask |= SP_SONG_ALBUM_ARTIST_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_RELEASE, (unsigned char *)song_info->release, sizeof(song_info->release))) {
        song_info->info_mask |= SP_SONG_RELEASE_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_COPYRIGHT, (unsigned char *)song_info->copyright, sizeof(song_info->copyright))) {
        song_info->info_mask |= SP_SONG_COPYRIGHT_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_COMPOSER, (unsigned char *)song_info->composer, sizeof(song_info->composer))) {
        song_info->info_mask |= SP_SONG_COMPOSER_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_SOFTWARE, (unsigned char *)song_info->software, sizeof(song_info->software))) {
        song_info->info_mask |= SP_SONG_SOFTWARE_MASK;
    }
    if (spGetMp4MetaDataUTF8StringToLocaleCode(meta, SP_MP4_META_DATA_TYPE_COMMENT, (unsigned char *)song_info->comment, sizeof(song_info->comment))) {
        song_info->info_mask |= SP_SONG_COMMENT_MASK;
    }

    number = total = 0;
    if (spGetMp4MetaDataTrackNumber(meta, &number, &total)) {
        if (number > 0) {
            song_info->track = number;
            song_info->info_mask |= SP_SONG_TRACK_MASK;
        }
        if (total > 0) {
            song_info->track_total = total;
            song_info->info_mask |= SP_SONG_TRACK_TOTAL_MASK;
        }
    }

    number = total = 0;
    if (spGetMp4MetaDataDiscNumber(meta, &number, &total)) {
        if (number > 0) {
            song_info->disc = number;
            song_info->info_mask |= SP_SONG_DISC_MASK;
        }
        if (total > 0) {
            song_info->disc_total = total;
            song_info->info_mask |= SP_SONG_DISC_TOTAL_MASK;
        }
    }

    tempo = 0;
    if (spGetMp4MetaDataTempo(meta, &tempo)) {
        song_info->tempo = tempo;
        song_info->info_mask |= SP_SONG_TEMPO_MASK;
    }
    
    if (song_info->info_mask) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spAppendMp4SongInfo(spMp4MetaBox *meta, spSongInfoV2 *song_info)
{
    unsigned short number, total;
    
    if (meta == NULL || song_info == NULL || song_info->info_mask == 0) return SP_FALSE;

    spDebug(10, "spAppendMp4SongInfo", "song_info->info_mask = %lx\n", song_info->info_mask);
    
    if (meta->ilst == NULL) {
        if (spCreateMp4Box((spMp4Box *)meta, "ilst") == NULL) {
            return SP_FALSE;
        }
    }

    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_TITLE, 0, 0, (unsigned char *)song_info->title);
    }
    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_TITLE, 0, 0, (unsigned char *)song_info->title);
    }
    if (song_info->info_mask & SP_SONG_ARTIST_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_ARTIST, 0, 0, (unsigned char *)song_info->artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_ARTIST_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_ALBUM_ARTIST, 0, 0, (unsigned char *)song_info->album_artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_ALBUM, 0, 0, (unsigned char *)song_info->album);
    }
    if (song_info->info_mask & SP_SONG_COMPOSER_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_COMPOSER, 0, 0, (unsigned char *)song_info->composer);
    }
    if (song_info->info_mask & SP_SONG_GENRE_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_GENRE, 0, 0, (unsigned char *)song_info->genre);
    }
    if (song_info->info_mask & SP_SONG_RELEASE_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_RELEASE, 0, 0, (unsigned char *)song_info->release);
    }
    if (song_info->info_mask & SP_SONG_SOFTWARE_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_SOFTWARE, 0, 0, (unsigned char *)song_info->software);
    }
    if (song_info->info_mask & SP_SONG_COPYRIGHT_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_COPYRIGHT, 0, 0, (unsigned char *)song_info->copyright);
    }
    if (song_info->info_mask & SP_SONG_COMMENT_MASK) {
        spAppendMp4MetaDataUTF8StringFromLocaleCode(meta, SP_MP4_META_DATA_TYPE_COMMENT, 0, 0, (unsigned char *)song_info->comment);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_TRACK_MASK) {
        number = (unsigned short)song_info->track;
        if (song_info->info_mask & SP_SONG_TRACK_TOTAL_MASK) {
            total = (unsigned short)song_info->track_total;
        }
        spAppendMp4MetaDataTrackNumber(meta, number, total);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_DISC_MASK) {
        number = (unsigned short)song_info->disc;
        if (song_info->info_mask & SP_SONG_DISC_TOTAL_MASK) {
            total = (unsigned short)song_info->disc_total;
        }
        spAppendMp4MetaDataDiscNumber(meta, number, total);
    }

    if (song_info->info_mask & SP_SONG_TEMPO_MASK && song_info->tempo > 0) {
        spAppendMp4MetaDataTempo(meta, song_info->tempo);
    }

    return SP_TRUE;
}

static void showMp4MatrixInfo(long *matrix)
{
    long k;
    double dmatrix[9];
    
    spConvMp4MatrixToDouble(matrix, dmatrix);

    printf("Matrix =");
    for (k = 0; k < 9; k++) {
        printf(" %.2f", dmatrix[k]);
    }
    
    return;
}

static void showMp4BoxInfo(spMp4Box *box, long depth)
{
    long k;
    spMp4Box *child_box;
    char type_string[5];
    char buf[SP_MAX_MESSAGE];
    spMp4BoxSize box_size;
    spMp4BoxSize content_size;
    spMp4BoxSize margin_size;

    box_size = spGetMp4BoxSize(box, type_string);
    content_size = spGetMp4BoxContentSize(box, NULL);

    if (box_size > 0) {
        margin_size = spGetMp4BoxContentMarginSize(box);
        
        for (k = 1; k < depth; k++) printf("  ");
        if (margin_size > 0) {
            printf("'%s' (size: %lu (%lu + %lu) bytes, margin: %lu bytes)\n",
                   type_string, (unsigned long)box_size, (unsigned long)(box_size - content_size), (unsigned long)content_size, (unsigned long)margin_size);
        } else {
            printf("'%s' (size: %lu (%lu + %lu) bytes)\n", type_string, (unsigned long)box_size, (unsigned long)(box_size - content_size), (unsigned long)content_size);
        }

        if (strneq(type_string, "ftyp", 4)) {
            long n;
            spMp4FileTypeBox *ftyp = (spMp4FileTypeBox *)box;
        
            printf("  '%c%c%c%c' (minor version %ld)\n",
                   ftyp->major_brand[0], ftyp->major_brand[1],
                   ftyp->major_brand[2], ftyp->major_brand[3],
                   ftyp->minor_version);

            for (n = 0; n < ftyp->nbrands; n++) {
                printf("\tcompatible_brands[%ld] = '%c%c%c%c'\n", n,
                       ftyp->compatible_brands[n * 4 + 0],
                       ftyp->compatible_brands[n * 4 + 1],
                       ftyp->compatible_brands[n * 4 + 2],
                       ftyp->compatible_brands[n * 4 + 3]);
            }
        } else if (strneq(type_string, "mvhd", 4)) {
            spMp4MovieHeaderBox *mvhd = (spMp4MovieHeaderBox *)box;
            
            for (k = 1; k <= depth; k++) printf("  ");
            
            if (mvhd->header.version == 1) {
                double mvhd_duration;
                
#ifdef SP_SUPPORT_64BIT_INT
                mvhd_duration = (double)spC64ToLong64(mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#else
                mvhd_duration = (double)spC64ToULong(mvhd->u.version1.duration, SP_AIFF_NEED_SWAP);
#endif
                
                printf("Version 1, Time scale = %lu, Duration = %.0f, ",
                       mvhd->u.version1.timescale, mvhd_duration);
            } else {
                printf("Time scale = %lu, Duration = %lu, ",
                       mvhd->u.version0.timescale, mvhd->u.version0.duration);
            }
#if 0
            printf("Preferred rate = %lx, Preferred volume = %x\n", mvhd->rate, mvhd->volume);
#else
            printf("Preferred rate = %.2f, Preferred volume = %.2f\n",
                   (double)mvhd->rate / (double)(1 << 16), (double)mvhd->volume / (double)(1 << 8));
#endif

            if (memcmp(mvhd->matrix, sp_mp4_default_matrix, sizeof(long) * 9) != 0) {
                for (k = 1; k <= depth; k++) printf("  ");
                showMp4MatrixInfo(mvhd->matrix);
                printf("\n");
            }
        } else if (strneq(type_string, "trak", 4)) {
            spMp4SampleEntryBox *sebox;
            spMp4TrackBox *trak = (spMp4TrackBox *)box;
            
            spGetMp4BoxHandleType(box, buf);
            for (k = 1; k <= depth; k++) printf("  ");
            printf("Track type: '%c%c%c%c'\n", buf[0], buf[1], buf[2], buf[3]);

            if ((sebox = spGetMp4TrackSampleEntryBox(trak, 0)) != NULL) {
                if (strneq(buf, "soun", 4)) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("  channel = %d, sample size = %d, sample rate = %.2f\n",
                           sebox->u.soun.channelcount, sebox->u.soun.samplesize,
                           (double)sebox->u.soun.samplerate / 65536.0);

                    if (sebox->u.soun.version >= 1) {
                        for (k = 1; k <= depth; k++) printf("  ");
                        printf("  samples/packet = %lu, bytes/packet = %lu, bytes/frame = %lu\n",
                               sebox->u.soun.samples_per_packet, sebox->u.soun.bytes_per_packet,
                               sebox->u.soun.bytes_per_sample);
                    }
                } else if (strneq(buf, "vide", 4)) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("  width = %d, height = %d, H resulution = %.2f, V resolution = %.2f, frame count = %d, depth = %d\n",
                           sebox->u.vide.width, sebox->u.vide.height,
                           (double)sebox->u.vide.horizresolution / 65536.0, (double)sebox->u.vide.vertresolution / 65536.0,
                           sebox->u.vide.frame_count, sebox->u.vide.depth);
                }
            }
        } else if (strneq(type_string, "tkhd", 4)) {
            spMp4TrackHeaderBox *tkhd = (spMp4TrackHeaderBox *)box;

#if 0
            for (k = 1; k <= depth; k++) printf("  ");
            printf("  layer = %d, alternate group = %d, volume = %d, width = %lu, height = %lu\n",
                   tkhd->layer, tkhd->alternate_group, tkhd->volume,
                   tkhd->width, tkhd->height);
#endif
                    
            if (memcmp(tkhd->matrix, sp_mp4_default_matrix, sizeof(long) * 9) != 0) {
                for (k = 1; k <= depth; k++) printf("  ");
                printf("  ");
                showMp4MatrixInfo(tkhd->matrix);
                printf("\n");
            }
        } else if (strneq(type_string, "sdp ", 4)
                   || (strneq(type_string, "rtp ", 4) && spEqMp4BoxType(box->parent, "hnti"))) {
            spMp4DataOnlyBox *dobox = (spMp4DataOnlyBox *)box;
            for (k = 1; k <= depth; k++) printf("  ");
            printf("SDP String:\n%s\n", dobox->data);
        } else if (strneq(type_string, "data", 4)) {
            spMp4MetaDataValueBox *databox = (spMp4MetaDataValueBox *)box;

            if (spEqMp4BoxType(databox->parent, "trkn")) {
                unsigned short number, total;
                if (spParseMp4MetaDataTrackNumber(databox, &number, &total)) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("Track Number: %d / %d\n", number, total);
                }
            } else if (spEqMp4BoxType(databox->parent, "disc")) {
                unsigned short number, total;
                if (spParseMp4MetaDataDiscNumber(databox, &number, &total)) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("Disc Number: %d / %d\n", number, total);
                }
            } else if (databox->type_indicator == SP_MP4_META_DATA_TYPE_INDICATOR_UTF8) { /* UTF8 */
                spStrCopy(buf, sizeof(buf), databox->value);
                spConvertKanjiToLocaleCode((unsigned char *)buf, sizeof(buf), SP_KANJI_CODE_UTF8);
                for (k = 1; k <= depth; k++) printf("  ");
                printf("UTF-8 String: '%s'\n", buf);
            } else if (databox->type_indicator == SP_MP4_META_DATA_TYPE_INDICATOR_BE_SIGNED_INT) {
                unsigned long nbyte;
                long number;
                
                if ((nbyte = spParseMp4MetaDataBESignedInt(databox, &number)) > 0) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("Signed Integer (%lu bytes): %ld\n", nbyte, number);
                }
            } else if (databox->type_indicator == SP_MP4_META_DATA_TYPE_INDICATOR_BE_UNSIGNED_INT) {
                unsigned long nbyte;
                unsigned long number;
                
                if ((nbyte = spParseMp4MetaDataBEUnsignedInt(databox, &number)) > 0) {
                    for (k = 1; k <= depth; k++) printf("  ");
                    printf("Unsigned Integer (%lu bytes): %lu\n", nbyte, number);
                }
            } else {
                for (k = 1; k <= depth; k++) printf("  ");
                printf("Data Type: %lu\n", databox->type_indicator);
            }
        } else if (strneq(type_string, "hdlr", 4)) {
            spMp4HandlerReferenceBox *hrbox = (spMp4HandlerReferenceBox *)box;
            
            for (k = 1; k <= depth; k++) printf("  ");
            printf("Handler: '%c%c%c%c', String: '%s'\n",
                   hrbox->handler_type[0], hrbox->handler_type[1],
                   hrbox->handler_type[2], hrbox->handler_type[3],
                   hrbox->name);
        } else if (strneq(type_string, "url ", 4)
                   || strneq(type_string, "mean", 4)
                   || strneq(type_string, "name", 4)) {
            spMp4DataOnlyBox *dobox = (spMp4DataOnlyBox *)box;
            
            for (k = 1; k <= depth; k++) printf("  ");
            printf("String: '%s'\n", dobox->data);
        }
    }
    
    child_box = box->child;
    while (child_box != NULL) {
        showMp4BoxInfo(child_box, depth + 1);
        child_box = child_box->next;
    }

    return;
}

void spShowMp4HeaderInfo(spMp4Header *header)
{
    showMp4BoxInfo((spMp4Box *)header, 0);
    return;
}

#if 1
static long double_to_fixedlong(double value, long fracbits)
{
    return (long)spRound(value * (1 << fracbits));
}

static double fixedlong_to_double(long value, long fracbits)
{
    return (double)value / (double)(1 << fracbits);
}

void spConvMp4MatrixToDouble(long *imatrix, double *omatrix)
{
    int i;

    for (i = 0; i < 6; i++) {
        omatrix[i] = fixedlong_to_double(imatrix[i], 16);
    }
    for (; i < 9; i++) {
        omatrix[i] = fixedlong_to_double(imatrix[i], 30);
    }
    
    return;
}

void spConvDoubleMatrixToMp4(double *imatrix, long *omatrix)
{
    int i;

    for (i = 0; i < 6; i++) {
        omatrix[i] = double_to_fixedlong(imatrix[i], 16);
    }
    for (; i < 9; i++) {
        omatrix[i] = double_to_fixedlong(imatrix[i], 30);
    }
    
    return;
}

void spMultMp4MatrixToXY(long *matrix, long *x, long *y)
{
    long ox, oy;

    ox = (long)spRound((double)(*x * matrix[0] + *y * matrix[3] + matrix[6]) / (double)(1 << 16));
    oy = (long)spRound((double)(*x * matrix[1] + *y * matrix[4] + matrix[7]) / (double)(1 << 16));

    *x = ox;
    *y = oy;

    return;
}

void spMultMp4MatrixToXYZ(long *matrix, long *x, long *y, long *z)
{
    long ox, oy, oz;

    ox = (long)spRound((double)(*x * matrix[0] + *y * matrix[3] + matrix[6]) / (double)(1 << 16));
    oy = (long)spRound((double)(*x * matrix[1] + *y * matrix[4] + matrix[7]) / (double)(1 << 16));
    oz = (long)spRound((double)(*x * matrix[2] + *y * matrix[5] + matrix[8]) / (double)(1 << 30));

    *x = ox;
    *y = oy;
    *z = oz;
    
    return;
}
#endif
